/* SPDX-License-Identifier: GPL-2.0 */
/* Copyright (C) 2008  PaX Team <pageexec@freemail.hu> */

#ifndef _ASM_X86_UACCESS_64_H
#define _ASM_X86_UACCESS_64_H

/*
 * User space memory access functions
 */
#include <linux/compiler.h>
#include <linux/lockdep.h>
#include <linux/kasan-checks.h>
#include <asm/alternative.h>
#include <asm/cpufeatures.h>
#include <asm/page.h>

/*
 * Copy To/From Userspace
 */

/* Handles exceptions in both to and from, but doesn't do access_ok */
__must_check unsigned long
copy_user_enhanced_fast_string(void *to, const void *from, unsigned long len);
__must_check unsigned long
copy_user_generic_string(void *to, const void *from, unsigned long len);
__must_check unsigned long
copy_user_generic_unrolled(void *to, const void *from, unsigned long len);

static __always_inline __must_check __size_overflow(3) unsigned long
copy_user_generic(void *to, const void *from, unsigned long len)
{
	unsigned long ret;

	/*
	 * If CPU has ERMS feature, use copy_user_enhanced_fast_string.
	 * Otherwise, if CPU has rep_good feature, use copy_user_generic_string.
	 * Otherwise, use copy_user_generic_unrolled.
	 */
	alternative_call_2(copy_user_generic_unrolled,
			 copy_user_generic_string,
			 X86_FEATURE_REP_GOOD,
			 copy_user_enhanced_fast_string,
			 X86_FEATURE_ERMS,
			 ASM_OUTPUT2("=a" (ret), "=D" (to), "=S" (from),
				     "=d" (len)),
			 "1" (to), "2" (from), "3" (len)
			 : "memory", "rcx", "r8", "r9", "r10", "r11");
	return ret;
}

static __always_inline __must_check unsigned long
copy_to_user_mcsafe(void *to, const void *from, size_t len)
{
	unsigned long ret;

	__uaccess_begin();
	/*
	 * Note, __memcpy_mcsafe() is explicitly used since it can
	 * handle exceptions / faults.  memcpy_mcsafe() may fall back to
	 * memcpy() which lacks this handling.
	 */
	ret = __memcpy_mcsafe(to, from, len);
	__uaccess_end();
	return ret;
}

static __always_inline __must_check unsigned long
raw_copy_from_user(void *dst, const void __user *src, unsigned long size)
{
	unsigned long ret = 0;

	access_ok_nospec(src, size);

#ifdef CONFIG_PAX_MEMORY_UDEREF
	if (unlikely(segment_eq(get_fs(), KERNEL_DS) && (unsigned long)src < TASK_SIZE_MAX))
		return size;
	if (!access_ok(src, size))
		return size;
#endif

	if (!__builtin_constant_p(size))
		return copy_user_generic(dst, (__force_kernel const void *)src, size);
	switch (size) {
	case 1:
		__uaccess_begin();
		__get_user_asm_nozero(*(u8 *)dst, (const u8 __user *)src,
			      ret, "b", "b", "=q", 1);
		__uaccess_end();
		return ret;
	case 2:
		__uaccess_begin();
		__get_user_asm_nozero(*(u16 *)dst, (const u16 __user *)src,
			      ret, "w", "w", "=r", 2);
		__uaccess_end();
		return ret;
	case 4:
		__uaccess_begin();
		__get_user_asm_nozero(*(u32 *)dst, (const u32 __user *)src,
			      ret, "l", "k", "=r", 4);
		__uaccess_end();
		return ret;
	case 8:
		__uaccess_begin();
		__get_user_asm_nozero(*(u64 *)dst, (const u64 __user *)src,
			      ret, "q", "", "=r", 8);
		__uaccess_end();
		return ret;
	case 10: {
		unsigned long tmp = 0;

		__uaccess_begin();
		__get_user_asm_nozero(*(u64 *)dst, (const u64 __user *)src,
			       tmp, "q", "", "=r", 10);
		__get_user_asm_nozero(*(u16 *)(8 + (char *)dst),
			       (const u16 __user *)(8 + (const char __user *)src),
			       ret, "w", "w", "=r", 2);
		ret |= tmp;
		__uaccess_end();
		return ret;
	}
	case 16: {
		unsigned long tmp = 0;

		__uaccess_begin();
		__get_user_asm_nozero(*(u64 *)dst, (const u64 __user *)src,
			       tmp, "q", "", "=r", 16);
		__get_user_asm_nozero(*(u64 *)(8 + (char *)dst),
			       (const u64 __user *)(8 + (const char __user *)src),
			       ret, "q", "", "=r", 8);
		ret |= tmp;
		__uaccess_end();
		return ret;
	}
	default:
		return copy_user_generic(dst, (__force_kernel const void *)src, size);
	}
}

static __always_inline __must_check unsigned long
raw_copy_to_user(void __user *dst, const void *src, unsigned long size)
{
	unsigned long ret = 0;

#ifdef CONFIG_PAX_MEMORY_UDEREF
	if (unlikely(segment_eq(get_fs(), KERNEL_DS) && (unsigned long)dst < TASK_SIZE_MAX))
		return size;
	if (!access_ok(dst, size))
		return size;
#endif

	if (!__builtin_constant_p(size))
		return copy_user_generic((__force_kernel void *)dst, src, size);
	switch (size) {
	case 1:
		__uaccess_begin();
		__put_user_asm(*(const u8 *)src, (u8 __user *)dst,
			      ret, "b", "b", "iq", 1);
		__uaccess_end();
		return ret;
	case 2:
		__uaccess_begin();
		__put_user_asm(*(const u16 *)src, (u16 __user *)dst,
			      ret, "w", "w", "ir", 2);
		__uaccess_end();
		return ret;
	case 4:
		__uaccess_begin();
		__put_user_asm(*(const u32 *)src, (u32 __user *)dst,
			      ret, "l", "k", "ir", 4);
		__uaccess_end();
		return ret;
	case 8:
		__uaccess_begin();
		__put_user_asm(*(const u64 *)src, (u64 __user *)dst,
			      ret, "q", "", "er", 8);
		__uaccess_end();
		return ret;
	case 10:
		__uaccess_begin();
		__put_user_asm(*(const u64 *)src, (u64 __user *)dst,
			       ret, "q", "", "er", 10);
		if (likely(!ret)) {
			asm("":::"memory");
			__put_user_asm(4[(const u16 *)src], 4 + (u16 __user *)dst,
				       ret, "w", "w", "ir", 2);
		}
		__uaccess_end();
		return ret;
	case 16:
		__uaccess_begin();
		__put_user_asm(*(const u64 *)src, (u64 __user *)dst,
			       ret, "q", "", "er", 16);
		if (likely(!ret)) {
			asm("":::"memory");
			__put_user_asm(1[(const u64 *)src], 1 + (u64 __user *)dst,
				       ret, "q", "", "er", 8);
		}
		__uaccess_end();
		return ret;
	default:
		return copy_user_generic((__force_kernel void *)dst, src, size);
	}
}

static __always_inline __must_check
unsigned long raw_copy_in_user(void __user *dst, const void __user *src, unsigned long size)
{
	return copy_user_generic((__force_kernel void *)dst,
				 (__force_kernel void *)src, size);
}

extern unsigned long __copy_user_nocache(void *dst, const void __user *src,
				unsigned long size, int zerorest) __size_overflow(3);

extern long __copy_user_flushcache(void *dst, const void __user *src, size_t size) __size_overflow(3);
extern void memcpy_page_flushcache(char *to, struct page *page, size_t offset,
			   size_t len);

static inline unsigned long
__copy_from_user_inatomic_nocache(void *dst, const void __user *src,
				  unsigned long size)
{
	kasan_check_write(dst, size);

	if (size > INT_MAX)
		return size;

	if (check_object_size(dst, size, false))
		return size;

#ifdef CONFIG_PAX_MEMORY_UDEREF
	if (!access_ok(src, size))
		return size;
#endif

	return __copy_user_nocache(dst, src, size, 0);
}

static inline int
__copy_from_user_flushcache(void *dst, const void __user *src, size_t size)
{
	kasan_check_write(dst, size);
	return __copy_user_flushcache(dst, src, size);
}

unsigned long
mcsafe_handle_tail(char *to, char *from, unsigned long len) __size_overflow(3);

#endif /* _ASM_X86_UACCESS_64_H */
