/* SPDX-License-Identifier: GPL-2.0 */
/* thread_info.h: common low-level thread information accessors
 *
 * Copyright (C) 2002  David Howells (dhowells@redhat.com)
 * - Incorporating suggestions made by Linus Torvalds
 */

#ifndef _LINUX_THREAD_INFO_H
#define _LINUX_THREAD_INFO_H

#include <linux/types.h>
#include <linux/bug.h>
#include <linux/restart_block.h>
#include <linux/errno.h>

#ifdef CONFIG_THREAD_INFO_IN_TASK
/*
 * For CONFIG_THREAD_INFO_IN_TASK kernels we need <asm/current.h> for the
 * definition of current, but for !CONFIG_THREAD_INFO_IN_TASK kernels,
 * including <asm/current.h> can cause a circular dependency on some platforms.
 */
#include <asm/current.h>
#ifndef current_thread_info
static __always_inline struct thread_info *current_thread_info(void);
#endif
#endif

#include <linux/bitops.h>

/*
 * For per-arch arch_within_stack_frames() implementations, defined in
 * asm/thread_info.h.
 */
enum {
	BAD_STACK = -1,
	NOT_STACK = 0,
	GOOD_FRAME,
	GOOD_STACK,
};

#include <asm/thread_info.h>

#ifdef __KERNEL__

#ifndef arch_set_restart_data
#define arch_set_restart_data(restart) do { } while (0)
#endif

static inline long set_restart_fn(struct restart_block *restart,
					long (*fn)(struct restart_block *))
{
	restart->fn = fn;
	arch_set_restart_data(restart);
	return -ERESTART_RESTARTBLOCK;
}

#ifndef THREAD_ALIGN
#define THREAD_ALIGN	THREAD_SIZE
#endif

#define THREADINFO_GFP		(GFP_KERNEL_ACCOUNT | __GFP_ZERO)

/*
 * flag set/clear/test wrappers
 * - pass TIF_xxxx constants to these functions
 */

static inline void set_ti_thread_flag(struct thread_info *ti, int flag)
{
	set_bit(flag, (unsigned long *)&ti->flags);
}

static inline void clear_ti_thread_flag(struct thread_info *ti, int flag)
{
	clear_bit(flag, (unsigned long *)&ti->flags);
}

static inline void update_ti_thread_flag(struct thread_info *ti, int flag,
					 bool value)
{
	if (value)
		set_ti_thread_flag(ti, flag);
	else
		clear_ti_thread_flag(ti, flag);
}

static inline int test_and_set_ti_thread_flag(struct thread_info *ti, int flag)
{
	return test_and_set_bit(flag, (unsigned long *)&ti->flags);
}

static inline int test_and_clear_ti_thread_flag(struct thread_info *ti, int flag)
{
	return test_and_clear_bit(flag, (unsigned long *)&ti->flags);
}

static inline int test_ti_thread_flag(struct thread_info *ti, int flag)
{
	return test_bit(flag, (unsigned long *)&ti->flags);
}

#define set_thread_flag(flag) \
	set_ti_thread_flag(current_thread_info(), flag)
#define clear_thread_flag(flag) \
	clear_ti_thread_flag(current_thread_info(), flag)
#define update_thread_flag(flag, value) \
	update_ti_thread_flag(current_thread_info(), flag, value)
#define test_and_set_thread_flag(flag) \
	test_and_set_ti_thread_flag(current_thread_info(), flag)
#define test_and_clear_thread_flag(flag) \
	test_and_clear_ti_thread_flag(current_thread_info(), flag)
#define test_thread_flag(flag) \
	test_ti_thread_flag(current_thread_info(), flag)

#define tif_need_resched() test_thread_flag(TIF_NEED_RESCHED)

#ifndef CONFIG_HAVE_ARCH_WITHIN_STACK_FRAMES
static inline int arch_within_stack_frames(unsigned long stack,
					   unsigned long stackend,
					   unsigned long obj, unsigned long len)
{
	return GOOD_STACK;
}
#endif

#ifdef CONFIG_HARDENED_USERCOPY
long __must_check __check_object_size(const void *ptr, unsigned long n, bool to_user);

static __always_inline long __must_check check_object_size(const void *ptr, unsigned long n, bool to_user)
{
	return __builtin_constant_p(n) ? 0 : __check_object_size(ptr, n, to_user);
}
#else
static __always_inline long __must_check check_object_size(const void *ptr, unsigned long n, bool to_user)
{
	return 0;
}
#endif /* CONFIG_HARDENED_USERCOPY */

#define __bad_copy_from(sz, bytes) \
extern void __compiletime_error("copy source size is too small") \
__bad_copy_from_##sz##_##bytes(void);

#define __bad_copy_to(sz, bytes) \
extern void __compiletime_error("copy destination size is too small") \
__bad_copy_to##sz##_##bytes(void);

static inline void copy_overflow(const void *addr, size_t size, size_t count, bool is_source)
{
	WARN(1, "Buffer overflow detected %s %pS (%zu < %zu)!\n", is_source ? "from" : "to", addr, size, count);
}

static __always_inline __must_check bool
check_copy_size(const void *addr, size_t bytes, bool is_source)
{
	size_t sz = __compiletime_object_size(addr);

	if (check_object_size(addr, bytes, is_source))
		return false;

	if (unlikely(bytes > INT_MAX || bytes > sz)) {
#ifdef CONFIG_PAX_SIZE_OVERFLOW
		copy_overflow(addr, sz, bytes, is_source);
#else
		if (!__builtin_constant_p(bytes))
			copy_overflow(addr, sz, bytes, is_source);
		else if (is_source) {
			__bad_copy_to(sz, bytes);
		 } else {
			__bad_copy_from(sz, bytes);
		}
#endif
		return false;
	}
	return true;
}

#ifndef arch_setup_new_exec
static inline void arch_setup_new_exec(void) { }
#endif

#endif	/* __KERNEL__ */

#endif /* _LINUX_THREAD_INFO_H */
