/*
 * Copyright 2011 by Emese Revfy <re.emese@gmail.com>
 * Copyright 2011-2021 by PaX Team <pageexec@freemail.hu>
 * Licensed under the GPL v2
 */

#include "constify.h"

static bool check_attr_usage(tree type, tree name, const char *incompat_attr1, const char *incompat_attr2)
{
	if (!TYPE_P(type)) {
		error("%qE attribute applies to types only (%qD)", name, type);
		return false;
	}

	if (TREE_CODE(type) != RECORD_TYPE && TREE_CODE(type) != UNION_TYPE && TREE_CODE(type) != ARRAY_TYPE) {
		error("%qE attribute used on %qT applies to struct and union types only", name, type);
		return false;
	}

	if (lookup_attribute(IDENTIFIER_POINTER(name), TYPE_ATTRIBUTES(type))) {
		error("%qE attribute used on %qT is already applied to the type", name, type);
		return false;
	}

	if (lookup_attribute(incompat_attr1, TYPE_ATTRIBUTES(type))) {
		error("%qE attribute used on %qT is incompatible with '%s'", name, type, incompat_attr1);
		return false;
	}

	if (incompat_attr2 && lookup_attribute(incompat_attr2, TYPE_ATTRIBUTES(type))) {
		error("%qE attribute used on %qT is incompatible with '%s'", name, type, incompat_attr2);
		return false;
	}

	return true;
}

/*
 * the no_const attribute on a type ensures that the type won't be automatically constified.
 */
static tree handle_no_const_attribute(tree *node, tree name, tree args __unused, int flags __unused, bool *no_add_attrs)
{
	tree *type;

	*no_add_attrs = true;
	if (TREE_CODE(*node) == FUNCTION_DECL) {
		error_at(DECL_SOURCE_LOCATION(*node), "%qE attribute does not apply to functions (%qF)", name, *node);
		return NULL_TREE;
	}

	if (TREE_CODE(*node) == PARM_DECL) {
		error_at(DECL_SOURCE_LOCATION(*node), "%qE attribute does not apply to function parameters (%qD)", name, *node);
		return NULL_TREE;
	}

	if (VAR_P(*node)) {
		error_at(DECL_SOURCE_LOCATION(*node), "%qE attribute does not apply to variables (%qD)", name, *node);
		return NULL_TREE;
	}

	if (TYPE_P(*node)) {
		type = node;
		gcc_assert(flags & (int)ATTR_FLAG_TYPE_IN_PLACE);
	} else {
		if (TREE_CODE(*node) != TYPE_DECL) {
			error_at(DECL_SOURCE_LOCATION(*node), "%qE attribute does not apply to %qD (%qT)", name, *node, TREE_TYPE(*node));
			return NULL_TREE;
		}
		gcc_assert(!(flags & (int)ATTR_FLAG_TYPE_IN_PLACE));
		type = &TREE_TYPE(*node);
	}

	if (TYPE_P(*node)) {
		if (check_attr_usage(*type, name, "do_const", "mutable_const"))
			*no_add_attrs = false;
		return NULL_TREE;
	}

	if (!check_attr_usage(*type, name, "do_const", NULL))
		return NULL_TREE;

	if (!TYPE_P(*node)) {
		*type = build_variant_type_copy(*type);
		TYPE_ATTRIBUTES(*type) = tree_cons(name, NULL_TREE, TYPE_ATTRIBUTES(*type));
	}

	if (lookup_attribute("mutable_const", TYPE_ATTRIBUTES(*type))) {
		*no_add_attrs = false;
		return NULL_TREE;
	}

	if (enabled && TYPE_FIELDS(*type))
		error("1 %qE attribute used on type %qT that is not constified", name, *type);

	return NULL_TREE;
}

/*
 * the do_const attribute on a type ensures that the type will be forcibly constified (as if
 * the type declaration had been replaced with a const typedef).
 */
static tree handle_do_const_attribute(tree *node, tree name, tree args __unused, int flags __unused, bool *no_add_attrs)
{
	*no_add_attrs = true;

	if (!check_attr_usage(*node, name, "no_const", "mutable_const"))
		return NULL_TREE;

	*no_add_attrs = false;
	return NULL_TREE;
}

/*
 * the mutable_const attribute on a type ensures that all objects of the type with static storage
 * duration will be forcibly constified.
 */
static tree handle_mutable_const_attribute(tree *node, tree name, tree args __unused, int flags __unused, bool *no_add_attrs)
{
	*no_add_attrs = true;

	if (TYPE_P(*node)) {
		if (!check_attr_usage(*node, name, "no_const", "do_const"))
			return NULL_TREE;
	} else {
		gcc_assert(DECL_P(*node));

		switch (TREE_CODE(*node)) {
		default:
			error_at(DECL_SOURCE_LOCATION(*node), "%qE attribute applies to variables and parameters only, not %qD", name, *node);
			return NULL_TREE;

		case TYPE_DECL: {
			tree *type, orig_type, decl;

			decl = *node;
			gcc_assert(DECL_ORIGINAL_TYPE(decl) == NULL_TREE);
			gcc_assert(!(flags & (int)ATTR_FLAG_TYPE_IN_PLACE));

			type = &TREE_TYPE(decl);
			orig_type = TREE_TYPE(decl);
			if (lookup_attribute(IDENTIFIER_POINTER(name), TYPE_ATTRIBUTES(orig_type)))
				return NULL_TREE;

			gcc_assert(!POINTER_TYPE_P(orig_type));
			*type = build_variant_type_copy(orig_type);

			TYPE_ATTRIBUTES(*type) = tree_cons(name, NULL_TREE, TYPE_ATTRIBUTES(orig_type));
			break;
		}

		case VAR_DECL:
			break;
		}
	}

	*no_add_attrs = false;
	return NULL_TREE;
}

struct attribute_spec no_const_attr = ATTRIBUTE_SPEC_INIT("no_const", 0, 0, false, false, false, handle_no_const_attribute, true, NULL);
struct attribute_spec do_const_attr = ATTRIBUTE_SPEC_INIT("do_const", 0, 0, false, true, false, handle_do_const_attribute, true, NULL);
struct attribute_spec mutable_const_attr = ATTRIBUTE_SPEC_INIT("mutable_const", 0, 0, false, false, false, handle_mutable_const_attribute, true, NULL);
