/*
 * Copyright 2014-2019 by Open Source Security, Inc., Brad Spengler <spender@grsecurity.net>
 *                   and PaX Team <pageexec@freemail.hu>
 * Licensed under the GPL v2
 *
 * Note: the choice of the license means that the compilation process is
 *       NOT 'eligible' as defined by gcc's library exception to the GPL v3,
 *       but for the kernel it doesn't matter since it doesn't link against
 *       any of the gcc libraries
 *
 * Usage:
 * $ # for 4.5/4.6/C based 4.7
 * $ gcc -I`gcc -print-file-name=plugin`/include -I`gcc -print-file-name=plugin`/include/c-family -fPIC -shared -O2 -o randomize_layout_plugin.so randomize_layout_plugin.c
 * $ # for C++ based 4.7/4.8+
 * $ g++ -I`g++ -print-file-name=plugin`/include -I`g++ -print-file-name=plugin`/include/c-family -fPIC -shared -O2 -o randomize_layout_plugin.so randomize_layout_plugin.c
 * $ gcc -fplugin=./randomize_layout_plugin.so test.c -O2
 */

#include "gcc-common.h"
#include "randomize_layout_seed.h"

#if BUILDING_GCC_MAJOR < 4 || (BUILDING_GCC_MAJOR == 4 && BUILDING_GCC_MINOR < 6) || \
   (BUILDING_GCC_MAJOR == 4 && BUILDING_GCC_MINOR == 6 && BUILDING_GCC_PATCHLEVEL < 4)
#error "The RANDSTRUCT plugin requires GCC 4.6.4 or newer."
#endif

#define ORIG_TYPE_NAME(node) \
	(TYPE_NAME(TYPE_MAIN_VARIANT(node)) != NULL_TREE ? \
	 ((const unsigned char *)TYPE_NAME_POINTER(TYPE_MAIN_VARIANT(node))) : \
	 (const unsigned char *)" anonymous ")

__visible int plugin_is_GPL_compatible;

static enum mode { full, performance } shuffle_mode = full;

static struct plugin_info randomize_layout_plugin_info = {
	.version	= "201904050928",
	.help		= "disable\t\tdo not activate plugin\n"
			  "performance-mode\tenable cacheline-aware layout randomization\n"
};

/* from old Linux dcache.h */
static inline unsigned long
partial_name_hash(unsigned long c, unsigned long prevhash)
{
	return (prevhash + (c << 4) + (c >> 4)) * 11;
}
static inline unsigned int
name_hash(const unsigned char *name)
{
	unsigned long hash = 0;
	unsigned int len = strlen((const char *)name);
	while (len--)
		hash = partial_name_hash(*name++, hash);
	return (unsigned int)hash;
}

static tree handle_randomize_layout_attr(tree *node, tree name, tree args, int flags, bool *no_add_attrs)
{
	tree type;
	location_t loc;

	*no_add_attrs = true;
	loc = DECL_P(*node) ? DECL_SOURCE_LOCATION(*node) : UNKNOWN_LOCATION;

	if (TREE_CODE(*node) == FUNCTION_DECL) {
		error_at(loc, "%qE attribute does not apply to functions (%qF)", name, *node);
		return NULL_TREE;
	}

	if (TREE_CODE(*node) == PARM_DECL) {
		error_at(loc, "%qE attribute does not apply to function parameters (%qD)", name, *node);
		return NULL_TREE;
	}

	if (VAR_P(*node)) {
		error_at(loc, "%qE attribute does not apply to variables (%qD)", name, *node);
		return NULL_TREE;
	}

	if (TYPE_P(*node)) {
		type = *node;
	} else if (TREE_CODE(*node) == FIELD_DECL) {
		*no_add_attrs = false;
		return NULL_TREE;
	} else {
		gcc_assert(TREE_CODE(*node) == TYPE_DECL);
		type = TREE_TYPE(*node);
	}

	if (TREE_CODE(type) != RECORD_TYPE) {
		error_at(loc, "%qE attribute used on %qT applies to struct types only", name, type);
		return NULL_TREE;
	}

	if (lookup_attribute(IDENTIFIER_POINTER(name), TYPE_ATTRIBUTES(type))) {
		error_at(loc, "%qE attribute is already applied to the type %qT", name, type);
		return NULL_TREE;
	}

	*no_add_attrs = false;

	return NULL_TREE;
}

/* set on complete types that we don't need to inspect further at all */
static tree handle_randomize_considered_attr(tree *node, tree name, tree args, int flags, bool *no_add_attrs)
{
	*no_add_attrs = false;
	return NULL_TREE;
}

/*
 * set on types that we've performed a shuffle on, to prevent re-shuffling
 * this does not preclude us from inspecting its fields for potential shuffles
 */
static tree handle_randomize_performed_attr(tree *node, tree name, tree args, int flags, bool *no_add_attrs)
{
	*no_add_attrs = false;
	return NULL_TREE;
}

/*
 * 64bit variant of Bob Jenkins' public domain PRNG
 * 256 bits of internal state
 */

typedef unsigned long long u64;

typedef struct ranctx { u64 a; u64 b; u64 c; u64 d; } ranctx;

#define rot(x,k) (((x)<<(k))|((x)>>(64-(k))))
static u64 ranval(ranctx *x) {
	u64 e = x->a - rot(x->b, 7);
	x->a = x->b ^ rot(x->c, 13);
	x->b = x->c + rot(x->d, 37);
	x->c = x->d + e;
	x->d = e + x->a;
	return x->d;
}

static void raninit(ranctx *x, u64 *seed) {
	int i;

	x->a = seed[0];
	x->b = seed[1];
	x->c = seed[2];
	x->d = seed[3];

	for (i=0; i < 30; ++i)
		(void)ranval(x);
}

static u64 shuffle_seed[4];

static inline void fisher_yates_durstenfeld_shuffle(void *base, size_t nmemb, size_t size, ranctx *prng_state, void (*swap)(void *, void *))
{
	for (nmemb--; nmemb > 0; nmemb--)
		swap((unsigned char *)base + size * nmemb, (unsigned char *)base + size * (ranval(prng_state) % (nmemb + 1)));
}

struct partition_group {
	unsigned long start;
	unsigned long length;
	bool has_bitfields;
};

static inline bool new_partition_group_p(enum mode mode, bool prev_bit_field, bool curr_bit_field, unsigned long accum_size, unsigned long prev_size, unsigned long curr_size, tree field)
{
	/*
	 * 0-sized bitfields are special as they force an alignment of subsequent
	 * bitfields.  Creating a new partition is therefore necessary to ensure
	 * we shuffle the following bitfields on their own.
	 */
	if (prev_bit_field && prev_size == 0)
		return true;

	if (prev_bit_field != curr_bit_field)
		return true;

	/* Even in full mode we want to group bitfields together */
	if (prev_bit_field && curr_bit_field)
		return false;

	if (mode == full)
		return true;

	if (accum_size >= 64 * 8)
		return true;

	/* don't force a new partition on observing a struct of size 0 and a non-zero alignment */
	if (curr_size && prev_size != curr_size && DECL_ALIGN(field) > tree_to_uhwi(DECL_SIZE(field)))
		return true;

	return false;
}

static unsigned long partition_struct(tree *fields, unsigned long start, unsigned long length, struct partition_group *size_groups, enum mode mode)
{
	unsigned long i, prev_size, accum_size = 0, group_idx = 0;
	bool prev_bit_field;

	gcc_assert(length < INT_MAX);

	/*
	 * partition_struct() can be called on shuffled structures, so
	 * we can't rely on the offset of some field in the struct until
	 * the structure is fully laid out again
	 */
	for (i = 0 ; i < start; i++)
		accum_size += tree_to_uhwi(DECL_SIZE(fields[i]));

	prev_bit_field = DECL_BIT_FIELD_TYPE(fields[start]) != NULL_TREE;
	prev_size = tree_to_uhwi(DECL_SIZE(fields[start]));
	accum_size %= 64 * 8;

	for (i = start; i < start + length; i++) {
		bool curr_bit_field = DECL_BIT_FIELD_TYPE(fields[i]) != NULL_TREE;
		unsigned long curr_size = tree_to_uhwi(DECL_SIZE(fields[i]));

		if (i > start && new_partition_group_p(mode, prev_bit_field, curr_bit_field, accum_size, prev_size, curr_size, fields[i])) {
			group_idx++;
			accum_size %= 64 * 8;
		}

		if (size_groups[group_idx].length == 0)
			size_groups[group_idx].start = i;
		size_groups[group_idx].has_bitfields |= curr_bit_field;
		size_groups[group_idx].length++;

		prev_bit_field = curr_bit_field;
		prev_size = curr_size;
		accum_size += curr_size;
	}

	return group_idx + 1;
}

#define swap_type(name, type) \
static void swap_##name(void *_a, void *_b) \
{ \
	type *a = (type *)_a; \
	type *b = (type *)_b; \
	type tmp; \
\
	tmp = *a; \
	*a = *b; \
	*b = tmp; \
}

swap_type(partition_group, struct partition_group);
swap_type(tree, tree);

#undef swap_type

static unsigned long partition_struct_and_shuffle_groups(tree *newtree, unsigned long start, unsigned long length, struct partition_group *size_group, ranctx *prng_state, enum mode mode)
{
	tree tree_copy[length];
	unsigned long x, y, num_groups;

	memset(tree_copy, 0, sizeof(tree) * length);

	num_groups = partition_struct(newtree, start, length, size_group, mode);
	fisher_yates_durstenfeld_shuffle(size_group, num_groups, sizeof(size_group[0]), prng_state, swap_partition_group);

	for (x = 0, y = 0; x < num_groups; x++) {
		memcpy(tree_copy + y, newtree + size_group[x].start, sizeof(newtree[0]) * size_group[x].length);
		/*
		 * performance mode will shuffle bitfields as a group below.  for full mode, we
		 * will need to do it here
		 * the only way in full mode to have a group of length > 1 is to have a run of
		 * bitfields
		 */
		if (mode == full && size_group[x].length > 1) {
			tree last_field;
			unsigned long shuffle_length;

			last_field = tree_copy[y + size_group[x].length - 1];
			gcc_assert(DECL_BIT_FIELD_TYPE(last_field) != NULL_TREE);

			shuffle_length = size_group[x].length;

			/*
			 * don't randomize the placement of the 0-width bitfield used to force
			 * alignment of subsequent bitfields
			 */
			if (tree_to_uhwi(DECL_SIZE(last_field)) == 0)
				shuffle_length--;

			fisher_yates_durstenfeld_shuffle(tree_copy + y, shuffle_length, sizeof(tree_copy[0]), prng_state, swap_tree);
		}
		size_group[x].start = start + y;
		y += size_group[x].length;
	}
	memcpy(newtree + start, tree_copy, sizeof(newtree[0]) * length);

	return num_groups;
}

/* modern in-place Fisher-Yates-Durstenfeld shuffle */
static void shuffle(const_tree type, tree *newtree, unsigned long start, unsigned long length)
{
	unsigned long i, x, num_groups;
	u64 seed[4];
	ranctx prng_state;
	const unsigned char *structname;
	struct partition_group size_group[length];
	enum mode mode = shuffle_mode;

	if (length == 0)
		return;

	gcc_assert(TREE_CODE(type) == RECORD_TYPE);

	structname = ORIG_TYPE_NAME(type);

#ifdef __DEBUG_PLUGIN
	fprintf(stderr, "Shuffling struct %s %p\n", (const char *)structname, type);
#ifdef __DEBUG_VERBOSE
	debug_tree((tree)type);
#endif
#endif

	for (i = 0; i < 4; i++) {
		seed[i] = shuffle_seed[i];
		seed[i] ^= name_hash(structname);
	}

	raninit(&prng_state, (u64 *)&seed);

	memset(size_group, 0, sizeof(struct partition_group) * length);

	num_groups = partition_struct_and_shuffle_groups(newtree, start, length, size_group, &prng_state, mode);

	if (mode == performance && !lookup_attribute("packed", TYPE_ATTRIBUTES(type))) {
		for (x = 0; x < num_groups; x++) {
			/*
			 * in performance mode, runs of bitfields can exist within a group of other
			 * fields.  in order to not harm the randomization, the easiest thing to do is
			 * partition the groups with bitfields a second time as if it was in full mode
			 * and shuffle them via the groups created
			 */
			if (size_group[x].has_bitfields) {
				struct partition_group sub_size_group[size_group[x].length];
				tree last_field;
				unsigned long shuffle_length;

				last_field = newtree[size_group[x].start + size_group[x].length - 1];
				gcc_assert(DECL_BIT_FIELD_TYPE(last_field) != NULL_TREE);
				memset(sub_size_group, 0, sizeof(struct partition_group) * size_group[x].length);

				shuffle_length = size_group[x].length;

				/*
				 * don't randomize the placement of the 0-width bitfield used to force
				 * alignment of subsequent bitfields
				 */
				if (tree_to_uhwi(DECL_SIZE(last_field)) == 0)
					shuffle_length--;

				partition_struct_and_shuffle_groups(newtree + size_group[x].start, 0, shuffle_length, sub_size_group, &prng_state, full);
			} else
				fisher_yates_durstenfeld_shuffle(newtree + size_group[x].start, size_group[x].length, sizeof(newtree[0]), &prng_state, swap_tree);
		}
	}
}

static bool is_flexible_array(const_tree field)
{
	const_tree fieldtype;
	const_tree typesize;
	const_tree elemtype;
	const_tree elemsize;

	fieldtype = TREE_TYPE(field);
	typesize = TYPE_SIZE(fieldtype);

	if (TREE_CODE(fieldtype) != ARRAY_TYPE)
		return false;

	elemtype = TREE_TYPE(fieldtype);
	elemsize = TYPE_SIZE(elemtype);

	/* size of type is represented in bits */

	if (typesize == NULL_TREE && TYPE_DOMAIN(fieldtype) != NULL_TREE &&
	    TYPE_MAX_VALUE(TYPE_DOMAIN(fieldtype)) == NULL_TREE)
		return true;

	if (typesize != NULL_TREE &&
	    (TREE_CONSTANT(typesize) && (!tree_to_uhwi(typesize) ||
	     tree_to_uhwi(typesize) == tree_to_uhwi(elemsize))))
		return true;

	return false;
}

static int relayout_struct(tree type)
{
	unsigned long num_fields = list_length(TYPE_FIELDS(type));
	unsigned long shuffle_start = 0, shuffle_length = num_fields;
	tree field;
	tree newtree[num_fields];
	unsigned long i;
	tree list;
	tree variant;
	tree main_variant;
	expanded_location xloc;
	bool has_flexarray = false;

	gcc_assert(TREE_CODE(type) == RECORD_TYPE);

	gcc_assert(num_fields < INT_MAX);

	if (lookup_attribute("randomize_performed", TYPE_ATTRIBUTES(type)) ||
	    lookup_attribute("no_randomize_layout", TYPE_ATTRIBUTES(TYPE_MAIN_VARIANT(type))))
		return 0;

	/* Workaround for 3rd-party VirtualBox source that we can't modify ourselves */
	if (!strcmp((const char *)ORIG_TYPE_NAME(type), "INTNETTRUNKFACTORY") ||
	    !strcmp((const char *)ORIG_TYPE_NAME(type), "RAWPCIFACTORY"))
		return 0;

	/* throw out any structs in uapi */
	xloc = expand_location(DECL_SOURCE_LOCATION(TYPE_FIELDS(type)));

	if (strstr(xloc.file, "/uapi/")) {
		error_at(DECL_SOURCE_LOCATION(TYPE_FIELDS(type)), G_("attempted to randomize userland API struct %s"), ORIG_TYPE_NAME(type));
		return 0;
	}

	for (field = TYPE_FIELDS(type), i = 0; field; field = TREE_CHAIN(field), i++) {
		gcc_assert(TREE_CODE(field) == FIELD_DECL);

		newtree[i] = field;

		/*
		 * enforce that we don't randomize the layout of the last
		 * element of a struct if it's a 0 or 1-length array
		 * or a proper flexible array
		 */
		if (i == num_fields - 1 && is_flexible_array(field)) {
			has_flexarray = true;
			shuffle_length--;
		} else if (lookup_attribute("no_randomize_layout", DECL_ATTRIBUTES(field))) {
			if (i != shuffle_start) {
				error_at(DECL_SOURCE_LOCATION(field), "non-randomized field %qD of type %qT must not be preceeded by randomized fields", field, type);
				return 0;
			}

			if (i == num_fields - 2) {
				error_at(DECL_SOURCE_LOCATION(field), "no fields can be randomized in type %qT", type);
				return 0;
			}

			shuffle_start++;
			shuffle_length--;
		}
	}

	shuffle(type, newtree, shuffle_start, shuffle_length);

	/*
	 * set up a bogus anonymous struct field designed to error out on unnamed struct initializers
	 * as gcc provides no other way to detect such code
	 */
	list = make_node(FIELD_DECL);
	TREE_CHAIN(list) = newtree[0];
	TREE_TYPE(list) = void_type_node;
	DECL_SIZE(list) = bitsize_zero_node;
	DECL_NONADDRESSABLE_P(list) = 1;
	DECL_FIELD_BIT_OFFSET(list) = bitsize_zero_node;
	DECL_SIZE_UNIT(list) = size_zero_node;
	DECL_FIELD_OFFSET(list) = size_zero_node;
	DECL_CONTEXT(list) = type;

	for (i = 0; i < num_fields - 1; i++)
		TREE_CHAIN(newtree[i]) = newtree[i+1];
	TREE_CHAIN(newtree[num_fields - 1]) = NULL_TREE;

	add_type_attr(type, "randomize_performed", NULL_TREE);
	if (has_flexarray)
		add_type_attr(type, "has_flexarray", NULL_TREE);

	main_variant = TYPE_MAIN_VARIANT(type);
	for (variant = main_variant; variant; variant = TYPE_NEXT_VARIANT(variant))
		TYPE_FIELDS(variant) = list;

	/*
	 * force a re-layout of the main variant
	 * the TYPE_SIZE for all variants will be recomputed
	 * by finalize_type_size()
	 */
	TYPE_SIZE(main_variant) = NULL_TREE;
	layout_type(main_variant);
	gcc_assert(TYPE_SIZE(main_variant) != NULL_TREE);

	return 1;
}

/* from constify plugin */
static const_tree get_field_type(const_tree field)
{
	return strip_array_types(TREE_TYPE(field));
}

/* derived from constify plugin */
static bool is_pure_ops_struct(const_tree type)
{
	const_tree field;

	gcc_assert(TREE_CODE(type) == RECORD_TYPE || TREE_CODE(type) == UNION_TYPE);

	for (field = TYPE_FIELDS(type); field; field = TREE_CHAIN(field)) {
		const_tree fieldtype = get_field_type(field);
		enum tree_code code = TREE_CODE(fieldtype);

		if (type == fieldtype)
			continue;

		if (FUNCTION_POINTER_TYPE_P(fieldtype))
			continue;

		if (code != RECORD_TYPE && code != UNION_TYPE)
			return false;

		if (!is_pure_ops_struct(fieldtype))
			return false;
	}

	return true;
}

static bool is_type_randomizable(tree type)
{
	unsigned long num_fields;

	if (TYPE_FIELDS(type) == NULL_TREE)
		return false;

	num_fields = (unsigned long)list_length(TYPE_FIELDS(type));
	if (num_fields < 2)
		return false;

	if (lookup_attribute("randomize_layout", TYPE_ATTRIBUTES(TYPE_MAIN_VARIANT(type))))
		return true;

	if (is_pure_ops_struct(type))
		return true;

	return false;
}

static void randomize_type(tree type)
{
	gcc_assert(TREE_CODE(type) == RECORD_TYPE);
	gcc_assert(!lookup_attribute("randomize_considered", TYPE_ATTRIBUTES(type)));

	if (is_type_randomizable(type))
		relayout_struct(type);

	add_type_attr(type, "randomize_considered", NULL_TREE);
#ifdef __DEBUG_PLUGIN
	fprintf(stderr, "Marking randomize_considered on struct %s\n", ORIG_TYPE_NAME(type));
#ifdef __DEBUG_VERBOSE
	debug_tree(type);
#endif
#endif
}

static void update_decl_size(tree decl)
{
	tree lastval, lastidx, field, init, type, flexsize;
	unsigned HOST_WIDE_INT len;

	type = TREE_TYPE(decl);

	if (!lookup_attribute("has_flexarray", TYPE_ATTRIBUTES(type)))
		return;

	init = DECL_INITIAL(decl);
	if (init == NULL_TREE || init == error_mark_node)
		return;

	if (TREE_CODE(init) != CONSTRUCTOR)
		return;

	if (!CONSTRUCTOR_NELTS(init))
		return;

	lastval = CONSTRUCTOR_ELT(init, CONSTRUCTOR_NELTS(init) - 1)->value;
	lastidx = CONSTRUCTOR_ELT(init, CONSTRUCTOR_NELTS(init) - 1)->index;

	for (field = TYPE_FIELDS(TREE_TYPE(decl)); TREE_CHAIN(field); field = TREE_CHAIN(field))
		;

	if (lastidx != field)
		return;

	switch (TREE_CODE(lastval)) {
	case STRING_CST:
		len = TREE_STRING_LENGTH(lastval);
		break;
	case CONSTRUCTOR:
		len = tree_to_uhwi(CONSTRUCTOR_ELT(lastval, CONSTRUCTOR_NELTS(lastval) - 1)->index) + 1;
		break;
	default:
		error_at(DECL_SOURCE_LOCATION(decl), "Only string constants and array initializers are "
			 "supported as initializers for randomized structures with flexible arrays");
		return;
	}

	flexsize = bitsize_int(len * tree_to_uhwi(TYPE_SIZE(TREE_TYPE(TREE_TYPE(lastval)))));
	DECL_SIZE(decl) = size_binop(PLUS_EXPR, TYPE_SIZE(type), flexsize);
}


static void randomize_layout_finish_decl(void *event_data, void *data)
{
	tree decl = (tree)event_data;
	tree type;

	if (decl == NULL_TREE || decl == error_mark_node)
		return;

	type = TREE_TYPE(decl);

	if (!VAR_P(decl))
		return;

	if (!RECORD_OR_UNION_TYPE_P(type))
		return;

	if (!lookup_attribute("randomize_performed", TYPE_ATTRIBUTES(type)))
		return;

	DECL_SIZE(decl) = 0;
	DECL_SIZE_UNIT(decl) = 0;
	SET_DECL_ALIGN(decl, 0);
	SET_DECL_MODE(decl, VOIDmode);
	SET_DECL_RTL(decl, 0);
	update_decl_size(decl);
	layout_decl(decl, 0);
}

static void finish_type(void *event_data, void *data)
{
	tree type = (tree)event_data;

	if (type == NULL_TREE || type == error_mark_node)
		return;

	if (TREE_CODE(type) != RECORD_TYPE)
		return;

	if (TYPE_FIELDS(type) == NULL_TREE)
		return;

	if (lookup_attribute("randomize_considered", TYPE_ATTRIBUTES(type)))
		return;

#ifdef __DEBUG_PLUGIN
	fprintf(stderr, "Calling randomize_type on %s\n", ORIG_TYPE_NAME(type));
#endif
#ifdef __DEBUG_VERBOSE
	debug_tree(type);
#endif
	randomize_type(type);
}

static struct attribute_spec randomize_layout_attr = ATTRIBUTE_SPEC_INIT("randomize_layout", 0, 0, false, true, false, handle_randomize_layout_attr, true, NULL);
static struct attribute_spec no_randomize_layout_attr = ATTRIBUTE_SPEC_INIT("no_randomize_layout", 0, 0, false, false, false, handle_randomize_layout_attr, true, NULL);
static struct attribute_spec randomize_considered_attr = ATTRIBUTE_SPEC_INIT("randomize_considered", 0, 0, false, true, false, handle_randomize_considered_attr, false, NULL);
static struct attribute_spec randomize_performed_attr = ATTRIBUTE_SPEC_INIT("randomize_performed", 0, 0, false, true, false, handle_randomize_performed_attr, false, NULL);

static void register_attributes(void *event_data, void *data)
{
	register_attribute(&randomize_layout_attr);
	register_attribute(&no_randomize_layout_attr);
	register_attribute(&randomize_considered_attr);
	register_attribute(&randomize_performed_attr);
}

static void check_bad_casts_in_constructor(tree var, tree init)
{
	unsigned HOST_WIDE_INT idx;
	tree field, val;
	tree field_type, val_type;

	FOR_EACH_CONSTRUCTOR_ELT(CONSTRUCTOR_ELTS(init), idx, field, val) {
		if (TREE_CODE(val) == CONSTRUCTOR) {
			check_bad_casts_in_constructor(var, val);
			continue;
		}

		/* pipacs' plugin creates franken-arrays that differ from those produced by
		   normal code which all have valid 'field' trees. work around this */
		if (field == NULL_TREE)
			continue;
		field_type = TREE_TYPE(field);
		val_type = TREE_TYPE(val);

		if (TREE_CODE(field_type) != POINTER_TYPE || TREE_CODE(val_type) != POINTER_TYPE)
			continue;

		if (field_type == val_type)
			continue;

		field_type = TYPE_MAIN_VARIANT(strip_array_types(TYPE_MAIN_VARIANT(TREE_TYPE(field_type))));
		val_type = TYPE_MAIN_VARIANT(strip_array_types(TYPE_MAIN_VARIANT(TREE_TYPE(val_type))));

		if (field_type == void_type_node)
			continue;
		if (field_type == val_type)
			continue;
		if (TREE_CODE(val_type) != RECORD_TYPE)
			continue;

		if (!lookup_attribute("randomize_performed", TYPE_ATTRIBUTES(val_type)))
			continue;
		inform(DECL_SOURCE_LOCATION(var), "found mismatched struct pointer types: %qT and %qT", TYPE_MAIN_VARIANT(field_type), TYPE_MAIN_VARIANT(val_type));
	}
}

/* derived from the constify plugin */
static void check_global_variables(void *event_data, void *data)
{
	varpool_node_ptr node;
	tree init;

	FOR_EACH_VARIABLE(node) {
		tree var = NODE_DECL(node);
		init = DECL_INITIAL(var);
		if (init == NULL_TREE)
			continue;

		if (TREE_CODE(init) != CONSTRUCTOR)
			continue;

		check_bad_casts_in_constructor(var, init);
	}
}

static bool dominated_by_is_err(const_tree rhs, basic_block bb)
{
	basic_block dom;
	gimple_stmt_ptr dom_stmt;
	gimple_stmt_ptr call_stmt;
	const_tree dom_lhs;
	const_tree poss_is_err_cond;
	const_tree poss_is_err_func;
	const_tree is_err_arg;

	dom = get_immediate_dominator(CDI_DOMINATORS, bb);
	if (!dom)
		return false;

	dom_stmt = last_stmt(dom);
	if (!dom_stmt)
		return false;

	if (gimple_code(dom_stmt) != GIMPLE_COND)
		return false;

	if (gimple_cond_code(dom_stmt) != NE_EXPR)
		return false;

	if (!integer_zerop(gimple_cond_rhs(dom_stmt)))
		return false;

	poss_is_err_cond = gimple_cond_lhs(dom_stmt);

	if (TREE_CODE(poss_is_err_cond) != SSA_NAME)
		return false;

	call_stmt = SSA_NAME_DEF_STMT(poss_is_err_cond);

	if (!is_gimple_call(call_stmt))
		return false;

	dom_lhs = gimple_get_lhs(call_stmt);
	poss_is_err_func = gimple_call_fndecl(call_stmt);
	if (!poss_is_err_func)
		return false;
	if (dom_lhs != poss_is_err_cond)
		return false;
	if (!id_equal(DECL_NAME(poss_is_err_func), "IS_ERR"))
		return false;

	is_err_arg = gimple_call_arg(call_stmt, 0);
	if (!is_err_arg)
		return false;

	if (is_err_arg != rhs)
		return false;

	return true;
}

static void handle_local_var_initializers(void)
{
	tree var;
	unsigned int i;

	FOR_EACH_LOCAL_DECL(cfun, i, var) {
		tree init = DECL_INITIAL(var);
		if (!init)
			continue;
		if (TREE_CODE(init) != CONSTRUCTOR)
			continue;
		check_bad_casts_in_constructor(var, init);
	}
}

/*
 * iterate over all statements to find "bad" casts:
 * those where the address of the start of a structure is cast
 * to a pointer of a structure of a different type, or a
 * structure pointer type is cast to a different structure pointer type
 */
static unsigned int find_bad_casts_execute(void)
{
	basic_block bb;

	handle_local_var_initializers();

	FOR_EACH_BB_FN(bb, cfun) {
		gimple_stmt_iterator gsi;

		for (gsi = gsi_start_bb(bb); !gsi_end_p(gsi); gsi_next(&gsi)) {
			gimple_stmt_ptr stmt;
			const_tree lhs;
			const_tree lhs_type;
			const_tree rhs1;
			const_tree rhs_type;
			const_tree ptr_lhs_type;
			const_tree ptr_rhs_type;
			const_tree op0;
			const_tree op0_type;
			enum tree_code rhs_code;

			stmt = gsi_stmt(gsi);

#ifdef __DEBUG_PLUGIN
#ifdef __DEBUG_VERBOSE
			debug_gimple_stmt(stmt);
			debug_tree(gimple_get_lhs(stmt));
#endif
#endif

			if (!is_gimple_assign(stmt))
				continue;

#ifdef __DEBUG_PLUGIN
#ifdef __DEBUG_VERBOSE
			debug_tree(gimple_assign_rhs1(stmt));
#endif
#endif

			rhs_code = gimple_assign_rhs_code(stmt);

			if (rhs_code != ADDR_EXPR && rhs_code != SSA_NAME)
				continue;

			lhs = gimple_get_lhs(stmt);
			lhs_type = TREE_TYPE(lhs);
			rhs1 = gimple_assign_rhs1(stmt);
			rhs_type = TREE_TYPE(rhs1);

			if (TREE_CODE(rhs_type) != POINTER_TYPE ||
			    TREE_CODE(lhs_type) != POINTER_TYPE)
				continue;

			ptr_lhs_type = TYPE_MAIN_VARIANT(strip_array_types(TYPE_MAIN_VARIANT(TREE_TYPE(lhs_type))));
			ptr_rhs_type = TYPE_MAIN_VARIANT(strip_array_types(TYPE_MAIN_VARIANT(TREE_TYPE(rhs_type))));

			if (ptr_rhs_type == void_type_node)
				continue;

			if (ptr_lhs_type == void_type_node)
				continue;

			if (dominated_by_is_err(rhs1, bb))
				continue;

			if (TREE_CODE(ptr_rhs_type) != RECORD_TYPE) {
#ifndef __DEBUG_PLUGIN
				if (lookup_attribute("randomize_performed", TYPE_ATTRIBUTES(ptr_lhs_type)))
#endif
				inform(gimple_location(stmt), "found mismatched struct pointer types: %qT and %qT", ptr_lhs_type, ptr_rhs_type);
				continue;
			}

			if (rhs_code == SSA_NAME && ptr_lhs_type == ptr_rhs_type)
				continue;

			if (rhs_code == ADDR_EXPR) {
				op0 = TREE_OPERAND(rhs1, 0);

				if (op0 == NULL_TREE)
					continue;

				if (!VAR_P(op0))
					continue;

				op0_type = TYPE_MAIN_VARIANT(strip_array_types(TYPE_MAIN_VARIANT(TREE_TYPE(op0))));
				if (op0_type == ptr_lhs_type)
					continue;

#ifndef __DEBUG_PLUGIN
				if (lookup_attribute("randomize_performed", TYPE_ATTRIBUTES(op0_type)))
#endif
				inform(gimple_location(stmt), "found mismatched struct pointer types: %qT and %qT", ptr_lhs_type, op0_type);
			} else {
				const_tree ssa_name_var = SSA_NAME_VAR(rhs1);
				/* skip bogus type casts introduced by container_of */
				if (ssa_name_var != NULL_TREE && DECL_NAME(ssa_name_var) &&
				    id_equal(DECL_NAME(ssa_name_var), "__mptr"))
					continue;
#ifndef __DEBUG_PLUGIN
				if (lookup_attribute("randomize_performed", TYPE_ATTRIBUTES(ptr_rhs_type)))
#endif
				inform(gimple_location(stmt), "found mismatched struct pointer types: %qT and %qT", ptr_lhs_type, ptr_rhs_type);
			}

		}
	}
	return 0;
}

#define PASS_NAME find_bad_casts
#define NO_GATE
#define TODO_FLAGS_FINISH TODO_dump_func
#include "gcc-generate-gimple-pass.h"

#if BUILDING_GCC_VERSION >= 6000
/*
 * the following functions ensure that debug information is emitted only for
 * the randomized version of the randomized types thus enabling gdb and other
 * tools to understand variables of these types.
 *
 * the core problem is that gcc can apparently emit debug information for an
 * extern variable (including its type) even before the underlying type has
 * been constructed and randomized and later emissions of the randomized type
 * do not override this first emission. therefore we want to make sure that
 * if a type is randomizable at all, only its randomized variant is emitted.
 *
 * the trick is that we hook into the debug information emission paths and
 * filter out the not-yet-randomized attempts. care has to be taken to filter
 * not only types but decls too since it is in fact the latter that (try to)
 * have a first go at emitting the not-yet-randomized type.
 *
 * the above also means that we lose debug info on extern variables but the
 * expectation is that we get it back from TUs that define them.
 */
static bool randomize_skip_type(tree type)
{
	if (TREE_CODE(type) != RECORD_TYPE && TREE_CODE(type) != UNION_TYPE)
		return false;

	if (!is_type_randomizable(type))
		return false;

	if (lookup_attribute("randomize_performed", TYPE_ATTRIBUTES(TYPE_MAIN_VARIANT(type))))
		return false;

	return true;
}

static const struct gcc_debug_hooks *old_debug_hooks;
static struct gcc_debug_hooks randomize_layout_debug_hooks;

static void randomize_layout_type_decl(tree decl, int local)
{
	tree type;

	gcc_assert(TREE_CODE(decl) == TYPE_DECL);
	type = TREE_TYPE(decl);

	if (randomize_skip_type(type))
		return;

//fprintf(stderr, "TYPE ");debug_tree(decl);
	if (old_debug_hooks && old_debug_hooks->type_decl)
		old_debug_hooks->type_decl(decl, local);
}

static void randomize_layout_early_global_decl(tree decl)
{
	tree type = TREE_TYPE(decl);

	if (randomize_skip_type(type))
		return;

//fprintf(stderr, "EARLYDECL ");debug_tree(decl);
	if (old_debug_hooks && old_debug_hooks->early_global_decl)
		old_debug_hooks->early_global_decl(decl);
}

static void randomize_layout_start_unit(void *gcc_data __unused, void *user_data __unused)
{
	if (debug_hooks)
		randomize_layout_debug_hooks = *debug_hooks;

	randomize_layout_debug_hooks.type_decl = randomize_layout_type_decl;
	randomize_layout_debug_hooks.early_global_decl = randomize_layout_early_global_decl;

	old_debug_hooks = debug_hooks;
	debug_hooks = &randomize_layout_debug_hooks;
}
#endif

__visible int plugin_init(struct plugin_name_args *plugin_info, struct plugin_gcc_version *version)
{
	int i;
	const char * const plugin_name = plugin_info->base_name;
	const int argc = plugin_info->argc;
	const struct plugin_argument * const argv = plugin_info->argv;
	bool enable = true;
	int obtained_seed = 0;

	PASS_INFO(find_bad_casts,			"ssa",		1, PASS_POS_INSERT_AFTER);

	if (!plugin_default_version_check(version, &gcc_version)) {
		error(G_("incompatible gcc/plugin versions"));
		return 1;
	}

	if (strncmp(lang_hooks.name, "GNU C", 5) || !strncmp(lang_hooks.name, "GNU C+", 6)) {
		inform(UNKNOWN_LOCATION, G_("%s supports C only, not %s"), plugin_name, lang_hooks.name);
		enable = false;
	}

	for (i = 0; i < argc; ++i) {
		if (!strcmp(argv[i].key, "disable")) {
			enable = false;
			continue;
		}
		if (!strcmp(argv[i].key, "performance-mode")) {
			shuffle_mode = performance;
			continue;
		}
		error(G_("unknown option '-fplugin-arg-%s-%s'"), plugin_name, argv[i].key);
	}

	if (strlen(randstruct_seed) != 64) {
		error(G_("invalid seed value supplied for %s plugin"), plugin_name);
		return 1;
	}
	obtained_seed = sscanf(randstruct_seed, "%016llx%016llx%016llx%016llx",
		&shuffle_seed[0], &shuffle_seed[1], &shuffle_seed[2], &shuffle_seed[3]);
	if (obtained_seed != 4) {
		error(G_("Invalid seed supplied for %s plugin"), plugin_name);
		return 1;
	}

	register_callback(plugin_name, PLUGIN_INFO, NULL, &randomize_layout_plugin_info);
	if (enable) {
#if BUILDING_GCC_VERSION >= 6000
		register_callback(plugin_name, PLUGIN_START_UNIT, randomize_layout_start_unit, NULL);
#endif
		register_callback(plugin_name, PLUGIN_ALL_IPA_PASSES_START, check_global_variables, NULL);
		register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &find_bad_casts_pass_info);
		register_callback(plugin_name, PLUGIN_FINISH_TYPE, finish_type, NULL);
		register_callback(plugin_name, PLUGIN_FINISH_DECL, randomize_layout_finish_decl, NULL);
	}
	register_callback(plugin_name, PLUGIN_ATTRIBUTES, register_attributes, NULL);

	return 0;
}
