/*
 * Copyright 2012-2020 by PaX Team <pageexec@freemail.hu>
 * Licensed under the GPL v2
 *
 * Homepage: http://pax.grsecurity.net/
 */

#include "rap.h"

#ifdef GCC_DF_H
static bool rap_xor_elimination_gate(void)
{
	rtx_insn *insn;

	if (!optimize)
		return false;

	// handle -fdump-passes
	if (!cfun->cfg)
		return true;

	if (rap_noreturn_function())
		return false;

#if BUILDING_GCC_VERSION >= 4009
	if (cfun->tail_call_marked)
		return false;
#endif

	if (rap_calls_complex_function_p(current_function_decl))
		return false;

	for (insn = get_insns(); insn; insn = NEXT_INSN(insn)) {
		rtx body;

		if (!CALL_P(insn) && !NONJUMP_INSN_P(insn))
			continue;

		body = PATTERN(insn);
		if (GET_CODE(body) == PARALLEL)
			body = XVECEXP(body, 0, 0);

		if (NONJUMP_INSN_P(insn)) {
			if (GET_CODE(body) == SET)
				body = SET_SRC(body);
			if (GET_CODE(body) != ASM_OPERANDS)
				continue;
			if (regexec(&asm_call_insn_regex,  ASM_OPERANDS_TEMPLATE(body), 0, NULL, 0))
				continue;
			return false;
		}

		if (GET_CODE(body) == SET)
			body = SET_SRC(body);

		if (GET_CODE(body) != CALL)
			return false;

		body = XEXP(body, 0);
		if (!MEM_P(body))
			return false;

		body = XEXP(body, 0);
		if (GET_CODE(body) != SYMBOL_REF)
			return false;
		gcc_assert(SYMBOL_REF_FUNCTION_P(body));

		if (rap_is_complex_function(SYMBOL_REF_DECL(body)))
			return false;
	}

	return true;
}

enum rap_xor { xor_none, xor_entry, xor_exit };

static enum rap_xor rap_match_xor(rtx body, bool both)
{
	rtx dst, src, src1, src2;
	tree expr;
	enum rap_xor ret;

	if (GET_CODE(body) != SET)
		return xor_none;

	dst = SET_DEST(body);
	if (!REG_P(dst) || !REG_EXPR(dst))
		return xor_none;
	expr = REG_EXPR(dst);
	if (!DECL_P(expr))
		return xor_none;
	if (!DECL_NAME(expr))
		return xor_none;
	if (!strncmp(DECL_NAME_POINTER(expr), "rap_retaddr_entry_", sizeof("rap_retaddr_entry_") - 1))
		ret = xor_entry;
	else if (!strncmp(DECL_NAME_POINTER(expr), "rap_retaddr_exit_", sizeof("rap_retaddr_exit_") - 1))
		ret = xor_exit;
	else if (!strncmp(DECL_NAME_POINTER(expr), "rap_cookie_entry", sizeof("rap_cookie_entry") - 1))
		ret = xor_entry;
	else if (!strncmp(DECL_NAME_POINTER(expr), "rap_cookie_exit", sizeof("rap_cookie_exit") - 1))
		ret = xor_exit;
	else
		return xor_none;

	src = SET_SRC(body);
	if (GET_CODE(src) != XOR)
		return xor_none;
	if (GET_MODE(dst) != GET_MODE(src))
		return xor_none;

	src1 = XEXP(src, 0);
	if (!REG_P(src1))
		return xor_none;
#if BUILDING_GCC_VERSION <= 4008
	if (!REG_EXPR(src1) || !DECL_NAME(REG_EXPR(src1)) || strncmp(DECL_NAME_POINTER(REG_EXPR(src1)), "rap_retaddr_", sizeof("rap_retaddr_") - 1))
		return xor_none;
#endif

	if (both) {
		src2 = XEXP(src, 1);

		if (!REG_P(src2) || !REG_EXPR(src2))
			return xor_none;
		if (!DECL_NAME(REG_EXPR(src2)))
			return xor_none;
		if (strncmp(DECL_NAME_POINTER(REG_EXPR(src2)), "rap_cookie", sizeof("rap_cookie") - 1))
			return xor_none;
	}

	return REGNO(dst) == REGNO(src1) ? ret : xor_none;
}

enum rap_mov { mov_none, mov_entry, mov_exit };

static enum rap_mov rap_match_mov(rtx body)
{
	rtx dst, src;
	tree expr;
	enum rap_mov ret;

	if (GET_CODE(body) != SET)
		return mov_none;

	dst = SET_DEST(body);
	if (!REG_P(dst) || !REG_EXPR(dst))
		return mov_none;
	expr = REG_EXPR(dst);
	gcc_assert(DECL_P(expr));
	if (!DECL_NAME(expr))
		return mov_none;
	if (!strncmp(DECL_NAME_POINTER(expr), "rap_retaddr_entry_", sizeof("rap_retaddr_entry_") - 1))
		ret = mov_entry;
	else if (!strncmp(DECL_NAME_POINTER(expr), "rap_retaddr_exit_", sizeof("rap_retaddr_exit_") - 1))
		ret = mov_exit;
	else if (!strncmp(DECL_NAME_POINTER(expr), "rap_cookie_entry", sizeof("rap_cookie_entry") - 1))
		ret = mov_entry;
	else if (!strncmp(DECL_NAME_POINTER(expr), "rap_cookie_exit", sizeof("rap_cookie_exit") - 1))
		ret = mov_exit;
	else
		return mov_none;

	src = SET_SRC(body);
	if (!REG_P(src) || !REG_EXPR(src))
		return mov_none;
	if (!DECL_NAME(REG_EXPR(src)) || !id_equal(DECL_NAME(REG_EXPR(src)), "rap_cookie"))
		return mov_none;
	if (GET_MODE(dst) != GET_MODE(src))
		return mov_none;
	return ret;
}

static void rap_fuse_mov_xor(rtx_insn *mov1, rtx xor1)
{
	rtx body, src;

	body = PATTERN(xor1);
	if (GET_CODE(body) == PARALLEL)
		body = XVECEXP(body, 0, 0);
	src = SET_SRC(body);
	SET_SRC(PATTERN(mov1)) = XEXP(src, 1);
	df_insn_rescan(mov1);
}

static bool rap_xorreg_spills(rtx_insn *xor1, rtx xorreg1, rtx_insn *xor2, unsigned int xorregno2)
{
	df_ref use;

	for (use = DF_REG_USE_CHAIN(REGNO(xorreg1)); use; use = DF_REF_NEXT_REG(use)) {
		rtx_insn *insn;
		rtx body;

		if (DF_REF_IS_ARTIFICIAL(use))
			continue;

		insn = DF_REF_INSN(use);

		if (insn == xor1)
			continue;

		if (insn == xor2)
			continue;

		if (!insn_dominates_insn_p(CDI_DOMINATORS, xor1, insn))
			continue;
		if (!insn_dominates_insn_p(CDI_POST_DOMINATORS, xor2, insn))
			continue;

		body = PATTERN(insn);
		if (GET_CODE(body) == PARALLEL)
			body = XVECEXP(body, 0, 0);
		if (GET_CODE(body) != SET)
			continue;
		if (!REG_P(SET_SRC(body)))
			continue;
		if (ORIGINAL_REGNO(xorreg1) != ORIGINAL_REGNO(SET_SRC(body)))
			continue;

		switch (GET_CODE(SET_DEST(body))) {
		default:
			print_rtl_single(stderr, insn);
			gcc_unreachable();

		case MEM:
			return true;

		case REG:
			if (REGNO(SET_DEST(body)) == REGNO(SET_SRC(body))) {
#if BUILDING_GCC_VERSION >= 4008
				continue;
#else
				gcc_unreachable();
#endif
			}
			if (ORIGINAL_REGNO(SET_DEST(body)) == ORIGINAL_REGNO(SET_SRC(body)))
#if BUILDING_GCC_VERSION >= 4008
				continue;
#else
				gcc_unreachable();
#endif

			if (rap_xorreg_spills(insn, SET_DEST(body), xor2, xorregno2))
				return true;
			continue;
		}
	}

//	gcc_assert(ORIGINAL_REGNO(xorreg1) == xorregno2);
	return false;
}

// eliminate the two xors if the result register can't be clobbered in-between
static unsigned int rap_xor_elimination_execute(void)
{
	rtx_insn *mov1 = 0, *xor1 = 0, *mov2 = 0, *xor2 = 0;
	rtx movreg1 = NULL_RTX, movreg2 = NULL_RTX, xorreg1 = NULL_RTX, xorreg2 = NULL_RTX;
	basic_block xorbb1, xorbb2;
	df_ref use;
	bool spills;

	if (!optimize) {
		calculate_dominance_info(CDI_DOMINATORS);
		calculate_dominance_info(CDI_POST_DOMINATORS);

		df_finish_pass(true);
		df_scan_add_problem();
		df_scan_alloc(NULL);

		df_lr_add_problem();
		df_live_add_problem();
		df_live_set_all_dirty();

		df_chain_add_problem(DF_DU_CHAIN | DF_UD_CHAIN);
		df_md_add_problem();
#if BUILDING_GCC_VERSION >= 6000
		df_mir_add_problem();
#endif
		df_note_add_problem();

		df_rd_add_problem();
		df_scan_blocks();
		df_analyze();
	}

	for (use = DF_REG_USE_CHAIN(rap_reg_id); use && !(mov1 != NULL_RTX && mov2 != NULL_RTX && xor1 != NULL_RTX && xor2 != NULL_RTX); use = DF_REF_NEXT_REG(use)) {
		rtx_insn *insn;
		rtx body;

		if (DF_REF_IS_ARTIFICIAL(use))
			continue;

		insn = DF_REF_INSN(use);
		if (!NONJUMP_INSN_P(insn))
			continue;

		body = PATTERN(insn);
		if (GET_CODE(body) == PARALLEL)
			body = XVECEXP(body, 0, 0);

		if (xor2 == NULL_RTX && rap_match_xor(body, true) != xor_none) {
			xor2 = insn;
			xorreg2 = XEXP(SET_SRC(body), 0);
			continue;
		}

		if (xor1 == NULL_RTX && rap_match_xor(body, true) == xor_entry) {
			xor1 = insn;
			xorreg1 = SET_DEST(body);
			continue;
		}

		if (mov1 == NULL_RTX && rap_match_mov(body) == mov_entry) {
			mov1 = insn;
			movreg1 = SET_DEST(body);
			continue;
		}

		if (mov2 == NULL_RTX && rap_match_mov(body) == mov_exit) {
			mov2 = insn;
			movreg2 = SET_DEST(body);
			continue;
		}
	}

	if (xor1 == NULL_RTX && mov1 == NULL_RTX)
		return 0;

	if (xor2 == NULL_RTX && mov2 == NULL_RTX) {
		debug_tree(current_function_decl);
		gcc_unreachable();
	}

	if (mov1 != NULL_RTX) {
		gcc_assert(xor1 == NULL_RTX);
		for (use = DF_REG_USE_CHAIN(REGNO(movreg1)); use; use = DF_REF_NEXT_REG(use)) {
			rtx_insn *insn;
			rtx body;

			if (DF_REF_IS_ARTIFICIAL(use))
				continue;

			insn = DF_REF_INSN(use);
			if (!NONJUMP_INSN_P(insn))
				continue;

			if (insn == xor2)
				continue;

			body = PATTERN(insn);
			if (GET_CODE(body) == PARALLEL)
				body = XVECEXP(body, 0, 0);

			if (rap_match_xor(body, false) == xor_entry) {
				xor1 = insn;
				xorreg1 = SET_DEST(body);
				break;
			}
		}
	}
	gcc_assert(xor1 != NULL_RTX);

	if (mov2 != NULL_RTX) {
		gcc_assert(xor2 == NULL_RTX);
		for (use = DF_REG_USE_CHAIN(REGNO(movreg2)); use; use = DF_REF_NEXT_REG(use)) {
			rtx_insn *insn;
			rtx body;

			if (DF_REF_IS_ARTIFICIAL(use))
				continue;

			insn = DF_REF_INSN(use);
			if (!NONJUMP_INSN_P(insn))
				continue;

			if (insn == xor1)
				continue;

			body = PATTERN(insn);
			if (GET_CODE(body) == PARALLEL)
				body = XVECEXP(body, 0, 0);

			if (rap_match_xor(body, false) == xor_exit) {
				xor2 = insn;
				xorreg2 = SET_DEST(body);
				break;
			}
		}
	}
	gcc_assert(xor2 != NULL_RTX);

	xorbb1 = BLOCK_FOR_INSN(xor1);
	xorbb2 = BLOCK_FOR_INSN(xor2);

	calculate_dominance_info(CDI_DOMINATORS);
	calculate_dominance_info(CDI_POST_DOMINATORS);

	if (!dominated_by_p(CDI_DOMINATORS, xorbb2, xorbb1)) {
		rtx_insn *temp_insn;
		rtx temp;

		gcc_assert(dominated_by_p(CDI_DOMINATORS, xorbb1, xorbb2));

		temp_insn = xor1;
		xor1 = xor2;
		xor2 = temp_insn;

		temp_insn = mov1;
		mov1 = mov2;
		mov2 = temp_insn;

		xorbb1 = BLOCK_FOR_INSN(xor1);
		xorbb2 = BLOCK_FOR_INSN(xor2);

		temp = PATTERN(xor1);
		if (GET_CODE(temp) == PARALLEL)
			temp = XVECEXP(temp, 0, 0);
		xorreg1 = SET_DEST(temp);

		temp = PATTERN(xor2);
		if (GET_CODE(temp) == PARALLEL)
			temp = XVECEXP(temp, 0, 0);
		xorreg2 = XEXP(SET_SRC(temp), 0);
	} else
		gcc_assert(dominated_by_p(CDI_DOMINATORS, xorbb2, xorbb1));

	spills = rap_xorreg_spills(xor1, xorreg1, xor2, ORIGINAL_REGNO(xorreg2));

	free_dominance_info(CDI_DOMINATORS);
	free_dominance_info(CDI_POST_DOMINATORS);

	if (spills)
		return 0;

	if (mov1)
		rap_fuse_mov_xor(mov1, xor1);
	if (mov2)
		rap_fuse_mov_xor(mov2, xor2);
	delete_insn_and_edges(xor1);
	delete_insn_and_edges(xor2);

	return 0;
}

#define PASS_NAME rap_xor_elimination
#define TODO_FLAGS_FINISH TODO_dump_func | TODO_verify_rtl_sharing | TODO_verify_flow | TODO_df_verify | TODO_df_finish
#include "gcc-generate-rtl-pass.h"
#endif
