/*
 * Copyright 2018-2020 by PaX Team <pageexec@freemail.hu>
 * Licensed under the GPL v2
 *
 * Note: the choice of the license means that the compilation process is
 *       NOT 'eligible' as defined by gcc's library exception to the GPL v3,
 *       but for the kernel it doesn't matter since it doesn't link against
 *       any of the gcc libraries
 *
 * gcc plugin to defend against the Spectre vulnerabilities by instrumentation
 *
 * TODO:
 * - spectre v1/v1.1/v1.2
 *   - static analysis
 *
 * - spectre v2
 *   - arch support
 *   - benchmarking
 *
 * - spectre v4
 *   - static analysis
 *
 * BUGS:
 * - see the TODOs
 */

#include "gcc-common.h"

__visible int plugin_is_GPL_compatible;

enum array_index_type {
	INDEX_TYPE_U8 = 0,
	INDEX_TYPE_S8,
	INDEX_TYPE_U16,
	INDEX_TYPE_S16,
	INDEX_TYPE_U32,
	INDEX_TYPE_S32,
	INDEX_TYPE_U64,
	INDEX_TYPE_S64,
	INDEX_TYPE_MAX
};

#ifdef __cplusplus
static inline enum array_index_type& operator++(enum array_index_type& index_type)
{
	if (index_type >= INDEX_TYPE_MAX)
		gcc_unreachable();

	index_type = static_cast<enum array_index_type>(static_cast<unsigned int>(index_type) + 1);
	return index_type;
}
#endif

static GTY(()) struct {
	const char *name;
	tree decl;
} array_index_nospecs[INDEX_TYPE_MAX] = {
#define INIT_ARRAY_INDEX_NOSPEC(bits)			\
[INDEX_TYPE_##U##bits] = {				\
		.name = "array_index_nospec_u" #bits,	\
		.decl = NULL_TREE,			\
},							\
[INDEX_TYPE_##S##bits] = {				\
		.name = "array_index_nospec_s" #bits,	\
		.decl = NULL_TREE,			\
}

	INIT_ARRAY_INDEX_NOSPEC(8),
	INIT_ARRAY_INDEX_NOSPEC(16),
	INIT_ARRAY_INDEX_NOSPEC(32),
	INIT_ARRAY_INDEX_NOSPEC(64),

#undef INIT_ARRAY_INDEX_NOSPEC
};

static GTY(()) tree barrier_nospec_decl;
static GTY(()) tree barrier_ssb_decl;

static bool verbose;
static bool loopindex;
static bool ssb;
static bool ssb_all;

static struct plugin_info respectre_plugin_info = {
	.version	= "201909181430",
	.help		= "verbose\tlog Spectre instrumented statements\n"
			  "loopindex\tinstrument loop index variable dependent Spectre v1 instances\n"
			  "ssb\t\tinstrument most likely Spectre v4 instances\n"
			  "ssb_all\tinstrument all Spectre v4 instances\n"
};

enum placement { before, after };
enum adjustment { keep, inc };
enum defense { mask, mask_adjust, mask_after, fence };
enum indextype { bounded, notbounded, skipped };

static tree create_new_var(tree type, const char *name)
{
	tree var;

	var = create_tmp_var(type, name);
	add_referenced_var(var);
	mark_sym_for_renaming(var);
	return var;
}

static bool simple_assign_p(gimple_stmt_ptr assign_stmt)
{
	if (!is_gimple_assign(assign_stmt))
		return false;

	switch (gimple_assign_rhs_code(assign_stmt)) {
	default:
		return false;

	case COMPONENT_REF:
#if BUILDING_GCC_VERSION >= 4006
	case MEM_REF:
#endif
	case NOP_EXPR:
	case PARM_DECL:
	case SSA_NAME:
	case VAR_DECL:
		return true;
	}
}

/*
 * compute the bound used for masking
 *
 * this is not a trivial exercise in case the bound check isn't against the index itself but some derived value
 * we have to undo the computation on the bound and use the original index instead of the derived value
 *
 * another bound adjustment is necessary to compensate for equality tests where the bound is off-by-one for masking purposes
 *
 * to top it all off, sometimes the above two adjustments cancel each other out so we don't need to do anything in the end
 */
static tree respectre_compute_bound(gimple_stmt_iterator *gsi, location_t mask_loc, tree block, tree bound, enum adjustment adjust, gimple_stmt_ptr assign_stmt)
{
	tree rhs2;
	enum tree_code op;

	if (!assign_stmt && adjust == keep)
		return bound;

	// adjust bound if it's against a derived index value, see respectre_choose_defense
	if (assign_stmt) {
		rhs2 = gimple_assign_rhs2(assign_stmt);
#if 0
		// this can actually trigger on gcc 8.2 specifically due to an unknown bug/misfeature
		if (TREE_CODE(rhs2) == INTEGER_CST)
			gcc_assert(!integer_zerop(rhs2));
#endif

		switch (gimple_assign_rhs_code(assign_stmt)) {
		default:
			debug_gimple_stmt(assign_stmt);
			debug_tree(rhs2);
			debug_tree(bound);
			gcc_unreachable();

		// check that the bound type is consistent with the op type
		case LSHIFT_EXPR:
		case RSHIFT_EXPR: {
			tree bound_type, rhs1_type;

			bound_type = TREE_TYPE(bound);
			rhs1_type = TREE_TYPE(gimple_assign_rhs1(assign_stmt));
			if (TYPE_MAIN_VARIANT(bound_type) == TYPE_MAIN_VARIANT(rhs1_type))
				break;

			// TODO: care about the type mismatch?
			break;
		}

		case MINUS_EXPR:
		case MULT_EXPR:
		case PLUS_EXPR: {
			tree bound_type, rhs2_type;

			bound_type = TREE_TYPE(bound);
			rhs2_type = TREE_TYPE(rhs2);
			if (TYPE_MAIN_VARIANT(bound_type) == TYPE_MAIN_VARIANT(rhs2_type))
				break;

			if (TYPE_SIZE(bound_type) != TYPE_SIZE(rhs2_type) || TYPE_UNSIGNED(bound_type) != TYPE_UNSIGNED(rhs2_type)) {
				gcc_assert(fold_convertible_p(rhs2_type, bound));
				if (TREE_CODE(bound) == INTEGER_CST)
					bound = fold_convert_loc(mask_loc, rhs2_type, bound);
				else {
					gimple_stmt_ptr stmt;
					tree new_bound;

					new_bound = create_new_var(rhs2_type, "respectre_bound0");
					new_bound = make_ssa_name(new_bound, NULL);
					stmt = gimple_build_assign(new_bound, fold_convert_loc(mask_loc, rhs2_type, bound));
					SSA_NAME_DEF_STMT(new_bound) = stmt;
					gimple_set_location(stmt, mask_loc);
					gimple_set_block(stmt, block);
					gsi_insert_before(gsi, stmt, GSI_SAME_STMT);
					update_stmt(stmt);
					bound = new_bound;
				}
			}
			break;
		}
		}

		switch (gimple_assign_rhs_code(assign_stmt)) {
		default:
			debug_gimple_stmt(assign_stmt);
			debug_tree(rhs2);
			debug_tree(bound);
			gcc_unreachable();

		case LSHIFT_EXPR:
			op = RSHIFT_EXPR;
			break;

		case MINUS_EXPR:
			op = PLUS_EXPR;
			break;

		case MULT_EXPR:
			op = CEIL_DIV_EXPR;
			break;

		case PLUS_EXPR:
			op = MINUS_EXPR;
			break;

		case RSHIFT_EXPR:
			op = LSHIFT_EXPR;
			break;
		}
	}

	if (TREE_CODE(bound) == INTEGER_CST && (!assign_stmt || TREE_CODE(rhs2) == INTEGER_CST)) {
		// adjust bound if it's against a derived index value
		if (assign_stmt) {
			// careful with left shifts, we need 1s in the incoming bits, not 0s
			if (op == LSHIFT_EXPR) {
				tree lsbmask;

				bound = fold_binary_to_constant(op, TREE_TYPE(bound), bound, rhs2);
				lsbmask = fold_binary_to_constant(op, TREE_TYPE(bound), integer_one_node, rhs2);
				lsbmask = fold_binary_to_constant(MINUS_EXPR, TREE_TYPE(bound), lsbmask, integer_one_node);
				bound = fold_binary_to_constant(PLUS_EXPR, TREE_TYPE(bound), bound, lsbmask);
			} else
				bound = fold_binary_to_constant(op, TREE_TYPE(bound), bound, rhs2);
			gcc_assert(bound);
		}

		// adjust bound if it's off-by-one
		if (adjust == inc) {
//			gcc_assert(TREE_CODE(TREE_TYPE(bound)) != INTEGER_TYPE || !tree_int_cst_equal(bound, TYPE_MAX_VALUE(TREE_TYPE(bound))));
			bound = fold_binary_to_constant(PLUS_EXPR, TREE_TYPE(bound), bound, integer_one_node);
			gcc_assert(bound);
		}

		return bound;
	} else {
		gimple_stmt_ptr stmt;
		tree respectre_bound = NULL_TREE;
		bool adjusted = false;

		// adjust bound if it's against a derived index value
		if (assign_stmt) {
			if (adjust == inc) {
				switch (op) {
				default:
					break;

				case MINUS_EXPR:
				case PLUS_EXPR:
					if (TREE_CODE(rhs2) == INTEGER_CST) {
						if (op == PLUS_EXPR)
							gcc_assert(!tree_int_cst_equal(rhs2, TYPE_MAX_VALUE(TREE_TYPE(rhs2))));
						else
							gcc_assert(!tree_int_cst_equal(rhs2, TYPE_MIN_VALUE(TREE_TYPE(rhs2))));
						rhs2 = fold_binary_to_constant(op, TREE_TYPE(rhs2), rhs2, integer_one_node);
						gcc_assert(rhs2);

						// no need to do anything for 0
						if (integer_zerop(rhs2))
							return bound;
					} else if (TREE_CODE(bound) == INTEGER_CST) {
						gcc_assert(!tree_int_cst_equal(bound, TYPE_MAX_VALUE(TREE_TYPE(bound))));
						bound = fold_binary_to_constant(PLUS_EXPR, TREE_TYPE(bound), bound, integer_one_node);
						gcc_assert(bound);
					} else
						break;

					adjusted = true;
					break;
				}
			}

			respectre_bound = create_new_var(TREE_TYPE(bound), "respectre_bound1");
			respectre_bound = make_ssa_name(respectre_bound, NULL);
			stmt = gimple_build_assign_with_ops(op, respectre_bound, bound, rhs2);
			gimple_set_location(stmt, mask_loc);
			gimple_set_block(stmt, block);
			gsi_insert_before(gsi, stmt, GSI_SAME_STMT);
			update_stmt(stmt);
		}

		// adjust bound if it's off-by-one
		if (adjust == inc && !adjusted) {
			if (respectre_bound) {
				bound = respectre_bound;
				respectre_bound = copy_ssa_name(respectre_bound, NULL);
			} else {
				respectre_bound = create_new_var(TREE_TYPE(bound), "respectre_bound2");
				respectre_bound = make_ssa_name(respectre_bound, NULL);
			}

			if (POINTER_TYPE_P(TREE_TYPE(respectre_bound)))
				stmt = gimple_build_assign_with_ops(POINTER_PLUS_EXPR, respectre_bound, bound, build_int_cstu(sizetype, 1));
			else
				stmt = gimple_build_assign_with_ops(PLUS_EXPR, respectre_bound, bound, build_int_cstu(TREE_TYPE(bound), 1));
			gimple_set_location(stmt, mask_loc);
			gimple_set_block(stmt, block);
			gsi_insert_before(gsi, stmt, GSI_SAME_STMT);
			update_stmt(stmt);
		}

		gcc_assert(respectre_bound);
		return respectre_bound;
	}
}

/*
 * compute then apply a mask against index
 * gcc will compute the PHIs as necessary
 */
static tree __respectre_mask_array_index(gimple_stmt_ptr *array_stmt,
					 basic_block in_bound_bb,
					 tree *array_index,
					 tree index,
					 tree bound,
					 enum adjustment adjust,
					 enum defense defense,
					 gimple_stmt_ptr assign_stmt,
					 location_t array_loc)
{
	location_t mask_loc;
	gimple_stmt_ptr stmt;
	gimple_stmt_ptr reference;
	gimple_stmt_iterator gsi;
	gcall *gcallee;
	cgraph_node_ptr node;
	int frequency;
	tree array_index_nospec, respectre_bound, respectre_index, block;
	enum array_index_type index_type;

	gcc_assert(TREE_CODE(index) == SSA_NAME);
	gcc_assert(TREE_CODE(TYPE_SIZE(TREE_TYPE(index))) == INTEGER_CST);

	if (defense == mask_after) {
		gsi = gsi_for_stmt(*array_stmt);
		gsi_next(&gsi);
		reference = *array_stmt;
	} else {
		basic_block reference_bb = in_bound_bb;

		gsi = gsi_start_nondebug_after_labels_bb(reference_bb);
		reference = gsi_end_p(gsi) ? NULL : gsi_stmt(gsi);
		while (!reference && single_succ_p(reference_bb)) {
			gimple_stmt_iterator gsi2;

			reference_bb = single_succ(reference_bb);
			gsi2 = gsi_start_nondebug_after_labels_bb(reference_bb);
			if (!gsi_end_p(gsi2))
				reference = gsi_stmt(gsi2);
		}
	}
	gcc_assert(reference);

	mask_loc = gimple_expansion_point_location(reference);

	block = gimple_block(reference);
	if (!block)
		block = DECL_INITIAL(current_function_decl);
	gcc_assert(block);

	respectre_bound = respectre_compute_bound(&gsi, mask_loc, block, bound, adjust, defense == mask_adjust ? assign_stmt : NULL);

	switch (defense) {
	default:
	case fence:
		gcc_unreachable();

	case mask:
		break;

	case mask_adjust:
		index = *array_index;
		break;

	case mask_after: {
		// create a new index variable if necessary
		tree temp_index;

		temp_index = create_new_var(TREE_TYPE(gimple_assign_lhs(*array_stmt)), "respectre_index0");
		temp_index = make_ssa_name(temp_index, NULL);
		index = gimple_assign_lhs(*array_stmt);
		gcc_assert(TREE_CODE(index) == SSA_NAME);

		gimple_assign_set_lhs(*array_stmt, temp_index);
		update_stmt(*array_stmt);

		stmt = gimple_build_assign(index, temp_index);
		SSA_NAME_DEF_STMT(index) = stmt;
		gimple_set_location(stmt, mask_loc);
		gimple_set_block(stmt, block);
		gsi_insert_before(&gsi, stmt, GSI_NEW_STMT);
		update_stmt(stmt);

		*array_index = index = gimple_assign_lhs(*array_stmt);
		*array_stmt = stmt;
	}
	}

	// find the proper array_index_nospec_* function to use
	switch (TYPE_PRECISION(TREE_TYPE(index))) {
	default:
		debug_tree(index);
		fflush(stderr);
		error_at(array_loc, "array_index_nospec for %u is not defined", TYPE_PRECISION(TREE_TYPE(index)));
		return NULL_TREE;

	case 1 ... 8:
		index_type = INDEX_TYPE_U8;
		break;

	case 9 ... 16:
		index_type = INDEX_TYPE_U16;
		break;

	case 17 ... 32:
		index_type = INDEX_TYPE_U32;
		break;

	case 33 ... 64:
		index_type = INDEX_TYPE_U64;
		break;
	}

	if (!TYPE_UNSIGNED(TREE_TYPE(index)))
		++index_type;

	array_index_nospec = array_index_nospecs[index_type].decl;
	node = cgraph_get_node(array_index_nospec);

	if (array_index_nospec == NULL_TREE || !node) {
		error_at(array_loc, "%s is not defined", array_index_nospecs[index_type].name);
		return NULL_TREE;
	}

	// check the index variable
	if (TREE_CODE(index) == INTEGER_CST) {
		debug_bb(in_bound_bb);
		debug_tree(index);
		gcc_unreachable();
	}

	// convert index to an SSA_NAME if necessary
	if (TREE_CODE(index) != SSA_NAME) {
		respectre_index = create_new_var(TREE_TYPE(index), "respectre_index1");
		respectre_index = make_ssa_name(respectre_index, NULL);
		stmt = gimple_build_assign(respectre_index, fold_convert_loc(mask_loc, TREE_TYPE(respectre_index), index));
		SSA_NAME_DEF_STMT(respectre_index) = stmt;
		gimple_set_location(stmt, mask_loc);
		gimple_set_block(stmt, block);
		gsi_insert_before(&gsi, stmt, GSI_SAME_STMT);
		update_stmt(stmt);

		index = respectre_index;
	}

	// convert index to the type expected by array_index_nospec
	if (!useless_type_conversion_p(TREE_TYPE(TREE_TYPE(array_index_nospec)), TREE_TYPE(index))) {
		gcc_assert(fold_convertible_p(TREE_TYPE(TREE_TYPE(array_index_nospec)), index));
		respectre_index = create_new_var(TREE_VALUE(TYPE_ARG_TYPES(TREE_TYPE(array_index_nospec))), "respectre_index2");
		respectre_index = make_ssa_name(respectre_index, NULL);
		stmt = gimple_build_assign(respectre_index, fold_convert_loc(mask_loc, TREE_TYPE(respectre_index), index));
		SSA_NAME_DEF_STMT(respectre_index) = stmt;
		gimple_set_location(stmt, mask_loc);
		gimple_set_block(stmt, block);
		gsi_insert_before(&gsi, stmt, GSI_SAME_STMT);
		update_stmt(stmt);

		index = respectre_index;
	} else
		respectre_index = index;

	// insert call to array_index_nospec
	if (adjust == inc && bound != respectre_bound && TREE_CODE(bound) != INTEGER_CST)
		stmt = gimple_build_call(array_index_nospec, 3, respectre_index, respectre_bound, boolean_true_node);
	else
		stmt = gimple_build_call(array_index_nospec, 3, respectre_index, respectre_bound, boolean_false_node);
	gimple_set_location(stmt, mask_loc);
	gimple_set_block(stmt, block);
	gcallee = as_a_gcall(stmt);
	gimple_call_set_lhs(gcallee, copy_ssa_name(respectre_index, NULL));
	gsi_insert_before(&gsi, stmt, GSI_SAME_STMT);
	update_stmt(stmt);

	index = respectre_index = gimple_call_lhs(gcallee);

	// convert index to the type of array_index
	if (!useless_type_conversion_p(TREE_TYPE(*array_index), TREE_TYPE(index))) {
		gcc_assert(fold_convertible_p(TREE_TYPE(*array_index), index));
		if (TREE_CODE(*array_index) == SSA_NAME) {
			stmt = gimple_build_assign(copy_ssa_name(*array_index, NULL), fold_convert_loc(mask_loc, TREE_TYPE(*array_index), index));
			SSA_NAME_DEF_STMT(gimple_assign_lhs(stmt)) = stmt;
		} else {
			respectre_index = create_new_var(TREE_TYPE(*array_index), "respectre_index3");
			respectre_index = make_ssa_name(respectre_index, NULL);
			stmt = gimple_build_assign(respectre_index, fold_convert_loc(mask_loc, TREE_TYPE(respectre_index), index));
			SSA_NAME_DEF_STMT(respectre_index) = stmt;
		}
		respectre_index = gimple_assign_lhs(stmt);
		gimple_set_location(stmt, mask_loc);
		gimple_set_block(stmt, block);
		gsi_insert_before(&gsi, stmt, GSI_SAME_STMT);
		update_stmt(stmt);
	}

	// update the cgraph
	frequency = compute_call_stmt_bb_frequency(current_function_decl, in_bound_bb);
	cgraph_create_edge(cgraph_get_node(current_function_decl), node, gcallee, in_bound_bb->count, frequency, in_bound_bb->loop_depth);

	if (verbose) {
		inform(array_loc, "Spectre v1 array index bound %qE", bound);
		inform(mask_loc, "Spectre v1 array index mask adjust: %s constbound: %s", adjust == keep ? "keep" : "inc", TREE_CODE(respectre_bound) == INTEGER_CST ? "yes" : "no");
	}

	switch (TREE_CODE(bound)) {
	case SSA_NAME:
		if (verbose) {
			fflush(stderr);
			fprintf(stderr, "Spectre v1 bound def stmt ");
			print_gimple_stmt(stderr, SSA_NAME_DEF_STMT(bound), 0, TDF_LINENO);
		}
		break;

	case INTEGER_CST:
		break;

	default:
		if (verbose) {
			fflush(stderr);
			fprintf(stderr, "UNKNOWN BOUND TYPE ");
			debug_tree(bound);
		}
		break;
	}

	return respectre_index;
}

static bool reachable_from_p(const_basic_block bb1, const_basic_block bb2)
{
	return dominated_by_p(CDI_DOMINATORS, bb1, bb2);
}

// split_edge can reallocate a PHI stmt so we need to update our own reference to it
// also fix up the post-dominator info while at it
static basic_block respectre_split_edge(edge e, gimple_stmt_ptr *array_stmt, gimple_stmt_ptr *assign_stmt)
{
	basic_block bb;
	tree array_res = NULL_TREE, assign_res = NULL_TREE;

	if (gimple_code(*array_stmt) == GIMPLE_PHI)
		array_res = gimple_phi_result(*array_stmt);
	if (gimple_code(*assign_stmt) == GIMPLE_PHI)
		assign_res = gimple_phi_result(*assign_stmt);

	bb = plugin_split_edge(e);
	gcc_assert(single_succ_p(e->dest));

	if (array_res != NULL_TREE)
		*array_stmt = SSA_NAME_DEF_STMT(array_res);
	if (assign_res != NULL_TREE)
		*assign_stmt = SSA_NAME_DEF_STMT(assign_res);

	return bb;
}

/*
 * when computing the proper mask for the array index would be too cumbersome
 * the strategy is to simply fence the array access itself
 */
static bool respectre_fence_array_index(gimple_stmt_ptr assign_stmt, char spectre_kind)
{
	gimple_stmt_ptr stmt;
	gimple_stmt_ptr reference;
	gimple_stmt_iterator gsi;
	gcall *gcallee;
	cgraph_node_ptr node;
	int frequency;
	tree block, barrier_decl;
	basic_block bb;
	location_t fence_loc;

	switch (spectre_kind) {
	default:
		error_at(gimple_location(assign_stmt), "unknown spectre kind %c", spectre_kind);
		gcc_unreachable();

	case '1':
		barrier_decl = barrier_nospec_decl;
		break;

	case '4':
		barrier_decl = barrier_ssb_decl;
		break;
	}

	node = cgraph_get_node(barrier_decl);
	if (barrier_decl == NULL_TREE || !node) {
		error_at(gimple_location(assign_stmt), "%s is not defined", spectre_kind == '1' ? "barrier_nospec" : "barrier_ssb");
		return false;
	}

	switch (gimple_code(assign_stmt)) {
	default:
		debug_gimple_stmt(assign_stmt);
		gcc_unreachable();

	case GIMPLE_ASSIGN:
		gsi = gsi_for_stmt(assign_stmt);
		reference = assign_stmt;
		break;

	case GIMPLE_PHI:
		gsi = gsi_start_nondebug_after_labels_bb(gimple_bb(assign_stmt));
		reference = gsi_end_p(gsi) ? NULL : gsi_stmt(gsi);
		break;
	}

	gcc_assert(reference);
	fence_loc = reference ? gimple_expansion_point_location(reference) : UNKNOWN_LOCATION;
	block = gimple_block(reference);
	if (!block)
		block = DECL_INITIAL(current_function_decl);
	gcc_assert(block);
	bb = gimple_bb(assign_stmt);
	gcc_assert(bb);

	// insert call to barrier_decl
	stmt = gimple_build_call(barrier_decl, 0);
	gimple_set_location(stmt, fence_loc);
	gimple_set_block(stmt, block);
	gcallee = as_a_gcall(stmt);
	gsi_insert_before(&gsi, stmt, GSI_SAME_STMT);
	update_stmt(stmt);

	// update the cgraph
	frequency = compute_call_stmt_bb_frequency(current_function_decl, bb);
	cgraph_create_edge(cgraph_get_node(current_function_decl), node, gcallee, bb->count, frequency, bb->loop_depth);

	if (verbose) {
		if (spectre_kind == '1')
			inform(fence_loc, "Spectre v1 array index fence");
		else
			inform(fence_loc, "Spectre v4 speculative store bypass fence");
	}

	return true;
}

// handle switch cases as a 'index <= max_label_value' bound check
static bool respectre_extract_switch_cond(const gswitch *switch_stmt,
					  const_basic_block assign_bb,
					  tree index,
					  enum tree_code *cond,
					  tree *lhs,
					  tree *rhs,
					  basic_block *then_bb,
					  basic_block *else_bb)
{
	unsigned int i;

	*cond = LE_EXPR;
	*lhs = index;
	*rhs = NULL_TREE;
	*then_bb = NULL;
	*else_bb = NULL;

	if (!operand_equal_p(gimple_switch_index(switch_stmt), index, 0))
		return false;

	for (i = 0; i < gimple_switch_num_labels(switch_stmt); i++) {
		tree label;
		basic_block case_bb;

		label = gimple_switch_label(switch_stmt, i);
		case_bb = label_to_block_fn(cfun, CASE_LABEL(label));

		if (dominated_by_p(CDI_DOMINATORS, assign_bb, case_bb)) {
			tree label_min, label_max;

			// we can't handle here an assign_stmt in the default case
			if (!i)
				return false;

			// punt on negative case values
			label_min = CASE_LOW(label);
			gcc_assert(label_min);
			if (tree_int_cst_lt(label_min, integer_zero_node))
				return false;

			label_max = CASE_HIGH(label) ? : label_min;
			if (*rhs) {
				if (tree_int_cst_lt(*rhs, label_max)) {
					*rhs = label_max;
					*then_bb = case_bb;
				}
			} else {
				*rhs = label_max;
				*then_bb = case_bb;
			}
		} else if (!*else_bb)
			*else_bb = case_bb;
	}

	if (*rhs == NULL_TREE)
		return false;
	gcc_assert(*then_bb);
	gcc_assert(*else_bb);
	return true;
}

static void respectre_extract_then_else(basic_block cond_bb, basic_block *then_bb, basic_block *else_bb)
{
	edge true_edge, false_edge;

	gcc_assert(EDGE_COUNT(cond_bb->succs) == 2);

	extract_true_false_edges_from_block(cond_bb, &true_edge, &false_edge);
	*then_bb = true_edge->dest;
	*else_bb = false_edge->dest;
}

static void respectre_propagate_respectre_index(gimple_stmt_ptr array_stmt, tree array_index,
						  tree reference, tree respectre_index)
{
	imm_use_iterator imm_iter;
	gimple_stmt_ptr stmt, def_stmt;
	basic_block def_bb;
	bool found;

	def_stmt = SSA_NAME_DEF_STMT(respectre_index);
	def_bb = gimple_bb(def_stmt);

	found = false;
	FOR_EACH_IMM_USE_STMT(stmt, imm_iter, reference) {
		// TODO: replace all uses?
		if (stmt == def_stmt)
			continue;

		switch (gimple_code(stmt)) {
		default:
			debug_gimple_stmt(stmt);
			gcc_unreachable();

		case GIMPLE_ASM: {
			use_operand_p use;

			if (!stmt_dominates_stmt_p(def_stmt, stmt))
				break;

			// TODO: care about this?
			// e.g., fs/xfs/libxfs/xfs_attr_leaf.c:xfs_attr3_leaf_verify and barrier()
			if (TREE_CODE(array_index) != SSA_NAME)
				break;

			FOR_EACH_IMM_USE_ON_STMT(use, imm_iter) {
				SET_USE(use, respectre_index);
			}

			update_stmt(stmt);
			gcc_assert(stmt != array_stmt);
			break;
		}

		case GIMPLE_ASSIGN: {
			use_operand_p use;

			if (!stmt_dominates_stmt_p(def_stmt, stmt))
				break;
			if (TREE_CODE(array_index) != SSA_NAME && !operand_equal_p(array_index, gimple_assign_rhs1(stmt), 0))
				break;

			FOR_EACH_IMM_USE_ON_STMT(use, imm_iter) {
				SET_USE(use, respectre_index);
			}

			if (TREE_CODE(array_index) != SSA_NAME) {
				gcc_assert(operand_equal_p(array_index, gimple_assign_rhs1(stmt), 0));
				gcc_assert(may_propagate_copy_into_stmt(stmt, respectre_index));
				gimple_assign_set_rhs1(stmt, respectre_index);
				gimple_set_vuse(stmt, NULL_TREE);
			}

			update_stmt(stmt);
			if (stmt == array_stmt)
				found = true;
			break;
		}

		case GIMPLE_CALL: {
			use_operand_p use;

			if (!stmt_dominates_stmt_p(def_stmt, stmt))
				break;
			// TODO
			if (TREE_CODE(array_index) != SSA_NAME)
				break;

			FOR_EACH_IMM_USE_ON_STMT(use, imm_iter) {
				SET_USE(use, respectre_index);
			}

			update_stmt(stmt);
			if (stmt == array_stmt)
				found = true;
			break;
		}

		case GIMPLE_COND: {
			use_operand_p use;

			if (!stmt_dominates_stmt_p(def_stmt, stmt))
				break;

			if (TREE_CODE(array_index) != SSA_NAME) {
				debug_gimple_stmt(stmt);
				gcc_unreachable();
			}

			FOR_EACH_IMM_USE_ON_STMT(use, imm_iter) {
				SET_USE(use, respectre_index);
			}

			update_stmt(stmt);
			gcc_assert(stmt != array_stmt);
			break;
		}

		case GIMPLE_DEBUG: {
			use_operand_p use;

			if (!stmt_dominates_stmt_p(def_stmt, stmt))
				break;

			if (TREE_CODE(array_index) != SSA_NAME)
				break;

			FOR_EACH_IMM_USE_ON_STMT(use, imm_iter) {
				SET_USE(use, respectre_index);
			}

			update_stmt(stmt);
			gcc_assert(stmt != array_stmt);
			break;
		}

		case GIMPLE_PHI: {
			use_operand_p use;
			gphi *phi_stmt;

			if (TREE_CODE(array_index) != SSA_NAME)
				break;

			phi_stmt = as_a_gphi(stmt);

			if (stmt != array_stmt && !dominated_by_p(CDI_DOMINATORS, gimple_bb(stmt), def_bb))
				break;

			if (stmt == array_stmt && single_arg_phi_p(phi_stmt)) {
				gcc_assert(is_gimple_assign(def_stmt));
				gimple_assign_set_lhs(def_stmt, gimple_phi_result(stmt));
				respectre_index = gimple_assign_lhs(def_stmt);
				gimple_phi_set_result(phi_stmt, copy_ssa_name(respectre_index, NULL));
				found = true;
				break;
			}

			FOR_EACH_IMM_USE_ON_STMT(use, imm_iter) {
				if (reference != USE_FROM_PTR(use))
					continue;

				if (!dominated_by_p(CDI_DOMINATORS, gimple_phi_arg_edge(phi_stmt, PHI_ARG_INDEX_FROM_USE(use))->src, def_bb))
					continue;

				SET_USE(use, respectre_index);
			}

			update_stmt(stmt);
			if (stmt == array_stmt)
				found = true;
			break;
		}

		case GIMPLE_RETURN: {
			use_operand_p use;

			if (!stmt_dominates_stmt_p(def_stmt, stmt))
				break;

			if (TREE_CODE(array_index) != SSA_NAME) {
				break;
				// TODO: do something about non-SSA indices here?
				// see drivers/pinctrl/sh-pfc/core.c:sh_pfc_phys_to_virt and window->phys
				debug_gimple_stmt(stmt);
				gcc_unreachable();
			}

			FOR_EACH_IMM_USE_ON_STMT(use, imm_iter) {
				SET_USE(use, respectre_index);
			}

			update_stmt(stmt);
			gcc_assert(stmt != array_stmt);
			break;
		}

		case GIMPLE_SWITCH: {
			use_operand_p use;

			if (!stmt_dominates_stmt_p(def_stmt, stmt))
				break;

			if (TREE_CODE(array_index) != SSA_NAME) {
				debug_gimple_stmt(stmt);
				gcc_unreachable();
			}

			FOR_EACH_IMM_USE_ON_STMT(use, imm_iter) {
				SET_USE(use, respectre_index);
			}

			update_stmt(stmt);
			gcc_assert(stmt != array_stmt);
			break;
		}
		}
	}

	// TODO: special case handling of simple casts, should be more generic
	gcc_assert(found);
	return;
}

/*
 * convert these (A and B are conditional expressions with A being an index bound check):
 *
 *   if (A & B)                 if (A | B)
 *   then { array access }      then { array access }
 *   { ... }                    { ... }
 *
 * to these:
 *
 *  if (A)                     if (!A)
 *  then {                     then {
 *    if (B)                     if (!B)
 *    then { array access }      then { goto after }
 *  }                          }
 *  { ... }                    { array access }
 *                             after: { ... }
 *
 * despite the slight visual asymmetry, these two graph conversions are isomorphic
 */
static void respectre_split_cond(basic_block cond_bb, basic_block then_bb, basic_block else_bb, gimple_stmt_ptr cond_stmt, gimple_stmt_ptr *array_stmt, gimple_stmt_ptr *assign_stmt)
{
	gimple_stmt_ptr stmt, andor_stmt;
	gcond *orig_cond;
	location_t loc;
	basic_block other_cond_bb, keep_bb;
	edge cond_edge, keep_edge, other_cond_edge, other_keep_edge;
	gimple_stmt_iterator gsi;
	gphi_iterator gsi2;
	gcond *other_cond;
	int cond_edge_flags, keep_edge_flags;

	stmt = gsi_stmt(gsi_last_nondebug_bb(cond_bb));
	gcc_assert(stmt && gimple_code(stmt) == GIMPLE_COND);

	orig_cond = as_a_gcond(stmt);
	andor_stmt = SSA_NAME_DEF_STMT(gimple_cond_lhs(orig_cond));
	if (is_gimple_call(andor_stmt))
		andor_stmt = SSA_NAME_DEF_STMT(gimple_assign_rhs1(SSA_NAME_DEF_STMT(gimple_call_arg(andor_stmt, 0))));
	gcc_assert(is_gimple_assign(andor_stmt));

	loc = gimple_location(orig_cond);
	switch (gimple_assign_rhs_code(andor_stmt)) {
	default:
		return;
		debug_gimple_stmt(andor_stmt);
		debug_gimple_stmt(cond_stmt);
		debug_gimple_stmt(*assign_stmt);
		debug_gimple_stmt(*array_stmt);
		gcc_unreachable();

	case BIT_AND_EXPR:
	case TRUTH_AND_EXPR:
		other_cond_bb = then_bb;
		keep_bb = else_bb;
		cond_edge_flags = EDGE_TRUE_VALUE;
		keep_edge_flags = EDGE_FALSE_VALUE;
		break;

	case BIT_IOR_EXPR:
	case TRUTH_OR_EXPR:
		other_cond_bb = else_bb;
		keep_bb = then_bb;
		cond_edge_flags = EDGE_FALSE_VALUE;
		keep_edge_flags = EDGE_TRUE_VALUE;
		break;
	}

	cond_edge = find_edge(cond_bb, other_cond_bb);
	keep_edge = find_edge(cond_bb, keep_bb);

	if (gimple_assign_lhs(cond_stmt) == gimple_assign_rhs1(andor_stmt)) {
		other_cond = gimple_build_cond(NE_EXPR, gimple_assign_rhs2(andor_stmt), fold_convert_loc(loc, TREE_TYPE(gimple_assign_rhs2(andor_stmt)), gimple_cond_rhs(orig_cond)), gimple_cond_true_label(orig_cond), gimple_cond_false_label(orig_cond));
		gimple_cond_set_lhs(orig_cond, gimple_assign_rhs1(andor_stmt));
	} else {
		gcc_assert(gimple_assign_lhs(cond_stmt) == gimple_assign_rhs2(andor_stmt));
		other_cond = gimple_build_cond(NE_EXPR, gimple_assign_rhs1(andor_stmt), fold_convert_loc(loc, (TREE_TYPE(gimple_assign_rhs1(andor_stmt))), gimple_cond_rhs(orig_cond)), gimple_cond_true_label(orig_cond), gimple_cond_false_label(orig_cond));
		gimple_cond_set_lhs(orig_cond, gimple_assign_rhs2(andor_stmt));
	}

	gimple_set_location(other_cond, loc);
	gimple_set_block(other_cond, gimple_block(orig_cond));

	gimple_cond_set_rhs(orig_cond, fold_convert_loc(loc, TREE_TYPE(gimple_cond_lhs(orig_cond)), gimple_cond_rhs(orig_cond)));
	update_stmt(orig_cond);

	other_cond_bb = respectre_split_edge(cond_edge, array_stmt, assign_stmt);

	other_cond_edge = single_succ_edge(other_cond_bb);
	other_cond_edge->flags = cond_edge_flags;
	other_cond_edge->probability = cond_edge->probability;

verify_loop_structure();

	other_keep_edge = make_edge(other_cond_bb, keep_bb, keep_edge_flags | (keep_edge->flags & EDGE_IRREDUCIBLE_LOOP));
	other_keep_edge->probability = keep_edge->probability;

	// we have to fix CDI_POST_DOMINATORS after make_edge
	set_immediate_dominator(CDI_POST_DOMINATORS, other_cond_bb, nearest_common_dominator(CDI_POST_DOMINATORS, other_cond_edge->dest, keep_bb));

	if (find_common_loop(keep_bb->loop_father, other_cond_bb->loop_father) != keep_bb->loop_father) {
		remove_bb_from_loops(other_cond_bb);
		add_bb_to_loop(other_cond_bb, keep_bb->loop_father);
	}
	rescan_loop_exit(other_keep_edge, true, false);
verify_loop_structure();

	for (gsi2 = gsi_start_phis(keep_bb); !gsi_end_p(gsi2); gsi_next(&gsi2)) {
		gphi *phi_stmt = gsi_gphi(gsi2);

		SET_PHI_ARG_DEF(phi_stmt, other_keep_edge->dest_idx, PHI_ARG_DEF_FROM_EDGE(phi_stmt, keep_edge));
		update_stmt(phi_stmt);
	}

	gsi = gsi_last_bb(other_cond_bb);
	gsi_insert_after(&gsi, other_cond, GSI_NEW_STMT);
}

/*
 * determine whether 'index' has a use that is a bound check and dominates 'assign_stmt'.
 *
 * 'array_stmt' and 'array_index' represent the memory access we want to ultimately protect,
 * whereas 'assign_stmt' and 'index' represent a potential derivation from 'array_index' that
 * is bound checked.
 *
 * the defense will protect 'array_index' but its kind and placement depend on its relation to 'index'.
 * strategy for masking the array index on the speculatively executed in-bound paths:
 *
 * - create a masked index variable
 *   - initialize it with the index itself (effectively masked with all 1s)
 *     - not all paths may do a bound check
 *       - warn about that?
 *   - for PARM_DECLs insert it in the first BB, split it if necessary
 * - replace the array use of the index with the masked index
 *   - this ensures that we'll get the same PHIs as the index, if any
 * - for each use in a condition:
 *   - determine if it's a bound check and find the in-bound path if it is
 *   - compute a new masked index version in the in-bound path
 *     - extract the mask from the condition
 *       - evaluate INTEGER_CST bounds at compile time
 * - compute PHI if necessary
 */
static bool respectre_mask_array_index(basic_block cond_bb,
				       basic_block then_bb, basic_block else_bb,
				       basic_block in_bound_bb,
				       edge in_bound_edge,
				       gimple_stmt_ptr cond_stmt,
				       gimple_stmt_ptr *array_stmt, tree array_index,
				       gimple_stmt_ptr assign_stmt, tree index, tree bound,
				       enum defense defense,
				       enum adjustment adjustment,
				       tree *respectre_index,
				       location_t array_loc)
{
	tree reference;

	// undo the frontend optimization since only the bound check condition must be instrumented
	if (is_gimple_assign(cond_stmt))
		respectre_split_cond(cond_bb, then_bb, else_bb, cond_stmt, array_stmt, &assign_stmt);

	// ensure that the masked index will continue to dominate its use
	if (!single_pred_p(in_bound_bb) && !single_arg_phi_p(*array_stmt)) {
		basic_block phi_bb;

		in_bound_bb = respectre_split_edge(in_bound_edge, array_stmt, &assign_stmt);
		phi_bb = single_succ(in_bound_bb);

		if (gimple_code(*array_stmt) == GIMPLE_PHI && single_succ(in_bound_bb) != gimple_bb(*array_stmt) && !single_pred_p(phi_bb)) {
			gphi *phi, *array_phi;
			edge e;
			edge_iterator ei;
			unsigned int i, num_args;

			phi = as_a_gphi(create_phi_node(copy_ssa_name(array_index, NULL), phi_bb));

			FOR_EACH_EDGE(e, ei, phi_bb->preds) {
				add_phi_arg(phi, array_index, e, gimple_location(assign_stmt));
			}

			update_stmt(phi);

			array_phi = as_a_gphi(*array_stmt);
			for (i = 0, num_args = gimple_phi_num_args(array_phi); i < num_args; i++) {
				if (gimple_phi_arg_def(array_phi, i) != array_index)
					continue;

				add_phi_arg(array_phi, gimple_phi_result(phi), gimple_phi_arg_edge(array_phi, i), gimple_location(assign_stmt));
				update_stmt(array_phi);
				break;
			}
			gcc_assert(i < num_args);

			if (*array_stmt == assign_stmt)
				assign_stmt = phi;
			*array_stmt = phi;
		}
	}

	*respectre_index = __respectre_mask_array_index(array_stmt, in_bound_bb, &array_index, index, bound, adjustment, defense, assign_stmt, array_loc);

	if (!*respectre_index)
		return false;

	reference = TREE_CODE(array_index) == SSA_NAME ? array_index : gimple_vuse(*array_stmt);

	// fake a PHI stmt if respectre_index doesn't dominate the array_stmt
	if (is_gimple_assign(*array_stmt)) {
		gimple_stmt_ptr def_stmt = SSA_NAME_DEF_STMT(*respectre_index);
		basic_block def_bb = gimple_bb(def_stmt);

		if (!dominated_by_p(CDI_DOMINATORS, gimple_bb(*array_stmt), def_bb)) {
			basic_block dom_bb, phi_bb;
			location_t ref_loc;
			gphi *phi;
			edge e;
			edge_iterator ei;

			phi_bb = gimple_bb(*array_stmt);
			dom_bb = get_immediate_dominator(CDI_DOMINATORS, def_bb);
			while (get_immediate_dominator(CDI_DOMINATORS, phi_bb) != dom_bb) {
				phi_bb = get_immediate_dominator(CDI_DOMINATORS, phi_bb);
				gcc_assert(phi_bb != ENTRY_BLOCK_PTR_FOR_FN(cfun));
			}

			if (TREE_CODE(array_index) == SSA_NAME)
				phi = as_a_gphi(create_phi_node(copy_ssa_name(array_index, NULL), phi_bb));
			else
				phi = as_a_gphi(create_phi_node(make_temp_ssa_name(TREE_TYPE(array_index), NULL, "respectre_index_phi"), phi_bb));

			ref_loc = gimple_location(SSA_NAME_DEF_STMT(reference));
			FOR_EACH_EDGE(e, ei, phi_bb->preds) {
				if (reachable_from_p(e->src, def_bb))
					add_phi_arg(phi, *respectre_index, e, gimple_location(def_stmt));
				else if (TREE_CODE(array_index) == SSA_NAME)
					add_phi_arg(phi, array_index, e, ref_loc);
				else {
					if (useless_type_conversion_p(TREE_TYPE(*respectre_index), TREE_TYPE(index)))
						add_phi_arg(phi, index, e, ref_loc);
					else {
						gimple_stmt_ptr index_def_stmt;

						index_def_stmt = SSA_NAME_DEF_STMT(index);
						gcc_assert(is_gimple_assign(index_def_stmt));
						gcc_assert(gimple_num_ops(index_def_stmt) == 2);
						gcc_assert(useless_type_conversion_p(TREE_TYPE(*respectre_index), TREE_TYPE(gimple_assign_rhs1(index_def_stmt))));
						add_phi_arg(phi, gimple_assign_rhs1(index_def_stmt), e, gimple_location(def_stmt));
					}
				}
			}
			update_stmt(phi);

			*respectre_index = gimple_phi_result(phi);
			gcc_assert(phi == SSA_NAME_DEF_STMT(*respectre_index));
			gcc_assert(dominated_by_p(CDI_DOMINATORS, gimple_bb(*array_stmt), gimple_bb(phi)));
		}
	}

	respectre_propagate_respectre_index(*array_stmt, array_index, reference, *respectre_index);

	return true;
}

static bool respectre_is_index_loop_variable(loop_p loop, tree index)
{
	loop_p ploop;
	unsigned int i;
	gphi_iterator gsi;
	gimple_stmt_ptr def_stmt;

	if (loopindex)
		return false;

	switch (TREE_CODE(index)) {
	default:
		debug_tree(index);
		gcc_unreachable();

	// TODO: handle these?
	case ADDR_EXPR:
	case ARRAY_REF:
	case BIT_FIELD_REF:
	case COMPONENT_REF:
	case CONSTRUCTOR:
	case INTEGER_CST:
	case INDIRECT_REF:
#if BUILDING_GCC_VERSION >= 4006
	case MEM_REF:
#endif
	case TARGET_MEM_REF:
	case VAR_DECL:
		return false;

	case IMAGPART_EXPR:
	case REALPART_EXPR:
		index = TREE_OPERAND(index, 0);
		// FALLTHROUGH

	case SSA_NAME:
		if (SSA_NAME_VAR(index) == NULL_TREE)
			return false;
		break;

	case PARM_DECL:
		break;
	}

	if (!loop)
		return false;
	if (loop->header == ENTRY_BLOCK_PTR_FOR_FN(cfun) && loop->latch == EXIT_BLOCK_PTR_FOR_FN(cfun))
		return false;

	for (gsi = gsi_start_phis(loop->header); !gsi_end_p(gsi); gsi_next(&gsi)) {
		gphi *phi;
		tree phires;

		phi = gsi_gphi(gsi);
		phires = PHI_RESULT(phi);

		if (SSA_NAME_OCCURS_IN_ABNORMAL_PHI(phires))
			continue;

		switch (TREE_CODE(index)) {
		default:
			gcc_unreachable();

		case PARM_DECL:
			if (SSA_NAME_VAR(phires) == index)
				return true;
			continue;

		case SSA_NAME:
			if (SSA_NAME_VAR(phires) == NULL_TREE)
				continue;

			if (SSA_NAME_VAR(phires) == SSA_NAME_VAR(index))
				return true;

			continue;
		}
	}

#if BUILDING_GCC_VERSION <= 4007
	FOR_EACH_VEC_ELT(loop_p, loop->superloops, i, ploop) {
#else
	FOR_EACH_VEC_SAFE_ELT(loop->superloops, i, ploop) {
#endif
		if (respectre_is_index_loop_variable(ploop, index))
			return true;
	}

	if (TREE_CODE(index) != SSA_NAME)
		return false;

	// walk the use-def chain a bit for more witnesses
	// TODO: simple heuristics for now:
	// - track only rhs1 of an influencing assignment
	// - a single argument call with a loop variable returns a loop variable
	def_stmt = SSA_NAME_DEF_STMT(index);
	switch (gimple_code(def_stmt)) {
	default:
		debug_gimple_stmt(def_stmt);
		gcc_unreachable();

	case GIMPLE_ASSIGN:
		return respectre_is_index_loop_variable(loop, gimple_assign_rhs1(def_stmt));

	case GIMPLE_CALL:
		if (gimple_call_num_args(def_stmt) < 1)
			return false;
		return respectre_is_index_loop_variable(loop, gimple_call_arg(def_stmt, 0));

	case GIMPLE_ASM:
	case GIMPLE_NOP:
	case GIMPLE_PHI:
		return false;
	}
}

static bool __respectre_is_interesting_assign(gimple_stmt_ptr use_stmt, tree index)
{
	gassign *assign_stmt;

	if (!is_gimple_assign(use_stmt))
		return false;
	if (!(gimple_bb(use_stmt)->flags & BB_REACHABLE))
		return false;

	assign_stmt = as_a_gassign(use_stmt);
	if (index == gimple_assign_lhs(assign_stmt))
		return false;
	if (TREE_CODE(gimple_assign_lhs(assign_stmt)) != SSA_NAME)
		return false;

	switch (gimple_assign_rhs_code(assign_stmt)) {
	default:
		error_at(gimple_location(assign_stmt), "code %s", get_tree_code_name(gimple_assign_rhs_code(assign_stmt)));
		fflush(stderr);
		debug_gimple_stmt(assign_stmt);
		debug_tree(index);
		gcc_unreachable();

	case ADDR_EXPR:
	case ARRAY_REF:
	case BIT_FIELD_REF:
	case EQ_EXPR:
	case INDIRECT_REF:
	case INTEGER_CST:
	case NE_EXPR:
	case POINTER_PLUS_EXPR:
#if BUILDING_GCC_VERSION >= 8000
	case POINTER_DIFF_EXPR: // TODO: should it be treated as MINUS_EXPR instead?
#endif
	case TARGET_MEM_REF:
	case TRUTH_NOT_EXPR:
		return false;

	case COMPONENT_REF:
#if BUILDING_GCC_VERSION >= 4006
	case MEM_REF:
#endif
		return operand_equal_p(index, gimple_assign_rhs1(assign_stmt), 0);

	case FIX_TRUNC_EXPR:
	case FLOAT_EXPR:
	case RDIV_EXPR:
		return false;

	case CEIL_DIV_EXPR:
	case EXACT_DIV_EXPR:
	case LROTATE_EXPR:
	case LSHIFT_EXPR:
	case RROTATE_EXPR:
	case RSHIFT_EXPR:
	case TRUNC_DIV_EXPR:
	case TRUNC_MOD_EXPR:
		if (index == gimple_assign_rhs2(assign_stmt))
			return false;
		// FALLTHROUGH

	// match a gcc frontend optimization that can coalesce two bound checks
	case GE_EXPR:
	case GT_EXPR:
	case LE_EXPR:
	case LT_EXPR:
		break;

	case ABS_EXPR:
	case BIT_AND_EXPR:
	case BIT_IOR_EXPR:
	case BIT_NOT_EXPR:
	case BIT_XOR_EXPR:
	case COND_EXPR:
	case CONVERT_EXPR:
	case IMAGPART_EXPR:
	case MAX_EXPR:
	case MIN_EXPR:
	case MINUS_EXPR:
	case MULT_EXPR:
	case NEGATE_EXPR:
	case NOP_EXPR:
	case PARM_DECL:
	case PLUS_EXPR:
	case REALPART_EXPR:
	case SSA_NAME:
	case VAR_DECL:
	case VIEW_CONVERT_EXPR:
	case WIDEN_MULT_EXPR:
		break;
	}

	return true;
}

/*
 * we can use index masking if the bound check is done against a value derived in a simple enough way:
 * - the index is cast to a different type for the bound check
 * - the index is adjusted by a simple numerical op for the bound check
 */
static enum defense respectre_choose_defense(tree index, gimple_stmt_ptr assign_stmt, gassign *use_stmt)
{
	tree rhs1, rhs2, next_index;

	next_index = gimple_assign_lhs(use_stmt);

	switch (TREE_CODE(next_index)) {
	default:
		return fence;

	case SSA_NAME:
		break;
	}

	rhs1 = gimple_assign_rhs1(use_stmt);

	if (simple_assign_p(use_stmt))
		return operand_equal_p(rhs1, index, 0) ? mask : fence;

	switch (gimple_assign_rhs_code(use_stmt)) {
	default:
		if (gimple_num_ops(use_stmt) < 3)
			return fence;

		if (rhs1 != index)
			return fence;

		rhs2 = gimple_assign_rhs2(use_stmt);
		if (TREE_CODE(rhs2) != INTEGER_CST)
			return fence;

		return fence;

	// TODO: support these unary ops?
	case BIT_NOT_EXPR:
	case NEGATE_EXPR:
	case PARM_DECL:
		return fence;

	// TODO: support these binary ops?
	case BIT_AND_EXPR:
	case BIT_IOR_EXPR:
	case BIT_XOR_EXPR:
	case MAX_EXPR:
	case MIN_EXPR:
	case TRUNC_DIV_EXPR:
	case TRUNC_MOD_EXPR:
		if (!is_gimple_assign(assign_stmt))
			return fence;

		if (gimple_assign_rhs_code(assign_stmt) != gimple_assign_rhs_code(use_stmt))
			return fence;

		rhs2 = gimple_assign_rhs2(use_stmt);
		if (!operand_equal_p(rhs2, gimple_assign_rhs2(assign_stmt), 0))
			return fence;

		if (!operand_equal_p(rhs1, gimple_assign_rhs1(assign_stmt), 0))
			return fence;
		return mask_after;

#if BUILDING_GCC_VERSION >= 4006
	case MEM_REF:
#endif
		return mask_after;

	case MULT_EXPR:
	case PLUS_EXPR:
		rhs2 = gimple_assign_rhs2(use_stmt);
		if (rhs2 == index) {
			swap_ssa_operands(use_stmt, gimple_assign_rhs1_ptr(use_stmt), gimple_assign_rhs2_ptr(use_stmt));
			update_stmt(use_stmt);
			return mask_adjust;
		}
		// FALLTHROUGH

	case LSHIFT_EXPR:
	case MINUS_EXPR:
	case RSHIFT_EXPR:
		if (rhs1 == index)
			return mask_adjust;

		return fence;
	}
}

static bool respectre_is_bound_check(gimple_stmt_ptr assign_stmt, tree index, tree vindex)
{
	tree rhs;

	if (!is_gimple_assign(assign_stmt))
		return false;

	switch (gimple_assign_rhs_code(assign_stmt)) {
	default:
		return false;

	case LE_EXPR:
	case LT_EXPR:
	case GE_EXPR:
	case GT_EXPR:
		rhs = gimple_assign_rhs1(assign_stmt);
		if (rhs == index)
			return true;
		else if (TREE_CODE(rhs) == SSA_NAME) {
			gimple_stmt_ptr def_stmt = SSA_NAME_DEF_STMT(rhs);
			if (is_gimple_assign(def_stmt) && gimple_assign_rhs1(def_stmt) == index)
				return true;
			if (is_gimple_assign(def_stmt) && vindex && operand_equal_p(gimple_assign_rhs1(def_stmt), vindex, 0))
				return true;
		}

		rhs = gimple_assign_rhs2(assign_stmt);
		if (rhs == index)
			return true;
		else if (TREE_CODE(rhs) == SSA_NAME) {
			gimple_stmt_ptr def_stmt = SSA_NAME_DEF_STMT(rhs);
			if (is_gimple_assign(def_stmt) && gimple_assign_rhs1(def_stmt) == index)
				return true;
			if (is_gimple_assign(def_stmt) && vindex && operand_equal_p(gimple_assign_rhs1(def_stmt), vindex, 0))
				return true;
		}

		return false;
	}
}

static tree respectre_is_bitmask_bounded(tree masked_index, tree bounded_index)
{
	gimple_stmt_ptr assign_stmt;
	tree and_rhs1, and_rhs2;

	if (TREE_CODE(masked_index) != SSA_NAME)
		return NULL_TREE;

	assign_stmt = SSA_NAME_DEF_STMT(masked_index);

	if (!is_gimple_assign(assign_stmt))
		return NULL_TREE;
	if (gimple_assign_rhs_code(assign_stmt) != BIT_AND_EXPR)
		return NULL_TREE;

	and_rhs1 = gimple_assign_rhs1(assign_stmt);
	and_rhs2 = gimple_assign_rhs2(assign_stmt);

	if (bounded_index == and_rhs1)
		return and_rhs2;

	return NULL_TREE;
}

// is rhs assigned/computed from prev_index?
static gimple_stmt_ptr respectre_is_index_assign(tree rhs, tree prev_index)
{
	gimple_stmt_ptr def_stmt;

	if (TREE_CODE(rhs) != SSA_NAME)
		return NULL;

	def_stmt = SSA_NAME_DEF_STMT(rhs);
	if (!__respectre_is_interesting_assign(def_stmt, prev_index))
		return NULL;

	rhs = gimple_assign_rhs1(def_stmt);
	if (operand_equal_p(prev_index, rhs, 0))
		return def_stmt;
	if (gimple_num_ops(def_stmt) < 3)
		return NULL;

	rhs = gimple_assign_rhs2(def_stmt);
	if (operand_equal_p(prev_index, rhs, 0))
		return def_stmt;

	return NULL;
}

/*
 * check whether the use of 'index' in 'cond_stmt' is a bound check and instrument it against speculative execution
 * TODO: check for both upper and lower bound checks?
 */
static bool __respectre_is_index_bounded(basic_block assign_bb,
					 basic_block cond_bb,
					 basic_block then_bb,
					 basic_block else_bb,
					 gimple_stmt_ptr cond_stmt,
					 enum tree_code cond,
					 tree lhs,
					 tree rhs,
					 gimple_stmt_ptr *assign_stmt,
					 tree index,
					 tree vindex,
					 tree *respectre_index,
					 location_t assign_loc)
{
	tree bounded_index, bound, prev_index;
	basic_block join_bb, in_bound_bb, out_of_bound_bb;
	edge in_bound_edge;
	gimple_stmt_ptr use_stmt;
	bool then_dom, else_dom;
	enum { then_side = 0, else_side } in_bound_side;
	enum defense defense;
	enum adjustment adjustment[2];

	// if there're paths from both the then_bb and the else_bb to the assign_bb then this may not be a real bound check
	join_bb = nearest_common_dominator(CDI_POST_DOMINATORS, then_bb, else_bb);
	if (!dominated_by_p(CDI_POST_DOMINATORS, assign_bb, join_bb))
		return false;

	switch (gimple_code(*assign_stmt)) {
	default:
		debug_gimple_stmt(*assign_stmt);
		gcc_unreachable();

	case GIMPLE_ASSIGN:
	case GIMPLE_CALL:
		if (assign_bb == join_bb)
			return false;
		break;

	case GIMPLE_PHI:
		break;
	}

	then_dom = reachable_from_p(assign_bb, then_bb);
	else_dom = reachable_from_p(assign_bb, else_bb);

	if (then_dom == else_dom)
		return false;

	if (then_dom) {
		in_bound_side = then_side;
		in_bound_bb = then_bb;
		out_of_bound_bb = else_bb;
	} else {
		in_bound_side = else_side;
		in_bound_bb = else_bb;
		out_of_bound_bb = then_bb;
	}

	if (join_bb != in_bound_bb && assign_bb != in_bound_bb && get_immediate_dominator(CDI_DOMINATORS, assign_bb) != in_bound_bb && find_edge(out_of_bound_bb, in_bound_bb))
		return false;

	switch (cond) {
	default:
		debug_gimple_stmt(cond_stmt);
		gcc_unreachable();

	case EQ_EXPR:
	case NE_EXPR: {
		// match a bound check disguised as a bitmasking op
		return false; // TODO: fix the breakage caused by instrumenting fs/xfs/xfs_log_recover.c:xlog_write_log_records
		if (cond == EQ_EXPR) {
			if (else_dom)
				return false;
			cond = LE_EXPR;
		} else {
			if (then_dom)
				return false;
			cond = GT_EXPR;
		}

		bounded_index = rhs;
		bound = respectre_is_bitmask_bounded(lhs, bounded_index);
		if (bound == NULL_TREE) {
			bounded_index = lhs;
			bound = respectre_is_bitmask_bounded(rhs, bounded_index);
			if (bound == NULL_TREE)
				return false;
		}
		break;
	}

	case LE_EXPR:
	case LT_EXPR:
		if (then_dom) {
			bounded_index = lhs;
			bound = rhs;
		} else {
			bounded_index = rhs;
			bound = lhs;
		}
		break;

	case GE_EXPR:
	case GT_EXPR:
		if (then_dom) {
			bounded_index = rhs;
			bound = lhs;
		} else {
			bounded_index = lhs;
			bound = rhs;
		}
		break;
	}

	// it can't be our bound checked index if it's not even an SSA variable
	if (TREE_CODE(bounded_index) != SSA_NAME)
		return false;

	// filter out false positive bound checks in preceding loops
	if (bb_loop_depth(cond_bb) && !flow_bb_inside_loop_p(loop_outermost(cond_bb->loop_father), in_bound_bb))
		return false;

	// instrument the in-bound block of a bound check
	in_bound_edge = find_edge(cond_bb, in_bound_bb);
	gcc_assert(in_bound_edge);

	// don't instrument if we have a PHI stmt and the in-bound block has a definition for it already
	if (gimple_code(*assign_stmt) == GIMPLE_PHI) {
		gphi *phi_stmt;

		phi_stmt = as_a_gphi(*assign_stmt);

		tree var = gimple_phi_arg_def(phi_stmt, in_bound_edge->dest_idx);
		switch (TREE_CODE(var)) {
		default:
			break;

		case INTEGER_CST:
			return false;

		case SSA_NAME: {
			gimple_stmt_ptr def_stmt = SSA_NAME_DEF_STMT(var);

			// TODO: should recursively traverse back the PHI and make this determination
			if (gimple_code(def_stmt) == GIMPLE_PHI)
				break;

			if (dominated_by_p(CDI_DOMINATORS, gimple_bb(def_stmt), in_bound_bb))
				return false;
			break;
		}
		}
	}

	/*
	 * decision table to detect a bound check and compute the bound
	 *
	 * cond/[i] | then  else
	 * ---------------------
	 *  i <= b  | b+1   -
	 *  b <= i  | -     b
	 *  i <  b  | b     -
	 *  b <  i  | -     b+1
	 *  i >= b  | -     b
	 *  b >= i  | b+1   -
	 *  i >  b  | -     b+1
	 *  b >  i  | b     -
	 */
	switch (cond) {
	default:
		debug_gimple_stmt(cond_stmt);
		debug_gimple_stmt(*assign_stmt);
		gcc_unreachable();

	case LE_EXPR:
	case GE_EXPR:
		adjustment[then_side] = inc;
		adjustment[else_side] = keep;
		break;

	case LT_EXPR:
	case GT_EXPR:
		adjustment[then_side] = keep;
		adjustment[else_side] = inc;
		break;
	}

	// collect the slice for the index computation (if any) by tracking the reverse data flow

	if (bounded_index == index) {
		// 2. handle a direct use in the condition
		defense = mask;
		use_stmt = *assign_stmt;
		prev_index = index;
	} else {
		// 3. handle simple index derivations in the condition
		// this is needed to get past casts and simple computations on the index
		// before a bound check is done on the result
		prev_index = vindex ? vindex : index;
		use_stmt = SSA_NAME_DEF_STMT(bounded_index);

		gcc_assert(*assign_stmt != use_stmt);

		if (!__respectre_is_interesting_assign(use_stmt, prev_index))
			return false;

		if (operand_equal_p(prev_index, gimple_assign_rhs1(use_stmt), 0)) {
			defense = respectre_choose_defense(prev_index, *assign_stmt, as_a_gassign(use_stmt));
		} else if (gimple_num_ops(use_stmt) == 3 && operand_equal_p(prev_index, gimple_assign_rhs2(use_stmt), 0)) {
			defense = respectre_choose_defense(prev_index, *assign_stmt, as_a_gassign(use_stmt));
		} else if (gimple_assign_cast_p(use_stmt)) {
			// 4. for SSA variables sometimes there's a cast between the computation and the bound check
			tree rhs1;

			rhs1 = gimple_assign_rhs1(use_stmt);
			use_stmt = respectre_is_index_assign(rhs1, prev_index);
			if (!use_stmt)
				return false;

			gcc_assert(*assign_stmt != use_stmt);
			defense = respectre_choose_defense(prev_index, *assign_stmt, as_a_gassign(use_stmt));
		} else {
			// 5. for non-SSA variables we look back another stmt since ops are done on SSA variables only
			tree rhs1;
			gimple_stmt_ptr use_stmt2;

			rhs1 = gimple_assign_rhs1(use_stmt);
			use_stmt2 = respectre_is_index_assign(rhs1, prev_index);
			if (!use_stmt2 && gimple_num_ops(use_stmt) == 3 && gimple_assign_rhs_code(use_stmt) != MINUS_EXPR) {
				rhs1 = gimple_assign_rhs2(use_stmt);
				use_stmt2 = respectre_is_index_assign(rhs1, prev_index);
			}
			if (!use_stmt2)
				return false;
			// TODO: handle more complex bound computations
			if (!simple_assign_p(use_stmt2))
				return false;

			gcc_assert(*assign_stmt != use_stmt2);
			defense = respectre_choose_defense(rhs1, *assign_stmt, as_a_gassign(use_stmt));
		}
		if (vindex)
			prev_index = unshare_expr(vindex);
	}

	switch (defense) {
	case fence:
		// for now bound checked uses in subsequent assignments are treated with a fence instead of masking
		return respectre_fence_array_index(use_stmt, '1');

	default:
		return respectre_mask_array_index(cond_bb, then_bb, else_bb, in_bound_bb, in_bound_edge, cond_stmt, assign_stmt, prev_index, use_stmt, bounded_index, bound, defense, adjustment[in_bound_side], respectre_index, assign_loc);
	}
}

/*
 * find uses of 'index' that are bound checks and instrument the tightest one against speculative execution
 * - determine whether 'index' has a use that is a bound check and dominates 'assign_stmt'
 * - instrument the bound checked uses of 'index'
 */
static enum indextype respectre_is_index_bounded(gimple_stmt_ptr assign_stmt, tree index, tree vindex, location_t assign_loc)
{
	basic_block assign_bb, cond_bb, index_bb;
	gimple_stmt_ptr def_stmt;
	enum indextype indextype = notbounded;

	assign_bb = gimple_bb(assign_stmt);
	gcc_assert(assign_bb);

	if (respectre_is_index_loop_variable(assign_bb->loop_father, vindex ? vindex : index))
		return skipped;

	// 1. walk up the dominator tree up until the index is defined and look for bound checks
	def_stmt = SSA_NAME_DEF_STMT(index);
	if (gimple_nop_p(def_stmt)) {
		gcc_assert(single_succ_p(ENTRY_BLOCK_PTR_FOR_FN(cfun)));
		index_bb = single_succ(ENTRY_BLOCK_PTR_FOR_FN(cfun));
	} else
		index_bb = gimple_bb(def_stmt);

	if (gimple_code(assign_stmt) == GIMPLE_PHI) {
		unsigned int i, num_args;
		gphi *phi_stmt;

		phi_stmt = as_a_gphi(assign_stmt);
		for (i = 0, num_args = gimple_phi_num_args(phi_stmt); i < num_args; i++) {
			if (index == gimple_phi_arg_def(phi_stmt, i))
				break;
		}
		gcc_assert(i < gimple_phi_num_args(phi_stmt));
		cond_bb = gimple_phi_arg_edge(phi_stmt, i)->src;
	} else
		cond_bb = get_immediate_dominator(CDI_DOMINATORS, assign_bb);

	for (;
	     indextype != bounded && cond_bb && dominated_by_p(CDI_DOMINATORS, cond_bb, index_bb);
	     cond_bb = get_immediate_dominator(CDI_DOMINATORS, cond_bb)) {
		gimple_stmt_iterator gsi;
		gimple_stmt_ptr cond_stmt;
		enum tree_code cond;
		tree lhs, rhs, respectre_index = NULL_TREE;
		basic_block then_bb, else_bb;

		gsi = gsi_last_nondebug_bb(cond_bb);
		if (gsi_end_p(gsi))
			continue;
		cond_stmt = gsi_stmt(gsi);
		if (!cond_stmt)
			continue;

		switch (gimple_code(cond_stmt)) {
		default:
			continue;

		case GIMPLE_COND:
			respectre_extract_then_else(cond_bb, &then_bb, &else_bb);
			lhs = gimple_cond_lhs(cond_stmt);
			rhs = gimple_cond_rhs(cond_stmt);
			cond = gimple_cond_code(cond_stmt);

			// match a few 'indirect' bound check patterns
			if (cond == NE_EXPR &&
			    TREE_CODE(lhs) == SSA_NAME &&
			    TREE_CODE(rhs) == INTEGER_CST &&
			    (TREE_CODE(TREE_TYPE(rhs)) == BOOLEAN_TYPE || TREE_CODE(TREE_TYPE(rhs)) == INTEGER_TYPE) &&
			    integer_zerop(rhs)) {
				def_stmt = SSA_NAME_DEF_STMT(lhs);

				// follow back through __builtin_expect
				if (gimple_call_builtin_p(def_stmt, BUILT_IN_EXPECT)) {
					tree arg0, arg1, rhs1, rhs2;

					arg1 = gimple_call_arg(def_stmt, 1);
					if (!integer_zerop(arg1))
						continue;

					arg0 = gimple_call_arg(def_stmt, 0);
					if (TREE_CODE(arg0) != SSA_NAME)
						continue;

					def_stmt = SSA_NAME_DEF_STMT(arg0);
					if (!is_gimple_assign(def_stmt))
						continue;
					if (gimple_assign_cast_p(def_stmt)) {
						arg0 = gimple_assign_rhs1(def_stmt);
						if (TREE_CODE(arg0) != SSA_NAME)
							continue;
						def_stmt = SSA_NAME_DEF_STMT(arg0);
						if (!is_gimple_assign(def_stmt))
							continue;
					}
					if (gimple_assign_rhs_code(def_stmt) != NE_EXPR)
						continue;

					rhs2 = gimple_assign_rhs2(def_stmt);
					if (!integer_zerop(rhs2))
						continue;

					rhs1 = gimple_assign_rhs1(def_stmt);
					if (TREE_CODE(rhs1) != SSA_NAME)
						continue;

					def_stmt = SSA_NAME_DEF_STMT(rhs1);
					if (gimple_assign_cast_p(def_stmt)) {
						rhs1 = gimple_assign_rhs1(def_stmt);
						if (TREE_CODE(rhs1) != SSA_NAME)
							continue;
						def_stmt = SSA_NAME_DEF_STMT(rhs1);
					}
				}

				if (!is_gimple_assign(def_stmt))
					continue;

				switch (gimple_assign_rhs_code(def_stmt)) {
				default:
					continue;

				case BIT_AND_EXPR:
				case BIT_IOR_EXPR:
				case TRUTH_AND_EXPR:
				case TRUTH_OR_EXPR: {
					gimple_stmt_ptr def_rhs1, def_rhs2;

					if (TREE_CODE(gimple_assign_rhs1(def_stmt)) != SSA_NAME)
						continue;
					if (TREE_CODE(gimple_assign_rhs2(def_stmt)) != SSA_NAME)
						continue;

					switch (TREE_CODE(TREE_TYPE(gimple_assign_rhs1(def_stmt)))) {
					default:
						debug_gimple_stmt(def_stmt);
						debug_tree(gimple_assign_rhs1(def_stmt));
						gcc_unreachable();

					case BOOLEAN_TYPE:
					case ENUMERAL_TYPE:
					case INTEGER_TYPE:
						break;
					}

					switch (TREE_CODE(TREE_TYPE(gimple_assign_rhs2(def_stmt)))) {
					default:
						debug_gimple_stmt(def_stmt);
						debug_tree(gimple_assign_rhs2(def_stmt));
						gcc_unreachable();

					case BOOLEAN_TYPE:
					case ENUMERAL_TYPE:
					case INTEGER_TYPE:
						break;
					}

					def_rhs1 = SSA_NAME_DEF_STMT(gimple_assign_rhs1(def_stmt));
					def_rhs2 = SSA_NAME_DEF_STMT(gimple_assign_rhs2(def_stmt));
					if (gimple_bb(def_rhs1) != gimple_bb(def_rhs2))
						continue;

					if (respectre_is_bound_check(def_rhs1, index, vindex)) {
						cond_stmt = def_rhs1;
						lhs = gimple_assign_rhs1(cond_stmt);
						rhs = gimple_assign_rhs2(cond_stmt);
						cond = gimple_assign_rhs_code(cond_stmt);
						if (__respectre_is_index_bounded(assign_bb, cond_bb, then_bb, else_bb, cond_stmt, cond, lhs, rhs, &assign_stmt, index, vindex, &respectre_index, assign_loc))
							indextype = bounded;
					}

					if (indextype == bounded || !respectre_is_bound_check(def_rhs2, index, vindex))
						continue;

					cond_stmt = def_rhs2;
					lhs = gimple_assign_rhs1(cond_stmt);
					rhs = gimple_assign_rhs2(cond_stmt);
					cond = gimple_assign_rhs_code(cond_stmt);
					break;
				}

				case GE_EXPR:
				case GT_EXPR:
				case LE_EXPR:
				case LT_EXPR:
					if (!respectre_is_bound_check(def_stmt, index, vindex))
						continue;

					cond_stmt = def_stmt;
					lhs = gimple_assign_rhs1(cond_stmt);
					rhs = gimple_assign_rhs2(cond_stmt);
					cond = gimple_assign_rhs_code(cond_stmt);
					break;
				}
			}
			break;

		case GIMPLE_SWITCH:
			if (!respectre_extract_switch_cond(as_a_const_gswitch(cond_stmt), assign_bb, vindex ? vindex : index, &cond, &lhs, &rhs, &then_bb, &else_bb))
				continue;
			break;
		}

		if (__respectre_is_index_bounded(assign_bb, cond_bb, then_bb, else_bb, cond_stmt, cond, lhs, rhs, &assign_stmt, index, vindex, &respectre_index, assign_loc))
			indextype = bounded;
	}

	return indextype;
}

static bool respectre_variable_p(tree index)
{
	if (TREE_CODE(index) != SSA_NAME)
		return false;

	if (SSA_NAME_VAR(index) == NULL_TREE)
		return false;

	if (!DECL_NAME(SSA_NAME_VAR(index)))
		return false;

	if (!strncmp(DECL_NAME_POINTER(SSA_NAME_VAR(index)), "respectre_index", sizeof("respectre_index") - 1))
		return true;

	if (!strncmp(DECL_NAME_POINTER(SSA_NAME_VAR(index)), "respectre_mask", sizeof("respectre_mask") - 1))
		return true;

	return false;
}

static tree address_taken_p(tree *tp, int *walk_subtrees, void *data)
{
	struct walk_stmt_info *wi = (struct walk_stmt_info *)data;
	tree index = (tree)wi->info;

	if (TYPE_P(*tp)) {
		*walk_subtrees = 0;
		return NULL_TREE;
	}

	if (TREE_CODE(*tp) != ADDR_EXPR)
		return NULL_TREE;

	return operand_equal_p(TREE_OPERAND(*tp, 0), index, 0) ? index : NULL_TREE;
}

static bool __speculative_store_bypass_possible_var_p(gimple_set *visited, gimple_stmt_iterator gsi, tree index)
{
	while (!gsi_end_p(gsi)) {
		gimple_stmt_ptr stmt;

		stmt = gsi_stmt(gsi);

		if (pointer_set_insert(visited, stmt)) {
			gsi_prev(&gsi);
			continue;
		}

		switch (gimple_code(stmt)) {
		default:
			debug_gimple_stmt(stmt);
			gcc_unreachable();

		case GIMPLE_ASM:
			// TODO: assume inline asm always behaves?
			break;

		case GIMPLE_ASSIGN: {
			struct walk_stmt_info wi = { };

			wi.info = index;
			if (walk_gimple_op(stmt, address_taken_p, &wi) != NULL_TREE)
				return true;
			break;
		}

		case GIMPLE_CALL: {
			unsigned int i, e;
			bool seen_index;

			seen_index = false;
			for (i = 0, e = gimple_call_num_args(stmt); i < e; ++i) {
				tree arg;

				arg = gimple_call_arg(stmt, i);

				switch (TREE_CODE(arg)) {
				default:
					break;

				case SSA_NAME: {
					gimple_stmt_ptr def_stmt;

					def_stmt = SSA_NAME_DEF_STMT(arg);
					if (!is_gimple_assign(def_stmt))
						break;
					if (gimple_assign_rhs_code(def_stmt) != ADDR_EXPR)
						break;

					arg = gimple_assign_rhs1(def_stmt);
					if (operand_equal_p(index, TREE_OPERAND(arg, 0), 0)) {
						pointer_set_insert(visited, def_stmt);
						seen_index = true;
						continue;
					}
					break;
				}

				case ADDR_EXPR:
					if (operand_equal_p(index, TREE_OPERAND(arg, 0), 0))
						seen_index = true;
					break;
				}
			}

			if (seen_index)
				return true;
			break;
		}

		case GIMPLE_COND:
		case GIMPLE_DEBUG:
		case GIMPLE_LABEL:
		case GIMPLE_NOP:
		case GIMPLE_PREDICT:
		case GIMPLE_SWITCH:
			break;
		}

		gsi_prev(&gsi);
	}

	return false;
}

static bool speculative_store_bypass_possible_var_p(gimple_set *visited, gassign *assign_stmt, tree index)
{
	gimple_stmt_iterator gsi;
	basic_block bb;

	gsi = gsi_for_stmt(assign_stmt);
	gsi_prev(&gsi);

	if (__speculative_store_bypass_possible_var_p(visited, gsi, index))
		return true;

	bb = get_immediate_dominator(CDI_DOMINATORS, gimple_bb(assign_stmt));
	while (bb) {
		gsi = gsi_last_nondebug_bb(bb);
		if (__speculative_store_bypass_possible_var_p(visited, gsi, index))
			return true;

		bb = get_immediate_dominator(CDI_DOMINATORS, bb);
	}

	return false;
}

static gassign *index_param_assign_p(tree var, tree index, gimple_stmt_ptr assign_stmt)
{
	gimple_stmt_ptr def_stmt;

	gcc_assert(TREE_CODE(var) == SSA_NAME);
	if (SSA_NAME_VAR(var))
		return NULL;

	def_stmt = SSA_NAME_DEF_STMT(var);
	if (!is_gimple_assign(def_stmt))
		return NULL;
	if (gimple_assign_rhs_code(def_stmt) != PARM_DECL)
		return NULL;
	if (gimple_assign_rhs1(def_stmt) != index)
		return NULL;
	if (!stmt_dominates_stmt_p(def_stmt, assign_stmt))
		return NULL;

	return as_a_gassign(def_stmt);
}

static bool speculative_store_bypass_possible_parm_p(gimple_set *visited, gassign *assign_stmt, tree index)
{
	size_t i;
	tree var;

	FOR_EACH_SSA_NAME(i, var, cfun) {
		gassign *def_stmt;

		def_stmt = index_param_assign_p(var, index, assign_stmt);
		if (!def_stmt)
			continue;

		if (speculative_store_bypass_possible_var_p(visited, def_stmt, index))
			return true;
	}

	return false;
}

/*
 * determine if operations on index can potentially result in SSB
 */
static bool speculative_store_bypass_possible_p(gassign *assign_stmt, tree index)
{
	bool possible = false;
	gimple_set *visited;

	if (!ssb)
		return false;

	if (ssb_all)
		return true;

	if (!TREE_ADDRESSABLE(index))
		return false;

	visited = gimple_pointer_set_create();
	switch (TREE_CODE(index)) {
	default:
		debug_tree(index);
		gcc_unreachable();

	case PARM_DECL:
		possible = speculative_store_bypass_possible_parm_p(visited, assign_stmt, index);
		break;

	case VAR_DECL:
		possible = speculative_store_bypass_possible_var_p(visited, assign_stmt, index);
		break;
	}
	pointer_set_destroy(visited);

	return possible;
}

static bool respectre_handle_index(gimple_set *visited, gimple_stmt_ptr assign_stmt, tree index, bool seen_vuse, location_t assign_loc);

/*
 * helper to respectre_handle_index that manages non-SSA index variables passed via 'vindex'
 */
static bool respectre_handle_vindex(gimple_set *visited, gimple_stmt_ptr assign_stmt, tree index, tree vindex, bool seen_vuse, location_t assign_loc)
{
	enum indextype indextype;
	ssa_op_iter iter;
	def_operand_p def_p __unused;
	imm_use_iterator imm_iter;
	gimple_stmt_ptr def_stmt, use_stmt;
	bool defd = false;

	if (index == NULL_TREE)
		return false;

	indextype = respectre_is_index_bounded(assign_stmt, index, vindex, assign_loc);
	if (indextype != notbounded)
		return indextype == bounded ? true : false;

	def_stmt = SSA_NAME_DEF_STMT(index);
	gcc_assert(def_stmt);

	FOR_EACH_IMM_USE_STMT(use_stmt, imm_iter, index) {
		if (use_stmt == def_stmt)
			continue;
		if (!is_gimple_assign(use_stmt))
			continue;
		if (!operand_equal_p(vindex, gimple_assign_rhs1(use_stmt), 0))
			continue;

		defd = true;
		BREAK_FROM_IMM_USE_STMT(imm_iter);
	}

	if (defd) {
		indextype = respectre_is_index_bounded(assign_stmt, index, vindex, assign_loc);
		if (indextype != notbounded)
			return indextype == bounded ? true : false;
		defd = false;
	}

	FOR_EACH_PHI_OR_STMT_DEF(def_p, def_stmt, iter, SSA_OP_ALL_DEFS) {
		switch (gimple_code(def_stmt)) {
		default:
			continue;

		case GIMPLE_ASSIGN:
			if (operand_equal_p(vindex, gimple_assign_lhs(def_stmt), 0))
				defd = true;
			break;

		case GIMPLE_CALL: {
			tree lhs;
			unsigned int i, e;

			lhs = gimple_call_lhs(def_stmt);
			if (lhs && operand_equal_p(vindex, lhs, 0))
				defd = true;
			else {
				// TODO: verify that a function arg is vindex taken by reference?
				for (i = 0, e = gimple_call_num_args(def_stmt); i < e && !defd; ++i) {
					tree arg, var;

					arg = gimple_call_arg(def_stmt, i);
					if (TREE_CODE(arg) != ADDR_EXPR)
						continue;

					arg = TREE_OPERAND(arg, 0);
					var = VAR_P(vindex) ? vindex : TREE_OPERAND(vindex, 0);
					if (operand_equal_p(var, arg, 0))
						defd = true;
				}
			}
			break;
		}

		case GIMPLE_PHI:
			if (operand_equal_p(vindex, gimple_phi_result(def_stmt), 0))
				defd = true;
			break;
		}

		if (defd)
			break;
	}

	if (!defd)
		return respectre_handle_vindex(visited, assign_stmt, gimple_vuse(def_stmt), vindex, seen_vuse, assign_loc);

	switch (gimple_code(def_stmt)) {
	default:
		debug_gimple_stmt(assign_stmt);
		debug_gimple_stmt(def_stmt);
		gcc_unreachable();

	case GIMPLE_ASSIGN:
		gcc_assert(gimple_num_ops(def_stmt) == 2);

		indextype = respectre_is_index_bounded(assign_stmt, index, vindex, assign_loc);
		if (indextype != notbounded)
			return indextype == bounded ? true : false;

		return respectre_handle_index(visited, def_stmt, gimple_assign_rhs1(def_stmt), seen_vuse || gimple_vuse(def_stmt) != NULL_TREE, assign_loc);

	// TODO: verify that vindex is actually defined by these stmts
	case GIMPLE_CALL:
		indextype = respectre_is_index_bounded(assign_stmt, index, vindex, assign_loc);
		if (indextype != notbounded)
			return indextype == bounded ? true : false;

		// TODO: trace data flow back inside the function in IPA?
		return false;
	}
}

/*
 * this is the core of the reverse data flow tracking engine that determines whether
 * the array index variable is
 *   - bound checked
 *     - thus subject to speculative execution with an out-of-bound value
 *   - not a loop counter
 *     - this is a heuristic for now to avoid instrumenting loop index related variables
 *        - may need refinement since the loop limit itself can be speculated
 *   - TODO: tainted (its value range is untrusted)
 *     - syscall argument
 *     - I/O data (network, USB, disk, etc)
 *
 * the 'textbook' case is to match this gimple construct:
 * if (x_4(D) < array1_size.0_3)
 *   goto <bb 3>;
 * else
 *   goto <bb 4>;
 * <bb 3>:
 * _5 = array1[x_4(D)];
 */
static bool respectre_handle_index(gimple_set *visited, gimple_stmt_ptr assign_stmt, tree index, bool seen_vuse, location_t assign_loc)
{
	tree vindex = NULL_TREE;
	enum indextype indextype;
	bool handled, vuse;

	switch (TREE_CODE(index)) {
	default:
		debug_gimple_stmt(assign_stmt);
		debug_tree(index);
		gcc_unreachable();

	case ADDR_EXPR:
	case INTEGER_CST:
	case REAL_CST:
		return false;

	case PARM_DECL: {
		size_t i;
		tree var;

		gcc_assert(gimple_assign_rhs_code(assign_stmt) == PARM_DECL);
		gcc_assert(gimple_assign_rhs1(assign_stmt) == index);
		// check uses of PARM_DECLs
		FOR_EACH_SSA_NAME(i, var, cfun) {
			if (!index_param_assign_p(var, index, assign_stmt))
				continue;

#if BUILDING_GCC_VERSION >= 4008
			gcc_assert(!virtual_operand_p(var));
#endif
			gimple_assign_set_rhs1(assign_stmt, var);
			update_stmt(assign_stmt);

			switch (respectre_is_index_bounded(assign_stmt, var, NULL_TREE, assign_loc)) {
			case notbounded:
				gimple_assign_set_rhs1(assign_stmt, index);
				update_stmt(assign_stmt);
				continue;

			case bounded:
				return true;

			case skipped:
				gimple_assign_set_rhs1(assign_stmt, index);
				update_stmt(assign_stmt);
				return false;
			}
		}

		// TODO: check arg passed by callers in IPA
		return false;
	}

	case ARRAY_REF:
	case BIT_FIELD_REF:
	case COMPONENT_REF:
#if BUILDING_GCC_VERSION >= 4006
	case MEM_REF:
#endif
	case VAR_DECL:
		vindex = index;
		index = gimple_vuse(assign_stmt);

		if (index == NULL_TREE)
			return false;
		indextype = respectre_is_index_bounded(assign_stmt, index, vindex, assign_loc);
		if (indextype != notbounded)
			return indextype == bounded ? true : false;

		return respectre_handle_vindex(visited, assign_stmt, index, vindex, seen_vuse, assign_loc);

	case INDIRECT_REF:
	case TARGET_MEM_REF:
		// TODO: verify dynamic and global variables
		return false;

	case REALPART_EXPR:
	case IMAGPART_EXPR: // TODO: allow only for DIVMOD but not overflow builtins
		index = TREE_OPERAND(index, 0);
		gcc_assert(TREE_CODE(index) == SSA_NAME);
		// TODO: how to handle modular arithmetic on the array index?
		// FALLTHROUGH

	case SSA_NAME: {
		gimple_stmt_ptr def_stmt;

		def_stmt = SSA_NAME_DEF_STMT(index);
		gcc_assert(def_stmt);

		if (respectre_variable_p(index))
			return false;

		switch (gimple_code(def_stmt)) {
		default:
			debug_gimple_stmt(def_stmt);
			gcc_unreachable();

		case GIMPLE_ASM:
			// TODO: would catch asm array_index_mask_nospec?
			return false;

		case GIMPLE_CALL:
			indextype = respectre_is_index_bounded(assign_stmt, index, vindex, assign_loc);
			if (indextype != notbounded)
				return indextype == bounded ? true : false;

			// TODO: check function return value in IPA
			// TODO: check overflow builtin args
			return false;

		case GIMPLE_PHI: {
			tree phi_res;
			gphi *phi_stmt = as_a_gphi(def_stmt);
			unsigned int i, num_args = gimple_phi_num_args(phi_stmt);
			edge preds[num_args];

			if (pointer_set_insert(visited, def_stmt))
				return false;

			// verify that none of the incoming PHI versions indicate a loop variable
			for (i = 0; !loopindex && i < num_args; i++) {
				tree index2;
				gimple_stmt_ptr def_stmt2;

				index2 = gimple_phi_arg_def(phi_stmt, i);
				if (TREE_CODE(index2) != SSA_NAME)
					continue;

				def_stmt2 = SSA_NAME_DEF_STMT(index2);

				// ignore definitions without a BB, typically for PARM_DECLs
				// and (potentially) uninitialized variables
				if (gimple_nop_p(def_stmt2)) {
					gcc_assert(SSA_NAME_VAR(index2));
					switch (TREE_CODE(SSA_NAME_VAR(index2))) {
					default:
						print_gimple_stmt(stderr, def_stmt, 0, TDF_LINENO);
						print_gimple_stmt(stderr, def_stmt2, 0, TDF_LINENO);
						debug_tree(index2);
						gcc_unreachable();

					case PARM_DECL:
						if (respectre_is_index_loop_variable(gimple_bb(def_stmt)->loop_father, index2))
							return false;
						continue;

					case VAR_DECL:
						continue;
					}
				}
				if (respectre_is_index_loop_variable(gimple_bb(def_stmt2)->loop_father, index2))
					return false;
			}

			// handle a somewhat degenerate case that some older gcc versions create during loop info construction:
			// if the def BB is part of a loop, gcc creates a new entry BB with single arg PHIs instead of simple assignments
			if (single_arg_phi_p(phi_stmt)) {
				tree block;
				location_t loc;
				basic_block phi_bb;
				gimple_stmt_iterator gsi;

				phi_bb = gimple_bb(phi_stmt);
				gsi = gsi_start_nondebug_after_labels_bb(phi_bb);
				if (gsi_end_p(gsi)) {
					tree index2;
					gimple_stmt_ptr def_stmt2;

					index2 = gimple_phi_arg_def(phi_stmt, 0);
					gcc_assert(TREE_CODE(index2) == SSA_NAME);
					gcc_assert(SSA_NAME_VAR(index2) != NULL_TREE);
					def_stmt2 = SSA_NAME_DEF_STMT(index2);
					loc = gimple_location(def_stmt2);
					block = gimple_block(def_stmt2);
				} else {
					loc = gimple_location(gsi_stmt(gsi));
					block = gimple_block(gsi_stmt(gsi));
				}

				def_stmt = gimple_build_assign(gimple_phi_result(phi_stmt), gimple_phi_arg_def(phi_stmt, 0));
				SSA_NAME_DEF_STMT(gimple_assign_lhs(def_stmt)) = def_stmt;
				gimple_set_location(def_stmt, loc);
				gimple_set_block(def_stmt, block);
				gsi_insert_before(&gsi, def_stmt, GSI_SAME_STMT);
				update_stmt(def_stmt);

				gsi = gsi_for_stmt(phi_stmt);
				remove_phi_node(&gsi, false);

				vuse = gimple_vuse(def_stmt) != NULL_TREE;
				handled = respectre_handle_index(visited, def_stmt, index, seen_vuse || vuse, assign_loc);
				return handled;
			}

			// fake an assignment if the reaching definition is a parameter,
			// the rest of the tracking logic can then deal with any bound checks
			for (i = 0; i < num_args; i++) {
				tree index2, block;
				gimple_stmt_ptr def_stmt2, use_stmt;
				edge e;
				gimple_stmt_iterator gsi;
				location_t loc;
				imm_use_iterator imm_iter;

				index2 = gimple_phi_arg_def(phi_stmt, i);
				if (TREE_CODE(index2) != SSA_NAME)
					continue;
				if (SSA_NAME_VAR(index2) == NULL_TREE)
					continue;
				if (TREE_CODE(SSA_NAME_VAR(index2)) != PARM_DECL)
					continue;

				def_stmt2 = SSA_NAME_DEF_STMT(index2);
				if (!gimple_nop_p(def_stmt2))
					continue;

				e = gimple_phi_arg_edge(phi_stmt, i);

				gcc_assert(e->src != ENTRY_BLOCK_PTR_FOR_FN(cfun));

				gsi = gsi_start_nondebug_after_labels_bb(e->src);
				if (gsi_end_p(gsi)) {
					loc = gimple_location(def_stmt2);
					block = gimple_block(def_stmt2);
				} else {
					loc = gimple_location(gsi_stmt(gsi));
					block = gimple_block(gsi_stmt(gsi));
				}
				if (loc == UNKNOWN_LOCATION)
					loc = DECL_SOURCE_LOCATION(SSA_NAME_VAR(index2));

				def_stmt2 = gimple_build_assign(copy_ssa_name(index2, NULL), index2);
				SSA_NAME_DEF_STMT(gimple_assign_lhs(def_stmt2)) = def_stmt2;
				gimple_set_location(def_stmt2, loc);
				gimple_set_block(def_stmt2, block);
				gsi_insert_before(&gsi, def_stmt2, GSI_SAME_STMT);
				update_stmt(def_stmt2);

				add_phi_arg(phi_stmt, gimple_assign_lhs(def_stmt2), e, loc);
				update_stmt(phi_stmt);

				// replace all dominated uses to re-establish the necessary data flow
				FOR_EACH_IMM_USE_STMT(use_stmt, imm_iter, index2) {
					use_operand_p use;

					if (use_stmt == def_stmt2)
						continue;
					if (!stmt_dominates_stmt_p(def_stmt2, use_stmt))
						continue;

					FOR_EACH_IMM_USE_ON_STMT(use, imm_iter) {
						SET_USE(use, gimple_assign_lhs(def_stmt2));
					}
				}
			}

			indextype = respectre_is_index_bounded(assign_stmt, index, vindex, assign_loc);
			if (indextype != notbounded)
				return indextype == bounded ? true : false;

			// collect the edges as they can be shuffled around when the phi stmt gets reallocated
			for (i = 0; i < num_args; i++)
				preds[i] = gimple_phi_arg_edge(phi_stmt, i);

			phi_res = gimple_phi_result(phi_stmt);
			handled = false;
			vuse = gimple_vuse(phi_stmt) != NULL_TREE;
			for (i = 0; i < num_args; i++) {
				tree index2 = PHI_ARG_DEF_FROM_EDGE(phi_stmt, preds[i]);

				if (TREE_CODE(index2) != SSA_NAME)
					continue;
				if (SSA_NAME_VAR(index) != SSA_NAME_VAR(index2))
					continue;

				handled |= respectre_handle_index(visited, phi_stmt, index2, seen_vuse || vuse, assign_loc);
				phi_stmt = as_a_gphi(SSA_NAME_DEF_STMT(phi_res));
				gcc_assert(num_args <= gimple_phi_num_args(phi_stmt));
			}

			return handled;
		}

		case GIMPLE_NOP:
			gcc_assert(TREE_CODE(index) == SSA_NAME);
			if (SSA_NAME_VAR(index) == NULL_TREE)
				return false;
			if (TREE_CODE(SSA_NAME_VAR(index)) != PARM_DECL)
				return false;

			indextype = respectre_is_index_bounded(assign_stmt, index, vindex, assign_loc);
			if (indextype != notbounded)
				return indextype == bounded ? true : false;

			// TODO: check arg passed by callers in IPA
			return false;

		case GIMPLE_ASSIGN: {
			gassign *next_assign_stmt;

			indextype = respectre_is_index_bounded(assign_stmt, index, vindex, assign_loc);
			if (indextype != notbounded)
				return indextype == bounded ? true : false;

			next_assign_stmt = as_a_gassign(def_stmt);

			handled = false;
			vuse = gimple_vuse(def_stmt) != NULL_TREE;
			switch (gimple_num_ops(def_stmt)) {
			default:
				debug_gimple_stmt(def_stmt);
				gcc_unreachable();

#if BUILDING_GCC_VERSION >= 4006
			case 4:
				gcc_assert(gimple_assign_rhs_code(def_stmt) == COND_EXPR);

				handled |= respectre_handle_index(visited, def_stmt, gimple_assign_rhs2(next_assign_stmt), seen_vuse || vuse, assign_loc);
				handled |= respectre_handle_index(visited, def_stmt, gimple_assign_rhs3(next_assign_stmt), seen_vuse || vuse, assign_loc);
				return handled;
#endif

			case 3:
				switch (gimple_assign_rhs_code(def_stmt)) {
				default:
					error_at(gimple_location(def_stmt), "code %s", get_tree_code_name(gimple_assign_rhs_code(def_stmt)));
					fflush(stderr);
					debug_gimple_stmt(def_stmt);
					gcc_unreachable();

				// TODO
				case BIT_AND_EXPR:
					// detect already instrumented variables
					if (respectre_variable_p(gimple_assign_rhs2(next_assign_stmt)))
						return false;
					// FALLTHROUGH

				case BIT_IOR_EXPR:
				case BIT_XOR_EXPR:
				case TRUTH_OR_EXPR:
					handled |= respectre_handle_index(visited, def_stmt, gimple_assign_rhs1(next_assign_stmt), seen_vuse || vuse, assign_loc);
					handled |= respectre_handle_index(visited, def_stmt, gimple_assign_rhs2(next_assign_stmt), seen_vuse || vuse, assign_loc);
					return handled;

				case BIT_NOT_EXPR:
				case NEGATE_EXPR:
					handled = respectre_handle_index(visited, def_stmt, gimple_assign_rhs1(next_assign_stmt), seen_vuse || vuse, assign_loc);
					return handled;

				// ignore rhs2 for asymmetric ops
				case EXACT_DIV_EXPR:
				case LROTATE_EXPR:
				case LSHIFT_EXPR:
#if BUILDING_GCC_VERSION >= 8000
				case POINTER_DIFF_EXPR:
#endif
				case RROTATE_EXPR:
				case RSHIFT_EXPR:
				case TRUNC_DIV_EXPR:
				case TRUNC_MOD_EXPR:
					handled = respectre_handle_index(visited, def_stmt, gimple_assign_rhs1(next_assign_stmt), seen_vuse || vuse, assign_loc);
					return handled;

				case ABS_EXPR:
				case EQ_EXPR:
				case GE_EXPR:
				case GT_EXPR:
				case LE_EXPR:
				case LT_EXPR:
				case MAX_EXPR:
				case MIN_EXPR:
				case MINUS_EXPR:
				case MULT_EXPR:
				case NE_EXPR:
				case NOP_EXPR:
				case PLUS_EXPR:
				case POINTER_PLUS_EXPR:
				case WIDEN_MULT_EXPR:
					handled |= respectre_handle_index(visited, def_stmt, gimple_assign_rhs1(next_assign_stmt), seen_vuse || vuse, assign_loc);
					handled |= respectre_handle_index(visited, def_stmt, gimple_assign_rhs2(next_assign_stmt), seen_vuse || vuse, assign_loc);
					return handled;
				}

			case 2:
				handled = respectre_handle_index(visited, def_stmt, gimple_assign_rhs1(next_assign_stmt), seen_vuse || vuse, assign_loc);
				if (!seen_vuse && vuse && speculative_store_bypass_possible_p(next_assign_stmt, gimple_assign_rhs1(next_assign_stmt)))
					respectre_fence_array_index(def_stmt, '4');
				return handled;
			}
		}
		}
	}
	}
}

/*
 * analyze and, if necessary, instrument the array or ptr+off access
 */
static void respectre_handle_array(gimple_stmt_ptr assign_stmt, tree index, tree array_min, tree array_max)
{
	gimple_set *visited;
	bool vuse;

	switch (TREE_CODE(index)) {
	default:
		print_gimple_stmt(stderr, assign_stmt, 0, TDF_LINENO);
		fprintf(stderr, "INDEX ");debug_tree(index);
		gcc_unreachable();

	case INTEGER_CST:
		return;

	case SSA_NAME:
		if (respectre_variable_p(index))
			return;

#if BUILDING_GCC_VERSION >= 4009
		wide_int index_min, index_max;

		switch (get_range_info(index, &index_min, &index_max)) {
		default:
			break;

		case VR_RANGE: {
			bool in_bounds = true;

			// determine if array_min <= index_min and index_max <= array_max
			if (array_min && TREE_CODE(array_min) == INTEGER_CST) {
				gcc_assert(tree_fits_uhwi_p(array_min));
				if (tree_to_uhwi(array_min) > index_min.to_uhwi())
					in_bounds = false;
			} else
				in_bounds = false;

			if (array_max && TREE_CODE(array_max) == INTEGER_CST) {
				gcc_assert(tree_fits_uhwi_p(array_max));
				if (tree_to_uhwi(array_max) + 1 < index_max.to_uhwi())
					in_bounds = false;
			} else
				in_bounds = false;

			if (in_bounds)
				return;
		}
		}
#endif
		break;
	}

	if (respectre_is_index_loop_variable(gimple_bb(assign_stmt)->loop_father, index))
		return;

	visited = gimple_pointer_set_create();
	vuse = gimple_vuse(assign_stmt) != NULL_TREE;
	respectre_handle_index(visited, assign_stmt, index, vuse, gimple_expansion_point_location(assign_stmt));
	pointer_set_destroy(visited);
}

/*
 * identify memory accesses that are or may lead to an array or ptr+off access
 */
static tree respectre_walk_tree(tree *tp, int *walk_subtrees, void *data)
{
	tree index = *tp;
	tree array_min, array_max;
	gassign *assign_stmt = (gassign *)data;

	*walk_subtrees = 0;

	switch (TREE_CODE(index)) {
	default:
		break;
		debug_gimple_stmt(assign_stmt);
		debug_tree(index);
		gcc_unreachable();

	case INDIRECT_REF: {
		tree ptr, off;
		gimple_stmt_ptr def_stmt;

		ptr = TREE_OPERAND(index, 0);

		if (TREE_CODE(ptr) != SSA_NAME)
			return NULL_TREE;

		def_stmt = SSA_NAME_DEF_STMT(ptr);
		switch (gimple_code(def_stmt)) {
		default:
			// TODO: handle anything besides POINTER_PLUS_EXPR?
			return NULL_TREE;
		}

		off = gimple_assign_rhs2(def_stmt);
		respectre_handle_array(def_stmt, off, NULL_TREE, NULL_TREE);
		break;
	}

#if BUILDING_GCC_VERSION >= 4006
	case MEM_REF: {
		tree ptr, off;
		gimple_stmt_ptr def_stmt;

		ptr = TREE_OPERAND(index, 0);
		off = TREE_OPERAND(index, 1);

		if (TREE_CODE(ptr) != SSA_NAME)
			return NULL_TREE;
		if (TREE_CODE(off) != INTEGER_CST) {
			debug_tree(current_function_decl);
			debug_gimple_stmt(assign_stmt);
			debug_tree(index);
			gcc_unreachable();
		}

		def_stmt = SSA_NAME_DEF_STMT(ptr);
		switch (gimple_code(def_stmt)) {
		default:
			// TODO: handle anything besides POINTER_PLUS_EXPR?
			return NULL_TREE;
		}

		off = gimple_assign_rhs2(def_stmt);
		respectre_handle_array(def_stmt, off, NULL_TREE, NULL_TREE);
		break;
	}
#endif

	// optimizations can turn an ARRAY_REF into a TARGET_MEM_REF
	case TARGET_MEM_REF: {
		tree base, domain;

		base = TMR_BASE(index);
		switch (TREE_CODE(base)) {
		default:
			return NULL_TREE; // TODO

		case ADDR_EXPR:
			break;
		}

		base = TREE_OPERAND(base, 0);
		switch (TREE_CODE(base)) {
		default:
			return NULL_TREE; // TODO

		case STRING_CST:
		case VAR_DECL:
			break;
		}

		switch (TREE_CODE(TREE_TYPE(base))) {
		default:
			return NULL_TREE; // TODO

		case ARRAY_TYPE:
			break;
		}

		domain = TYPE_DOMAIN(TREE_TYPE(base));
		if (domain) {
			array_min = TYPE_MIN_VALUE(domain);
			array_max = TYPE_MAX_VALUE(domain);
		} else {
			array_min = NULL_TREE;
			array_max = NULL_TREE;
		}

#if BUILDING_GCC_VERSION >= 4006
		if (!TMR_INDEX(index))
			index = TMR_INDEX2(index);
		else
#endif
			index = TMR_INDEX(index);
		gcc_assert(index);

		respectre_handle_array(assign_stmt, index, array_min, array_max);
		break;
	}

	case COMPONENT_REF:
		index = TREE_OPERAND(index, 0);

		switch (TREE_CODE(index)) {
		default:
		case VAR_DECL:
			break;

		case ARRAY_REF:
		case COMPONENT_REF:
//		case INDIRECT_REF:
#if BUILDING_GCC_VERSION >= 4006
		case MEM_REF:
#endif
			*walk_subtrees = 1;
		}

		return NULL_TREE;

	case ADDR_EXPR:
		*walk_subtrees = 1;
		return NULL_TREE;

	case ARRAY_REF:
		*walk_subtrees = 1;

		array_min = array_ref_low_bound(index);
		array_max = array_ref_up_bound(index);
		index = TREE_OPERAND(index, 1);

		respectre_handle_array(assign_stmt, index, array_min, array_max);
		break;
	}

	return NULL_TREE;
}

/*
 * Spectre v1 defense:
 * - find array (and pointer+offset) reads where the index is bound checked
 *   - check whether an index use
 *     - is a bound check
 *     - dominates (precedes) the array read
 *   - if none is then walk the use-def chain recursively and check the same
 * - instrument the bound checked index calculation
 *   - lfence
 *   - index masking
 */
static unsigned int respectre_execute(void)
{
	basic_block bb;

	loop_optimizer_init(LOOPS_NORMAL | LOOPS_HAVE_RECORDED_EXITS);
	gcc_assert(current_loops);

	calculate_dominance_info(CDI_DOMINATORS);
	calculate_dominance_info(CDI_POST_DOMINATORS);

	gcc_assert(dom_info_available_p(CDI_DOMINATORS));
	gcc_assert(dom_info_available_p(CDI_POST_DOMINATORS));

	scev_initialize();

	find_unreachable_blocks();

	// 1. loop through BBs and GIMPLE statements
	FOR_EACH_BB_FN(bb, cfun) {
		gimple_stmt_iterator gsi;

		for (gsi = gsi_start_bb(bb); !gsi_end_p(gsi); gsi_next(&gsi)) {
			tree index;
			gimple_stmt_ptr stmt;

			stmt = gsi_stmt(gsi);
			if (!is_gimple_assign(stmt))
				continue;

			// 2. handle Spectre v1 (Bounds Check Bypass)
			if (gimple_assign_rhs_code(stmt) == POINTER_PLUS_EXPR) {
				index = gimple_assign_rhs2(as_a_gassign(stmt));
				respectre_handle_array(stmt, index, NULL_TREE, NULL_TREE);
			}

			// 3. handle Spectre v1 (Bounds Check Bypass)
			index = gimple_assign_rhs1(as_a_gassign(stmt));
			walk_tree_without_duplicates(&index, respectre_walk_tree, stmt);

			// 4. handle Spectre v1.1 (Bounds Check Store Bypass)
			index = gimple_assign_lhs(as_a_gassign(stmt));
			walk_tree_without_duplicates(&index, respectre_walk_tree, stmt);
		}
	}

	scev_finalize();
	free_dominance_info(CDI_POST_DOMINATORS);
	free_dominance_info(CDI_DOMINATORS);
	loop_optimizer_finalize();

	return 0;
}

static bool respectre_gate(void)
{
	tree section;

	section = lookup_attribute("section", DECL_ATTRIBUTES(current_function_decl));
	if (!section || !TREE_VALUE(section))
		return true;

	section = TREE_VALUE(TREE_VALUE(section));

	gcc_assert(strncmp(TREE_STRING_POINTER(section), ".vsyscall_", 10));

	if (!strncmp(TREE_STRING_POINTER(section), ".init.text", 10))
		return false;
	if (!strncmp(TREE_STRING_POINTER(section), ".meminit.text", 13))
		return false;
	if (!strncmp(TREE_STRING_POINTER(section), ".head.text", 10))
		return false;

	return true;
}

#define PASS_NAME respectre
#define TODO_FLAGS_FINISH TODO_verify_ssa | TODO_verify_stmts | TODO_dump_func | TODO_remove_unused_locals | TODO_update_ssa | TODO_cleanup_cfg | TODO_ggc_collect | TODO_rebuild_cgraph_edges | TODO_verify_flow
#include "gcc-generate-gimple-pass.h"

/*
 * the TU must provide array_index_nospec_*. look them up by name early enough and mark them as used
 * otherwise gcc would delete them if the TU itself doesn't use them and before the instrumentation needs them
 */
static void find_nospec_decls(void *event_data __unused, void *data __unused)
{
	tree fndecl = (tree)event_data;
	const char *asmname;
	enum array_index_type index_type;

	if (fndecl == error_mark_node)
		return;

	if (TREE_CODE(fndecl) != FUNCTION_DECL)
		return;

	if (DECL_ARTIFICIAL(fndecl))
		return;

	if (DECL_ABSTRACT_ORIGIN(fndecl) && DECL_ABSTRACT_ORIGIN(fndecl) != fndecl)
		return;

	asmname = DECL_NAME_POINTER(fndecl);
	gcc_assert(asmname[0]);

	for (index_type = INDEX_TYPE_U8; index_type < INDEX_TYPE_MAX; ++index_type) {
		if (!id_equal(DECL_NAME(fndecl), array_index_nospecs[index_type].name))
			continue;

		gcc_assert(array_index_nospecs[index_type].decl == NULL_TREE);
		array_index_nospecs[index_type].decl = fndecl;
		return;
	}

	if (DECL_NAME_LENGTH(fndecl) == sizeof("barrier_nospec") - 1 &&
	    id_equal(DECL_NAME(fndecl), "barrier_nospec")) {
		gcc_assert(!barrier_nospec_decl);
		barrier_nospec_decl = fndecl;
		return;
	}

	if (DECL_NAME_LENGTH(fndecl) == sizeof("barrier_ssb") - 1 &&
	    id_equal(DECL_NAME(fndecl), "barrier_ssb")) {
		gcc_assert(!barrier_ssb_decl);
		barrier_ssb_decl = fndecl;
		return;
	}
}

__visible int plugin_init(struct plugin_name_args *plugin_info, struct plugin_gcc_version *version)
{
	const char * const plugin_name = plugin_info->base_name;
	const int argc = plugin_info->argc;
	const struct plugin_argument * const argv = plugin_info->argv;
	int i;

	static const struct ggc_root_tab gt_ggc_r_gt_respectre[] = {
		{
			.base = &array_index_nospecs[0].decl,
			.nelt = ARRAY_SIZE(array_index_nospecs),
			.stride = sizeof(array_index_nospecs[0]),
			.cb = &gt_ggc_mx_tree_node,
			.pchw = &gt_pch_nx_tree_node
		},
		{
			.base = &barrier_nospec_decl,
			.nelt = 1,
			.stride = sizeof(barrier_nospec_decl),
			.cb = &gt_ggc_mx_tree_node,
			.pchw = &gt_pch_nx_tree_node
		},
		{
			.base = &barrier_ssb_decl,
			.nelt = 1,
			.stride = sizeof(barrier_ssb_decl),
			.cb = &gt_ggc_mx_tree_node,
			.pchw = &gt_pch_nx_tree_node
		},
		LAST_GGC_ROOT_TAB
	};

#if BUILDING_GCC_VERSION >= 12000
	PASS_INFO(respectre, "early_uninit", 1, PASS_POS_INSERT_AFTER);
#else
	PASS_INFO(respectre, "*early_warn_uninitialized", 1, PASS_POS_INSERT_AFTER);
#endif

	if (!plugin_default_version_check(version, &gcc_version)) {
		error_gcc_version(version);
		return 1;
	}

	register_callback(plugin_name, PLUGIN_INFO, NULL, &respectre_plugin_info);

	for (i = 0; i < argc; ++i) {
		if (!strcmp(argv[i].key, "verbose")) {
			verbose = true;
			continue;
		}

		if (!strcmp(argv[i].key, "loopindex")) {
			loopindex = true;
			continue;
		}

		if (!strcmp(argv[i].key, "ssb")) {
			ssb = true;
			continue;
		}

		if (!strcmp(argv[i].key, "ssb_all")) {
			ssb = ssb_all = true;
			continue;
		}

		error(G_("unknown option '-fplugin-arg-%s-%s'"), plugin_name, argv[i].key);
	}

	register_callback(plugin_name, PLUGIN_PRE_GENERICIZE, find_nospec_decls, NULL);
	register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &respectre_pass_info);
	register_callback(plugin_name, PLUGIN_REGISTER_GGC_ROOTS, NULL, (void *)&gt_ggc_r_gt_respectre);

	return 0;
}
