/*
 * Copyright 2013-2022 by PaX Team <pageexec@freemail.hu>
 * Licensed under the GPL v2
 *
 * Note: the choice of the license means that the compilation process is
 *       NOT 'eligible' as defined by gcc's library exception to the GPL v3,
 *       but for the kernel it doesn't matter since it doesn't link against
 *       any of the gcc libraries
 *
 * gcc plugin to forcibly initialize certain local variables that could
 * otherwise leak kernel stack to userland if they aren't properly initialized
 * by later code
 *
 * Homepage: http://pax.grsecurity.net/
 *
 * Options:
 * -fplugin-arg-structleak_plugin-disable
 * -fplugin-arg-structleak_plugin-verbose
 * -fplugin-arg-structleak_plugin-byref
 * -fplugin-arg-structleak_plugin-byref-all
 *
 * Usage:
 * $ # for 4.5/4.6/C based 4.7
 * $ gcc -I`gcc -print-file-name=plugin`/include -I`gcc -print-file-name=plugin`/include/c-family -fPIC -shared -O2 -o structleak_plugin.so structleak_plugin.c
 * $ # for C++ based 4.7/4.8+
 * $ g++ -I`g++ -print-file-name=plugin`/include -I`g++ -print-file-name=plugin`/include/c-family -fPIC -shared -O2 -o structleak_plugin.so structleak_plugin.c
 * $ gcc -fplugin=./structleak_plugin.so test.c -O2
 *
 * TODO: eliminate redundant initializers
 *       increase type coverage
 */

#include "gcc-common.h"

// unused C type flag in all versions 4.5-11
#define TYPE_USERSPACE(TYPE) TYPE_LANG_FLAG_5(TYPE)

__visible int plugin_is_GPL_compatible;

static struct plugin_info structleak_plugin_info = {
	.version	= "202003091015",
	.help		= "disable\tdo not activate plugin\n"
			  "byref\tinit structs passed by reference\n"
			  "byref-all\tinit anything passed by reference\n"
			  "verbose\tprint all initialized variables\n",
};

enum byref { BYREF_NONE, BYREF_STRUCT, BYREF_ALL };

static bool verbose;
static enum byref byref = BYREF_NONE;

static tree handle_user_attribute(tree *node, tree name __unused, tree args __unused, int flags __unused, bool *no_add_attrs)
{
	*no_add_attrs = true;

	// check for types? for now accept everything linux has to offer
	if (TREE_CODE(*node) != FIELD_DECL)
		return NULL_TREE;

	*no_add_attrs = false;
	return NULL_TREE;
}

static struct attribute_spec user_attr = ATTRIBUTE_SPEC_INIT("user", 0, 0, false, false, false, handle_user_attribute, true, NULL);

static void register_attributes(void *event_data __unused, void *data __unused)
{
	register_attribute(&user_attr);
}

static tree get_field_type(tree field)
{
	return strip_array_types(TREE_TYPE(field));
}

static bool is_userspace_type(tree type)
{
	tree field;

	for (field = TYPE_FIELDS(type); field; field = TREE_CHAIN(field)) {
		tree fieldtype = get_field_type(field);
		enum tree_code code = TREE_CODE(fieldtype);

		if (code == RECORD_TYPE || code == UNION_TYPE)
			if (is_userspace_type(fieldtype))
				return true;

		if (lookup_attribute("user", DECL_ATTRIBUTES(field)))
			return true;
	}
	return false;
}

static void finish_type(void *event_data, void *data __unused)
{
	tree type = (tree)event_data;

	if (type == NULL_TREE || type == error_mark_node)
		return;

#if BUILDING_GCC_VERSION >= 5000
	if (TREE_CODE(type) == ENUMERAL_TYPE)
		return;
#endif

	if (TYPE_USERSPACE(type))
		return;

	if (is_userspace_type(type))
		TYPE_USERSPACE(type) = 1;
}

static void structleak_instrument(gimple_stmt_iterator *gsi, tree var)
{
	location_t loc;
	gimple_stmt_ptr init_stmt;
	tree type;

	loc = DECL_SOURCE_LOCATION(var);
#if BUILDING_GCC_VERSION >= 4008
	if (LOCATION_BLOCK(loc))
		loc = UNKNOWN_LOCATION;
#endif

	// build the initializer expression
	type = TREE_TYPE(var);
	if (AGGREGATE_TYPE_P(type))
		init_stmt = gimple_build_call(builtin_decl_implicit(BUILT_IN_MEMSET), 3, build_fold_addr_expr_loc(loc, var), integer_zero_node, TYPE_SIZE_UNIT(type));
	else
		init_stmt = gimple_build_assign(var, fold_convert_loc(loc, type, integer_zero_node));

	// build the initializer stmt
	gimple_set_location(init_stmt, loc);
//	gimple_set_block(init_stmt, TREE_BLOCK(var));
	gsi_insert_before(gsi, init_stmt, GSI_NEW_STMT);
	update_stmt(init_stmt);

	// these aren't the 0days you're looking for
	if (verbose)
		inform(loc, "%s variable %qD will be forcibly initialized", TYPE_USERSPACE(type) ? "userspace" : "byref", var);
}

static bool is_arg_var_ref(const_tree var, tree arg)
{
	arg = get_inner_arg(arg);

	switch (TREE_CODE(arg)) {
	default:
		return false;

	case ADDR_EXPR:
		return is_arg_var_ref(var, TREE_OPERAND(arg, 0));

	case SSA_NAME: {
		gimple_stmt_ptr def_stmt;

		def_stmt = SSA_NAME_DEF_STMT(arg);
		if (!is_gimple_assign(def_stmt))
			return false;
		if (gimple_assign_rhs_code(def_stmt) != ADDR_EXPR)
			return false;

		arg = TREE_OPERAND(gimple_assign_rhs1(def_stmt), 0);
		break;
	}

	case VAR_DECL:
		break;
	}

	gcc_assert(operand_equal_p(arg, var, 0) == (arg == var));
	return operand_equal_p(arg, var, 0);
}

enum access_kind { access_unknown, access_load, access_ctor, access_store };

static enum access_kind structleak_classify_stmt(const_gimple_stmt_ptr stmt, const_tree var)
{
	unsigned int i, e;

	switch (gimple_code(stmt)) {
	default:
		print_gimple_stmt(stderr, stmt, 0, TDF_LINENO);
		debug_tree(var);
		gcc_unreachable();

	case GIMPLE_COND:
	case GIMPLE_NOP:
	case GIMPLE_SWITCH: {
		enum access_kind kind = access_unknown;

		for (i = 0, e = gimple_num_ops(stmt); i < e; i++) {
			tree arg = gimple_op(stmt, i);
			if (!arg)
				continue;
			if (!is_arg_var_ref(var, arg))
				continue;
			if (TREE_CODE(arg) != ADDR_EXPR) {
				debug_gimple_stmt(stmt);
				debug_tree(var);
				debug_tree(arg);
				gcc_unreachable();
			}

			if (kind != access_unknown)
				continue;
			kind = access_load;

		}
		return kind;
	}

	case GIMPLE_DEBUG:
	case GIMPLE_GOTO:
	case GIMPLE_LABEL:
	case GIMPLE_RETURN:
		return access_unknown;

	case GIMPLE_ASM: {
		const gasm *asm_stmt = as_a_const_gasm(stmt);

		for (i = 0, e = gimple_asm_noutputs(asm_stmt); i < e; ++i) {
			tree arg;

			arg = TREE_VALUE(gimple_asm_output_op(asm_stmt, i));
			if (!is_arg_var_ref(var, arg))
				continue;

			return access_load;
		}

		for (i = 0, e = gimple_asm_ninputs(asm_stmt); i < e; ++i) {
			tree arg;

			arg = TREE_VALUE(gimple_asm_input_op(asm_stmt, i));
			if (!is_arg_var_ref(var, arg))
				continue;

			return access_store;
		}

		return access_unknown;
	}

	case GIMPLE_CALL: {
		tree arg, lhs, offset;
		poly_int64_pod bitsize, bitpos;
		enum machine_mode mode;
		int unsignedp, reversep, volatilep;

		lhs = gimple_call_lhs(stmt);
		if (lhs && operand_equal_p(var, get_inner_reference(lhs, &bitsize, &bitpos, &offset, &mode, &unsignedp, &reversep, &volatilep, true), 0))
			return access_store;

		for (i = 0, e = gimple_call_num_args(stmt); i < e; ++i) {
			arg = gimple_call_arg(stmt, i);
			if (is_arg_var_ref(var, arg))
				break;
		}

		if (i == e)
			return access_unknown;

		if (POINTER_TYPE_P(arg))
			return TYPE_READONLY(TREE_TYPE(TREE_TYPE(arg))) ? access_load : access_store;
		return access_load;
	}

	case GIMPLE_ASSIGN: {
		tree offset, lhs;
		poly_int64_pod bitsize, bitpos;
		enum machine_mode mode;
		int unsignedp, reversep, volatilep;

		lhs = gimple_assign_lhs(stmt);
		if (operand_equal_p(var, get_inner_reference(lhs, &bitsize, &bitpos, &offset, &mode, &unsignedp, &reversep, &volatilep, true), 0)) {
			tree typesize;

			if (gimple_clobber_p(stmt))
				return access_unknown;
			if (offset != NULL_TREE || known_gt_poly_int64(bitpos, 0))
				return access_store;
			typesize = TYPE_SIZE(TREE_TYPE(var));
			if (!INTEGRAL_TYPE_P(typesize))
				return access_store;
			if (known_ne_poly_int64(bitsize, TREE_INT_CST_LOW(TYPE_SIZE(TREE_TYPE(lhs)))))
				return access_store;

			return TREE_CODE(gimple_assign_rhs1(stmt)) == CONSTRUCTOR ? access_ctor : access_store;
		}

		for (i = 1, e = gimple_num_ops(stmt); i < e; ++i)
			if (is_arg_var_ref(var, gimple_op(stmt, i)))
				return access_load;

		return access_unknown;
	}
	}
}

static void structleak_initialize(int *postorder, unsigned int postorder_num, sbitmap var_use_blocks, tree var)
{
	unsigned int i, loopdepth;
	loop_p loop = NULL;
	basic_block var_scope_entry_bb;
	gimple_stmt_iterator gsi;
	gimple_stmt_ptr first_use = NULL;

	// find all BBs with uses of var
	for (i = postorder_num; i; i--) {
		basic_block bb;
		gphi_iterator gsi2;

		bb = BASIC_BLOCK_FOR_FN(cfun, postorder[i - 1]);

		for (gsi = gsi_start_bb(bb); !gsi_end_p(gsi); gsi_next(&gsi)) {
			gimple_stmt_ptr stmt = gsi_stmt(gsi);

			switch (structleak_classify_stmt(stmt, var)) {
			case access_load:
			case access_store:
				sbitmap_set_bit(var_use_blocks, bb->index);
				if (!first_use)
					first_use = stmt;
				if (bb_loop_depth(bb))
					loop = find_common_loop(loop, bb->loop_father);
				continue;

			case access_unknown:
				continue;

			case access_ctor:
				gcc_assert(!first_use);
				gcc_assert(sbitmap_empty_p(var_use_blocks));
				return;
			}
		}

		for (gsi2 = gsi_start_phis(bb); !gsi_end_p(gsi2); gsi_next(&gsi2)) {
			unsigned int i, num_args;
			gphi *phi_stmt = gsi_gphi(gsi2);

			for (i = 0, num_args = gimple_phi_num_args(phi_stmt); i < num_args; i++) {
				if (!is_arg_var_ref(var, gimple_phi_arg_def(phi_stmt, i)))
					continue;

				sbitmap_set_bit(var_use_blocks, bb->index);
				if (!first_use)
					first_use = phi_stmt;
				if (bb_loop_depth(bb))
					loop = find_common_loop(loop, bb->loop_father);
			}
		}
	}
	if (sbitmap_empty_p(var_use_blocks))
		return;

	if (loop && loop_depth(loop)) {
		sbitmap_set_bit(var_use_blocks, loop->header->index);
		loopdepth = loop_depth(loop);
	}
		loopdepth = 1;

	var_scope_entry_bb = nearest_common_dominator_for_sbitmap(CDI_DOMINATORS, var_use_blocks);
	gcc_assert(var_scope_entry_bb);
	gcc_assert(var_scope_entry_bb != ENTRY_BLOCK_PTR_FOR_FN(cfun));

	if (loop && loop_depth(loop) && bb_loop_depth(var_scope_entry_bb) >= loop_depth(loop) && sbitmap_bit_p(var_use_blocks, var_scope_entry_bb->index)) {
		var_scope_entry_bb = get_immediate_dominator(CDI_DOMINATORS, var_scope_entry_bb);
		gcc_assert(var_scope_entry_bb);
		gcc_assert(var_scope_entry_bb != ENTRY_BLOCK_PTR_FOR_FN(cfun));
	}

	while (bb_loop_depth(var_scope_entry_bb) >= loopdepth) {
		var_scope_entry_bb = get_immediate_dominator(CDI_DOMINATORS, var_scope_entry_bb);
		gcc_assert(var_scope_entry_bb);
	}
	gcc_assert(var_scope_entry_bb != ENTRY_BLOCK_PTR_FOR_FN(cfun));

	if (first_use && gimple_bb(first_use) == var_scope_entry_bb) {
		if (gimple_code(first_use) == GIMPLE_PHI) {
			var_scope_entry_bb = get_immediate_dominator(CDI_DOMINATORS, var_scope_entry_bb);
			gcc_assert(var_scope_entry_bb);
			gcc_assert(var_scope_entry_bb != ENTRY_BLOCK_PTR_FOR_FN(cfun));
			gsi = gsi_start_nondebug_after_labels_bb(var_scope_entry_bb);
		} else
			gsi = gsi_for_stmt(first_use);
	} else
		gsi = gsi_start_nondebug_after_labels_bb(var_scope_entry_bb);

	structleak_instrument(&gsi, var);
}

static unsigned int structleak_execute(void)
{
	tree var;
	unsigned int i;
	int *postorder;
	unsigned int postorder_num;
	sbitmap var_use_blocks;

	loop_optimizer_init(LOOPS_NORMAL | LOOPS_HAVE_RECORDED_EXITS);
	gcc_assert(current_loops);

	calculate_dominance_info(CDI_DOMINATORS);
	calculate_dominance_info(CDI_POST_DOMINATORS);

	// split the first bb where we can put the forced initializers
	get_single_pred_entry_block();

	var_use_blocks = sbitmap_alloc(last_basic_block_for_fn(cfun));

	postorder = XNEWVEC(int, n_basic_blocks_for_fn(cfun));
	postorder_num = post_order_compute(postorder, true, false);
	gcc_assert(ENTRY_BLOCK == postorder[postorder_num - 1]);

	// enumerate all local variables and forcibly initialize our targets
	FOR_EACH_LOCAL_DECL(cfun, i, var) {
		tree type = strip_array_types(TREE_TYPE(var));

		gcc_assert(DECL_P(var));
		// this is a simplified version of auto_var_in_fn_p as we only deal with actual variables
		if (DECL_CONTEXT(var) != current_function_decl)
			continue;
		// punt on volatile vars
		if (TREE_SIDE_EFFECTS(var))
			continue;

		switch (TREE_CODE(var)) {
		default:
			debug_tree(var);
			gcc_unreachable();

		case VAR_DECL:
			if (is_global_var(var))
				continue;
			break;
		}

		switch (byref) {
		case BYREF_NONE:
			if (!TYPE_USERSPACE(type))
				continue;
			// FALLTHROUGH

		case BYREF_STRUCT:
			// only care about structure types unless byref-all
			if (!RECORD_OR_UNION_TYPE_P(type))
				continue;
			// FALLTHROUGH

		case BYREF_ALL:
			if (byref != BYREF_NONE && !TREE_ADDRESSABLE(var))
				continue;
			break;
		}

		sbitmap_clear(var_use_blocks);
		structleak_initialize(postorder, postorder_num, var_use_blocks, var);
	}

	XDELETEVEC(postorder);
	sbitmap_free(var_use_blocks);
	free_dominance_info(CDI_DOMINATORS);
	free_dominance_info(CDI_POST_DOMINATORS);
	loop_optimizer_finalize();

	return 0;
}

#if BUILDING_GCC_VERSION >= 7000
static bool structleak_gate(void)
{
	if (asan_sanitize_use_after_scope()) {
		error_at(DECL_SOURCE_LOCATION(current_function_decl), "structleak is not compatible with -fsanitize-address-use-after-scope");
		return false;
	}

	return true;
}
#endif

#define PASS_NAME structleak
#if BUILDING_GCC_VERSION < 7000
#define NO_GATE
#endif
#define PROPERTIES_REQUIRED PROP_cfg
#define TODO_FLAGS_FINISH TODO_verify_il | TODO_verify_ssa | TODO_verify_stmts | TODO_dump_func | TODO_remove_unused_locals | TODO_update_ssa | TODO_ggc_collect | TODO_verify_flow
#include "gcc-generate-gimple-pass.h"

__visible int plugin_init(struct plugin_name_args *plugin_info, struct plugin_gcc_version *version)
{
	int i;
	const char * const plugin_name = plugin_info->base_name;
	const int argc = plugin_info->argc;
	const struct plugin_argument * const argv = plugin_info->argv;
	bool enable = true;

	PASS_INFO(structleak, "sra", 1, PASS_POS_INSERT_AFTER);

	if (!plugin_default_version_check(version, &gcc_version)) {
		error_gcc_version(version);
		return 1;
	}

	if (strncmp(lang_hooks.name, "GNU C", 5) || !strncmp(lang_hooks.name, "GNU C+", 6)) {
		inform(UNKNOWN_LOCATION, G_("%s supports C only, not %s"), plugin_name, lang_hooks.name);
		enable = false;
	}

	for (i = 0; i < argc; ++i) {
		if (!strcmp(argv[i].key, "disable")) {
			enable = false;
			continue;
		}
		if (!strcmp(argv[i].key, "verbose")) {
			verbose = true;
			continue;
		}
		if (!strcmp(argv[i].key, "byref")) {
			byref = BYREF_STRUCT;
			continue;
		}
		if (!strcmp(argv[i].key, "byref-all")) {
			byref = BYREF_ALL;
			continue;
		}

		error(G_("unknown option '-fplugin-arg-%s-%s'"), plugin_name, argv[i].key);
	}

	register_callback(plugin_name, PLUGIN_INFO, NULL, &structleak_plugin_info);
	if (enable) {
		register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &structleak_pass_info);
		register_callback(plugin_name, PLUGIN_FINISH_TYPE, finish_type, NULL);
	}
	register_callback(plugin_name, PLUGIN_ATTRIBUTES, register_attributes, NULL);

	return 0;
}
