/*
 * Copyright 2018-2020 by PaX Team <pageexec@freemail.hu>
 * Licensed under the GPL v2
 *
 * generic context tracking facility similar to the one provided by sparse
 */

#include "utilities.h"

static bool split_context_attribute(tree args, tree *lock, tree *in, tree *out)
{
	*in = TREE_VALUE(args);

	if (TREE_CODE(*in) != INTEGER_CST) {
		*lock = *in;
		args = TREE_CHAIN(args);
		*in = TREE_VALUE(args);
	} else
		*lock = NULL_TREE;

	args = TREE_CHAIN(args);
	if (*lock && !args)
		return false;

	*out = TREE_VALUE(args);
	return true;
}

static tree handle_context_attribute(tree *node, tree name, tree args, int flags, bool *no_add_attrs)
{
	*no_add_attrs = true;
	tree lock, in, out;

	if (TREE_CODE(*node) != FUNCTION_DECL) {
		error_at(DECL_P(*node) ? DECL_SOURCE_LOCATION(*node) : UNKNOWN_LOCATION, "%qE attribute applies to functions only (%qD)", name, *node);
		return NULL_TREE;
	}

	if (!split_context_attribute(args, &lock, &in, &out)) {
		error_at(DECL_SOURCE_LOCATION(*node), "%qE attribute needs two integers after the lock expression", name);
		return NULL_TREE;
	}

	if (TREE_CODE(in) != INTEGER_CST) {
		error_at(DECL_SOURCE_LOCATION(*node), "the 'in' argument of the %qE attribute must be an integer (%qE)", name, in);
		return NULL_TREE;
	}

	if (TREE_CODE(out) != INTEGER_CST) {
		error_at(DECL_SOURCE_LOCATION(*node), "the 'out' argument of the %qE attribute must be an integer (%qE)", name, out);
		return NULL_TREE;
	}

	*no_add_attrs = false;
	return NULL_TREE;
}

struct attribute_spec context_attr = ATTRIBUTE_SPEC_INIT("context", 2, 3, true, false, false, handle_context_attribute, true, NULL);

static const char context_function[] = "__context__";
GTY(()) tree context_function_decl;

static const char context_error[] = "__context_error__";
GTY(()) tree context_error_decl;

EXPORTED_CONST struct ggc_root_tab gt_ggc_r_gt_checker[] = {
	{
		.base = &context_function_decl,
		.nelt = 1,
		.stride = sizeof(context_function_decl),
		.cb = &gt_ggc_mx_tree_node,
		.pchw = &gt_pch_nx_tree_node
	},
	{
		.base = &context_error_decl,
		.nelt = 1,
		.stride = sizeof(context_error_decl),
		.cb = &gt_ggc_mx_tree_node,
		.pchw = &gt_pch_nx_tree_node
	},
	LAST_GGC_ROOT_TAB
};

void context_start_unit(void __unused *gcc_data, void __unused *user_data)
{
	tree fntype, attr;

	// void __context__(void *, int);
	fntype = build_function_type_list(void_type_node, ptr_type_node, integer_type_node, NULL_TREE);
	context_function_decl = build_fn_decl(context_function, fntype);

	TREE_PUBLIC(context_function_decl) = 1;
	TREE_USED(context_function_decl) = 1;
	DECL_EXTERNAL(context_function_decl) = 1;
	DECL_ARTIFICIAL(context_function_decl) = 1;
	DECL_PRESERVE_P(context_function_decl) = 1;
//	TREE_NOTHROW(context_function_decl) = 1;
//	DECL_UNINLINABLE(context_function_decl) = 1;
	DECL_ASSEMBLER_NAME(context_function_decl); // for LTO
	lang_hooks.decls.pushdecl(context_function_decl);

	invoke_plugin_callbacks(PLUGIN_PRE_GENERICIZE, context_function_decl);

	// void __context_error__(const void *, int) __attribute__((error("context error")));
	fntype = build_function_type_list(void_type_node, const_ptr_type_node, integer_type_node, NULL_TREE);
	context_error_decl = build_fn_decl(context_error, fntype);

	TREE_PUBLIC(context_error_decl) = 1;
	TREE_USED(context_error_decl) = 1;
	DECL_EXTERNAL(context_error_decl) = 1;
	DECL_ARTIFICIAL(context_error_decl) = 1;
	DECL_PRESERVE_P(context_error_decl) = 1;
//	TREE_NOTHROW(context_error_decl) = 1;
//	DECL_UNINLINABLE(context_error_decl) = 1;
	TREE_THIS_VOLATILE(context_error_decl) = 1;
	DECL_ASSEMBLER_NAME(context_error_decl);

	attr = tree_cons(NULL, build_const_char_string(14, "context error"), NULL);
	attr = tree_cons(get_identifier("error"), attr, NULL);
	decl_attributes(&context_error_decl, attr, 0);

	invoke_plugin_callbacks(PLUGIN_PRE_GENERICIZE, context_error_decl);
}

static basic_block verify_context_before(gimple_stmt_iterator *gsi, tree context, tree inout, tree error)
{
	gimple_stmt_ptr stmt;
	basic_block cond_bb, join_bb, true_bb;
	edge e;
	cgraph_node_ptr node;
	location_t loc;
	const char *file;
	int frequency, line;
	size_t len;
	tree filename, block;

	stmt = gsi_stmt(*gsi);
	cond_bb = gsi_bb(*gsi);

	block = gimple_block(stmt);
	if (!block) {
		gcc_assert(!bb_loop_depth(cond_bb));
		block = DECL_INITIAL(current_function_decl);
	}

	if (gimple_has_location(stmt)) {
		loc = gimple_location(stmt);
		file = gimple_filename(stmt);
		line = gimple_lineno(stmt);
	} else {
		loc = DECL_SOURCE_LOCATION(current_function_decl);
		file = DECL_SOURCE_FILE(current_function_decl);
		line = DECL_SOURCE_LINE(current_function_decl);
	}
	gcc_assert(file);

	// if (context != count) __context_error__(__FILE__, __LINE__);
	stmt = gimple_build_cond(NE_EXPR, context, inout, NULL_TREE, NULL_TREE);
	gimple_set_location(stmt, loc);
	gimple_set_block(stmt, block);
	gsi_insert_before(gsi, stmt, GSI_NEW_STMT);

	gcc_assert(!gsi_end_p(*gsi));
	gcc_assert(stmt == gsi_stmt(*gsi));

#if 1
	e = split_block(cond_bb, gsi_stmt(*gsi));
	gcc_assert(cond_bb == e->src);
	join_bb = e->dest;
	e->flags = EDGE_FALSE_VALUE;
	e->probability = REG_BR_PROB_EVEN;

	true_bb = create_empty_bb(EXIT_BLOCK_PTR_FOR_FN(cfun)->prev_bb);
	e = make_edge(cond_bb, true_bb, EDGE_TRUE_VALUE);
	e->probability = REG_BR_PROB_EVEN;
	true_bb->count = edge_count(e);

	e = make_single_succ_edge(true_bb, join_bb, EDGE_FALLTHRU);

	gcc_assert(dom_info_available_p(CDI_DOMINATORS));
	set_immediate_dominator(CDI_DOMINATORS, true_bb, cond_bb);
	set_immediate_dominator(CDI_DOMINATORS, join_bb, cond_bb);

	gcc_assert(cond_bb->loop_father == join_bb->loop_father);
	add_bb_to_loop(true_bb, cond_bb->loop_father);
#else
	true_bb = insert_cond_bb(cond_bb, gsi_stmt(*gsi), stmt, REG_BR_PROB_EVEN);
#endif
	loops_state_set(LOOPS_NEED_FIXUP);

	// insert call to builtin_trap or __context_error__
	*gsi = gsi_start_bb(true_bb);

#if 0
	error = builtin_decl_implicit(BUILT_IN_TRAP);
	stmt = gimple_build_call(error, 0);
#else
	len = strlen(file) + 1;
	filename = build_const_char_string(len, file);
	filename = build_fold_addr_expr_with_type_loc(loc, filename, const_ptr_type_node);
	stmt = gimple_build_call(error, 2, filename, build_int_cst(NULL_TREE, line));
#endif
	gimple_set_location(stmt, loc);
	gimple_set_block(stmt, block);
	gsi_insert_after(gsi, stmt, GSI_CONTINUE_LINKING);

	// update the cgraph
	node = cgraph_get_create_node(error);
	gcc_assert(node);
	frequency = compute_call_stmt_bb_frequency(current_function_decl, true_bb);
	cgraph_create_edge(cgraph_get_node(current_function_decl), node, as_a_gcall(stmt), true_bb->count, frequency, true_bb->loop_depth);

	*gsi = gsi_start_nondebug_bb(join_bb);
	return join_bb;
}

static void update_context(gimple_stmt_iterator *gsi, tree context, int diff)
{
	gimple_stmt_ptr assign;
	tree op;
	location_t loc = gimple_location(gsi_stmt(*gsi));

	op = fold_build2_loc(loc, PLUS_EXPR, integer_type_node, context, build_int_cst(integer_type_node, diff));
	assign = gimple_build_assign(context, op);
	gimple_set_location(assign, loc);
	gimple_set_block(assign, gimple_block(gsi_stmt(*gsi)));
	gsi_insert_after(gsi, assign, GSI_NEW_STMT);
	update_stmt(assign);
}

static basic_block track_context(basic_block bb, tree context)
{
	gimple_stmt_iterator gsi;

	// adjust context according to the context information on any call stmt
	for (gsi = gsi_start_bb(bb); !gsi_end_p(gsi); gsi_next(&gsi)) {
		gimple_stmt_ptr stmt = gsi_stmt(gsi);
		tree fndecl, context_attr;
		tree lock, in, out;
		int incount, outcount;

		if (!is_gimple_call(stmt))
			continue;

		fndecl = gimple_call_fndecl(stmt);
		if (!fndecl)
			continue;

		if (fndecl == context_function_decl) {
			unsigned int num_ops = gimple_num_ops(stmt);
			int diff = tree_to_shwi(gimple_op(stmt, num_ops - 1));

			gcc_assert(diff);
			update_context(&gsi, context, diff);
			gsi_prev(&gsi);
			gsi_remove(&gsi, true);
			continue;
		}

		context_attr = lookup_attribute("context", DECL_ATTRIBUTES(fndecl));
		if (!context_attr)
			continue;

		gcc_assert(split_context_attribute(TREE_VALUE(context_attr), &lock, &in, &out));
		incount = tree_to_shwi(in);
		outcount = tree_to_shwi(out);
		bb = verify_context_before(&gsi, context, in, context_error_decl);
		update_context(&gsi, context, outcount - incount);
	}

	return bb;
}

static unsigned int context_execute(void)
{
	basic_block bb;
	gimple_stmt_ptr assign;
	gimple_stmt_iterator gsi;
	tree context_attr, context;
	tree lock, in, out;

	if (seen_error())
		return 0;
#if BUILDING_GCC_VERSION >= 4009
	if (werrorcount)
		return 0;
#endif

	loop_optimizer_init(LOOPS_NORMAL | LOOPS_HAVE_RECORDED_EXITS);
	gcc_assert(current_loops);

	calculate_dominance_info(CDI_DOMINATORS);
	calculate_dominance_info(CDI_POST_DOMINATORS);

	context_attr = lookup_attribute("context", DECL_ATTRIBUTES(current_function_decl));
	if (context_attr) {
		gcc_assert(split_context_attribute(TREE_VALUE(context_attr), &lock, &in, &out));
	} else {
		lock = NULL_TREE;
		in = out = integer_zero_node;
	}

	// 1. create local context variable
	context = create_tmp_var(integer_type_node, "context");
	add_referenced_var(context);
	mark_sym_for_renaming(context);

	// 2. initialize local context variable
	gcc_assert(single_succ_p(ENTRY_BLOCK_PTR_FOR_FN(cfun)));
	bb = single_succ(ENTRY_BLOCK_PTR_FOR_FN(cfun));
	if (!single_pred_p(bb)) {
		gcc_assert(bb_any_loop(bb));
		plugin_split_edge(single_succ_edge(ENTRY_BLOCK_PTR_FOR_FN(cfun)));
		bb = single_succ(ENTRY_BLOCK_PTR_FOR_FN(cfun));
	}
	gsi = gsi_start_nondebug_after_labels_bb(bb);
	assign = gimple_build_assign(context, in);
//	gimple_set_location(assign, loc);
	gimple_set_block(assign, DECL_INITIAL(current_function_decl));
	gsi_insert_before(&gsi, assign, GSI_NEW_STMT);
	update_stmt(assign);

	// 3. instrument each BB to track the local context variable
	FOR_EACH_BB_FN(bb, cfun) {
		bb = track_context(bb, context);
	}

	// 4. verify the local context variable against the expected state
	if (EDGE_COUNT(EXIT_BLOCK_PTR_FOR_FN(cfun)->preds)) {
		gcc_assert(single_pred_p(EXIT_BLOCK_PTR_FOR_FN(cfun)));
		gsi = gsi_last_nondebug_bb(single_pred(EXIT_BLOCK_PTR_FOR_FN(cfun)));
		verify_context_before(&gsi, context, out, context_error_decl);
	}

	free_dominance_info(CDI_DOMINATORS);
	free_dominance_info(CDI_POST_DOMINATORS);
	loop_optimizer_finalize();
	return 0;
}

#define PASS_NAME context
#define NO_GATE
#define PROPERTIES_REQUIRED PROP_gimple_leh | PROP_cfg
//#define TODO_FLAGS_START TODO_verify_ssa | TODO_verify_flow | TODO_verify_stmts
#define TODO_FLAGS_FINISH TODO_verify_ssa | TODO_verify_stmts | TODO_dump_func | TODO_verify_flow | TODO_update_ssa
#include "gcc-generate-gimple-pass.h"
