/*
 * Written by Mathias Krause <minipli@grsecurity.net>
 * Copyright 2020 by Open Source Security, Inc.
 * Copyright 2020 by PaX Team
 * Licensed under the GPL v2
 *
 * make certain kernel constants available to gcc plugins
 */

#include "utilities.h"
#include "../gcc-ksyms.h"

static GTY(()) tree ksyms;

EXPORTED_CONST struct ggc_root_tab gt_ggc_r_gt_ksyms[] = {
	{
		.base = &ksyms,
		.nelt = 1,
		.stride = sizeof(ksyms),
		.cb = &gt_ggc_mx_tree_node,
		.pchw = &gt_pch_nx_tree_node
	},
#define DEFINE(sym, val)			\
	{					\
		.base = &(sym),			\
		.nelt = 1,			\
		.stride = sizeof(sym),		\
		.cb = &gt_ggc_mx_tree_node,	\
		.pchw = &gt_pch_nx_tree_node	\
	},
#define DEFINE_OPTIONAL(sym, val)		\
	{					\
		.base = &(sym),			\
		.nelt = 1,			\
		.stride = sizeof(sym),		\
		.cb = &gt_ggc_mx_tree_node,	\
		.pchw = &gt_pch_nx_tree_node	\
	},
#define DEFINE_ENUM(name, val)			\
	{					\
		.base = &(val),			\
		.nelt = 1,			\
		.stride = sizeof(val),		\
		.cb = &gt_ggc_mx_tree_node,	\
		.pchw = &gt_pch_nx_tree_node	\
	},
#define KSYMS_GENERATE_DEFINES
#include "ksyms.def"
	LAST_GGC_ROOT_TAB
};

/* generate ksyms' symbol trees, exported via gcc-ksyms.h */
#define DEFINE(sym, val) __visible tree /*GTY(())*/ sym;
#define DEFINE_OPTIONAL(sym, val) __visible tree /*GTY(())*/ sym;
#define DEFINE_ENUM(name, val) __visible tree /*GTY(())*/ val;
#define KSYMS_GENERATE_DEFINES
#include "ksyms.def"

static tree __unused ksyms_get_define_value(tree ksyms, const char *id, bool optional)
{
	tree index, value;
	unsigned int i;

	FOR_EACH_CONSTRUCTOR_ELT(CONSTRUCTOR_ELTS(DECL_INITIAL(ksyms)), i, index, value) {
		if (id_equal(DECL_NAME(index), id))
			return value;
	}

	if (!optional)
		error_at(DECL_SOURCE_LOCATION(ksyms), "missing definition of %qs", id);

	return NULL_TREE;
}

static tree __unused ksyms_get_enum_value(tree ksyms, const char *id)
{
	tree field;

	for (field = TYPE_FIELDS(TREE_TYPE(ksyms)); field; field = TREE_CHAIN(field)) {
		tree type;

		gcc_assert(DECL_NAME(field));
		if (strcmp(DECL_NAME_POINTER(field), id))
			continue;

		type = TREE_TYPE(field);
		if (TREE_CODE(type) != ENUMERAL_TYPE)
			continue;

		gcc_assert(TYPE_VALUES(type));
		for (field = TYPE_VALUES(type); field; field = TREE_CHAIN(field)) {
			if (strcmp(IDENTIFIER_POINTER(TREE_PURPOSE(field)), id))
				continue;

			return TREE_VALUE(field);
		}

		return NULL_TREE;
	}

	gcc_unreachable();
}

/* generate ksyms' symbols population code */
static void ksyms_populate(tree decl)
{
#define DEFINE(sym, val) sym = ksyms_get_define_value(decl, "ksym_" #sym, false);
#define DEFINE_OPTIONAL(sym, val) sym = ksyms_get_define_value(decl, "ksym_" #sym, true);
#define DEFINE_ENUM(name, val) val = ksyms_get_enum_value(decl, "ksym_" #val);
#define KSYMS_GENERATE_DEFINES
#include "ksyms.def"
	ksyms = decl;
}

void ksyms_find_decls(void *event_data, void *data __unused)
{
	tree decl = (tree)event_data;
	tree type, index, value, name;
	unsigned int i;

	if (seen_error())
		return;

	if (ksyms)
		return;

	if (decl == NULL_TREE || decl == error_mark_node)
		return;

	if (!VAR_P(decl))
		return;
	type = TREE_TYPE(decl);
	if (TREE_CODE(type) != RECORD_TYPE)
		return;

	name = DECL_NAME(decl);
	if (!name || TREE_CODE(name) != IDENTIFIER_NODE || !id_equal(name, "__plugin_ksyms"))
		return;

	name = TYPE_NAME(type);
	if (!name || TREE_CODE(name) != IDENTIFIER_NODE || !id_equal(name, "__plugin_ksyms"))
		return;

	/* sanity check type */
	if (DECL_INITIAL(decl) == NULL_TREE || TYPE_FIELDS(type) == NULL_TREE) {
		error("error parsing %qT", type);
		return;
	}

	FOR_EACH_CONSTRUCTOR_ELT(CONSTRUCTOR_ELTS(DECL_INITIAL(decl)), i, index, value) {
		gcc_assert(TREE_CODE(index) == FIELD_DECL);
		gcc_assert(TREE_CODE_CLASS(TREE_CODE(value)) == tcc_constant);

#if 0
		switch (TREE_CODE(value)) {
		default:
			debug_tree(index);
			debug_tree(value);
			gcc_unreachable();

		case INTEGER_CST:
			fprintf(stderr, ">>> added '%s' => %ld\n", DECL_NAME_POINTER(index), tree_to_shwi(value));
			break;

		case STRING_CST:
			fprintf(stderr, ">>> added '%s' => %s\n", DECL_NAME_POINTER(index), TREE_STRING_POINTER(value));
			break;
		}
#endif
	}

	ksyms_populate(decl);
	TREE_USED(decl) = 1;
}

#if BUILDING_GCC_VERSION < 4007
void ksyms_find_all_decls(void *event_data, void *data __unused)
{
	varpool_node_ptr vnode;

	FOR_EACH_VARIABLE(vnode)
		ksyms_find_decls(NODE_DECL(vnode), NULL);
}
#endif
